
/*-------------------------------------------------------------
  module: toesacQXsacZPXMPI.C
  author: Angel L. Diaz

  date: 12/5/96

  description: 
     Interpolates in parallel a factor of an nXn symetric toeplitz 
     determinant utilizing 
        - SACLIB's rational field and polynomial arithmetic for 
          black box construction 
        - SACLIB's modular arithmetic for black box evaluation
        - black box MPI interface for paralelization

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#include "PlugIns/saclib.h"
#include "PlugIns/maps.h"
#include "BlackBox/CommonObjects/bbtoeplitz.h"
#include "BlackBox/Aux/bbselectormpi.h"
#include "BlackBox/Aux/bbvector.h"
#include "BlackBox/Algorithms/bbfactors.h"
#include "BlackBox/HomomorphicMaps/bbfachmapmpi.h"
#include "BlackBox/Algorithms/bbsparse.h"
#include "BlackBox/MPI/bbservermpi.h"
#include "BlackBox/Aux/bbtime.h"

typedef BlackBoxSymToeDet< SaclibQ, SaclibQX >            BBSymToeDetSacQ;
typedef BlackBoxSymToeDet< SaclibZP, SaclibZPX >          BBSymToeDetSacZP;
typedef BlackBoxFactors< SaclibQ, SaclibQX, 
        BBSymToeDetSacQ >                                 BBFactors;
typedef BlackBoxFactorsHMapMPI< SaclibQ, SaclibQX, 
        SaclibZP, SaclibZPX, BBSymToeDetSacQ, 
        BBSymToeDetSacZP, SaclibQSaclibZP >               BBFactorsHMapMPI;
typedef BlackBoxSelectorMPI< SaclibZP, BBFactorsHMapMPI > BBFactorSacZPMPI;
typedef BlackBoxServerMPI< SaclibZP, SaclibZPX, 
        BBSymToeDetSacZP >                                BBServerMPI;

int main( int argc, char *argv[] ) 
{
   MPI_Init(&argc, &argv);
   Word stack;
   int    MyRank;                          // Rank of process 
   int    NumProc;                         // Number of processes 
   int    i,N, DegDet, DegVar,Fac,Seed,IsRestart;
   double Prob;
   char   FileName[BBDEFS_FILEZ];
   char   ProbName[BBDEFS_FILEZ];

   MP_INT MPCard;
   MPI_Comm_rank(MPI_COMM_WORLD, &MyRank);
   MPI_Comm_size(MPI_COMM_WORLD, &NumProc);

   mpz_init_set_ui( &MPCard, 100000007 ); 
   Prob      = 1.0;
   Seed      = 1;
   N         = 4;
   DegDet    = 4;
   DegVar    = 4;
   Fac       = 0;
   IsRestart = BBDEFS_FALSE;
   //IsRestart = 1;

   if ( MyRank == BBMPI_DRIVER_RANK ) {
      SaclibInitEnv(2000000,stack);
      // SaclibInitEnv(24000000,stack); // 99M
      SaclibSetPrime( &MPCard );
      BBSymToeDetSacQ   TSacQ( N, DegDet , Prob );
      BBSymToeDetSacZP  TSacZP( N, DegDet , Prob );

      sprintf( ProbName, "%dx%d%s", N, N, "toesacQXsacZPXMPI" );
      GetCPUTime( ProbName );

      BBFactors        FactorsTSacQ( TSacQ, Prob, Seed, &MPCard );
      SaclibQSaclibZP  h;
      BBFactorsHMapMPI FactorsTSacZPMPI( TSacZP, FactorsTSacQ, h );
      BBFactorSacZPMPI FirstFactTSacZP( FactorsTSacZPMPI, Fac ); 

      BBSPARSEAUX_VarVector Vars;
      BBSPARSEAUX_DegVector Degs;
      for( i=1; i<=N; i++ ) Vars.push_back(i);
      for( i=1; i<=N; i++ ) Degs.push_back(DegVar); 
      SaclibZPX SaclibZPXElm[BBSPARSEAUX_MAXTERMS];
      BBSPARSEAUX_VecDegVector PruneDegs;
      bbvector< SaclibZP >     PruneMons;
  
      SparseInterp( FirstFactTSacZP, Vars, Degs, DegDet, &MPCard, 
                    PruneDegs, PruneMons, SaclibZPXElm, ProbName, IsRestart,
                    NumProc );

      GetCPUTime( ProbName );
   }
   else {
      SaclibInitEnv(1000000,stack);
      // SaclibInitEnv(1000000,stack);  // 11M 
      SaclibSetPrime( &MPCard );
      BBSymToeDetSacZP  TSacZP( N, DegDet , Prob );
      sprintf( FileName, "%dx%d%s", N, N, "toesacQXsacZPXMPI" );
      BBServerMPI Server( BBMPI_DRIVER_RANK, MyRank, TSacZP, 
                          FileName, BBMPI_CLOCK_ON );
   }

   mpz_clear( &MPCard );
   SaclibCleanUpEnv();
   MPI_Finalize();
}

