
/*-------------------------------------------------------------
  module: toesacQXshoupZPX.C
  author: Angel L. Diaz

  date: 10/28/96

  description: 
     Interpolates a factor of an nXn symetric toeplitz determinant 
     utilizing SACLIB's rational field and polynomial arithmetic for
     black box construction and Shoup's modular arithmetic for
     black box evaluation. The mapping from SACLIB Q to Shoup ZP
     occurs via file I/O.

  arguments:
     N       - size of toeplitz matrix
     DegDet  - bound on the degree of the determiant
     DegVar  - bound on the degree of the variables

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#include "PlugIns/saclib.h"
#include "PlugIns/shoup.h"
#include "PlugIns/maps.h"
#include "BlackBox/CommonObjects/bbtoeplitz.h"
#include "BlackBox/Aux/bbselector.h"
#include "BlackBox/Aux/bbvector.h"
#include "BlackBox/Algorithms/bbfactors.h"
#include "BlackBox/HomomorphicMaps/bbfachmap.h"
#include "BlackBox/Algorithms/bbsparse.h"
#include "BlackBox/Aux/bbtime.h"

typedef BlackBoxSymToeDet< SaclibQ, SaclibQX >            BBSymToeDetSacQ;
typedef BlackBoxSymToeDet< SaclibZP, SaclibZPX >          BBSymToeDetSacZP;
typedef BlackBoxFactors< SaclibQ, SaclibQX, 
        BBSymToeDetSacQ >                                 BBFactors;
typedef BlackBoxFactorsHMap< SaclibQ, SaclibQX, SaclibZP, 
           SaclibZPX, BBSymToeDetSacQ, BBSymToeDetSacZP, 
           SaclibQSaclibZP >                              BBFactorsHMapSacZP;

typedef BlackBoxSymToeDet< ShoupZP, ShoupZPX >            BBSymToeDetShoupZP;
typedef BlackBoxFactorsHMap< ShoupZP, ShoupZPX, ShoupZP, 
           ShoupZPX, BBSymToeDetShoupZP, 
           BBSymToeDetShoupZP, SaclibQShoupZP >           BBFactorsHMapShoupZP;
typedef BlackBoxSelector< ShoupZP, BBFactorsHMapShoupZP > BBFactorShoupZP;

int main( int argc, char *argv[] ) 
{
   Word stack;
   SaclibInitEnv(1000000,stack);
   int    i,N, DegDet, DegVar,Fac,Seed,IsRestart;
   double Prob;
   char   FileName[BBDEFS_FILEZ];
   char   ProbName[BBDEFS_FILEZ];
   MP_INT MPCard;
   mpz_init_set_ui( &MPCard, 100000007 ); 

 
   if ( argc < 6 ) {
      cerr << "usage: n deg_det deg_var factor restart" << endl;
      SaclibCleanUpEnv();
      exit( -1 );
   }
   else {
      SaclibSetPrime( &MPCard );
      Prob   = 1.0;
      Seed   = 1;
      N      = atoi( argv[1] );
      DegDet = atoi( argv[2] );
      DegVar = atoi( argv[3] );
      Fac    = atoi( argv[4] );
      IsRestart = atoi( argv[5] ); // BBDEFS_FALSE indicates no restart

      sprintf( ProbName, "%dx%d%s", N, N, "toesacQXshoupZPX" );
      GetCPUTime( ProbName );

      //
      // construct utilizing SACLIB's rational arithmetic
      //

      BBSymToeDetSacQ  TSacQ( N, DegDet , Prob );
      BBSymToeDetSacZP TSacZP( N, DegDet , Prob );
      BBFactors        FactorsTSacQ( TSacQ, Prob, Seed, &MPCard );
      SaclibQSaclibZP  h;
      BBFactorsHMapSacZP    FactorsTSacZP( TSacZP, FactorsTSacQ, h );

      sprintf( FileName, "%s.ins", ProbName );
      ofstream OutFile( FileName, ios::out );
      if ( !OutFile ) {
         cerr << "toefacshmap : cannot open " << FileName
              << " for output" << endl;
         exit(-1);
      }

      OutFile << FactorsTSacZP;
      OutFile.close();
      SaclibCleanUpEnv();

      //
      // interpolate utilizing Shoup's modular arithmetic
      //

      ShoupInitEnv( &MPCard );

      BBSymToeDetShoupZP   TShoupZP( N, DegDet, Prob );
      BBFactorsHMapShoupZP FactorsTShoupZP( TShoupZP );

      ifstream InFile( FileName, ios::in );
      if ( !InFile ) {
         cerr << "toefacshmap : cannot open ``" << FileName
              << "'' for input" << endl;
      }

      InFile >> FactorsTShoupZP;
      InFile.close();

      BBFactorShoupZP FirstFacTShoupZP( FactorsTShoupZP, Fac );

      BBSPARSEAUX_VarVector  Vars;
      BBSPARSEAUX_DegVector  Degs;
      for( i=1; i<=N; i++ ) Vars.push_back(i);
      for( i=1; i<=N; i++ ) Degs.push_back(DegVar); 
      ShoupZPX ShoupZPXElm[BBSPARSEAUX_MAXTERMS];
      BBSPARSEAUX_VecDegVector PruneDegs;
      bbvector< ShoupZP >      PruneMons;

      SparseInterp( FirstFacTShoupZP, Vars, Degs, DegDet, &MPCard, 
                    PruneDegs, PruneMons, ShoupZPXElm, ProbName, IsRestart );

      mpz_clear( &MPCard );
      ShoupCleanUpEnv();
   }

   GetCPUTime( ProbName );

}

