
/*-------------------------------------------------------------
  module: toesacQXshoupZPXMap.C
  author: Angel L. Diaz

  date: 11/29/96

  description: 
     Interpolates a factor of an nXn symetric toeplitz determinant 
     utilizing SACLIB's rational field and polynomial arithmetic for
     black box construction and Shoup's modular arithmetic for
     black box evaluation. The mapping from SACLIB Q to Shoup ZP 
     occurs via a homorphic map.

  arguments:
     N       - size of toeplitz matrix
     DegDet  - bound on the degree of the determiant
     DegVar  - bound on the degree of the variables

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#include "PlugIns/saclib.h"
#include "PlugIns/shoup.h"
#include "PlugIns/maps.h"
#include "BlackBox/CommonObjects/bbtoeplitz.h"
#include "BlackBox/Aux/bbselector.h"
#include "BlackBox/Aux/bbvector.h"
#include "BlackBox/Algorithms/bbfactors.h"
#include "BlackBox/HomomorphicMaps/bbfachmap.h"
#include "BlackBox/Algorithms/bbsparse.h"
#include "BlackBox/Aux/bbtime.h"

typedef BlackBoxSymToeDet< SaclibQ, SaclibQX >            BBSymToeDetSacQ;
typedef BlackBoxSymToeDet< SaclibZP, SaclibZPX >          BBSymToeDetSacZP;
typedef BlackBoxFactors< SaclibQ, SaclibQX, 
   BBSymToeDetSacQ >                                      BBFactorsSacQ;
typedef BlackBoxSymToeDet< ShoupZP, ShoupZPX >            BBSymToeDetShoupZP;
typedef BlackBoxFactorsHMap< SaclibQ, SaclibQX, ShoupZP, 
   ShoupZPX, BBSymToeDetSacQ, BBSymToeDetShoupZP, 
   SaclibQShoupZP >                                       BBFactorsSacQHMapShoupZP;
typedef BlackBoxSelector< ShoupZP, 
   BBFactorsSacQHMapShoupZP >                             BBFactorShoupZP;

int main( int argc, char *argv[] ) 
{
   Word stack;
   SaclibInitEnv(1000000,stack);
   MP_INT MPCard;
   mpz_init_set_ui( &MPCard, 100000007 ); 
   //mpz_init_set_ui( &MPCard, 32771 );
   ShoupInitEnv( &MPCard );

   int    i,N, DegDet, DegVar,Fac,Seed,IsRestart;
   double Prob;
   char   ProbName[BBDEFS_FILEZ];
 
   if ( argc < 5 ) {
      cerr << "usage: n deg_det deg_var factor restart" << endl;
      SaclibCleanUpEnv();
      exit( -1 );
   }
   else {
      Prob   = 1.0;
      Seed   = 1;
      N      = atoi( argv[1] );
      DegDet = atoi( argv[2] );
      DegVar = atoi( argv[3] );
      Fac    = atoi( argv[4] );
      IsRestart = atoi( argv[5] ); // BBDEFS_FALSE indicates no restart

      sprintf( ProbName, "%dx%d%s", N, N, "toesacQXshoupZPXMap" );
      GetCPUTime( ProbName );

      //
      // construct utilizing SACLIB's rational arithmetic
      //

      BBSymToeDetSacQ BBToeQ( N, DegDet , Prob );
      BBFactorsSacQ   BBFactorsToeQ( BBToeQ, Prob, Seed, &MPCard );

      //
      // map to SHOUP's modular arithmetic
      //

      SaclibQShoupZP           h;
      BBSymToeDetShoupZP       BBToeZP( N, DegDet, Prob );
      BBFactorsSacQHMapShoupZP BBFactorsToeZP( BBToeZP, BBFactorsToeQ, h );
      BBFactorShoupZP          BBFirstFacToeZP( BBFactorsToeZP, Fac );

      SaclibCleanUpEnv();

      //
      // interpolate utilizing Shoup's modular arithmetic
      //

      BBSPARSEAUX_VarVector Vars;
      BBSPARSEAUX_DegVector  Degs;
      for( i=1; i<=N; i++ ) Vars.push_back(i);
      for( i=1; i<=N; i++ ) Degs.push_back(DegVar); 
      ShoupZPX ShoupZPXElm[BBSPARSEAUX_MAXTERMS];
      BBSPARSEAUX_VecDegVector PruneDegs;
      bbvector< ShoupZP >      PruneMons;

      SparseInterp( BBFirstFacToeZP, Vars, Degs, DegDet, &MPCard, 
                    PruneDegs, PruneMons, ShoupZPXElm, ProbName, IsRestart );

      ShoupCleanUpEnv();
   }

   mpz_clear( &MPCard );
   GetCPUTime( ProbName );

}

