
/*-------------------------------------------------------------
  module: bbfactors.h
  author: Angel L. Diaz

  date: 10/23/96

  description: 
     The class in this module realizes the black box factorization
     algorithm. In the FoxBox library, a factors black box object is a 
     function object which when supplied with values (a pointer to an 
     array of field elements) for the symbols, computes the values for 
     each of the factors of the input black box polynomial.

  class:
     BlackBoxFactors
     BlackBoxFactorsEx

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBFACTORS
#define _BBFACTORS

#include "BlackBox/Algorithms/bbfactoraux.h"

/*-------------------------------------------------------------
  class: BlackBoxFactors
  date: 10/23/96

  description:
     Provides for constructing a factors black box object.

-------------------------------------------------------------*/

template < class K, class P, class B > 
class BlackBoxFactors : public BlackBox< K > {
   private:
      B *BB;                     
      K RandA[BBFACTOR_MAXVAR];  
      K RandB[BBFACTOR_MAXVAR];  
      K RandC[BBFACTOR_MAXVAR];  
      MP_INT Card;               
      unsigned int Seed;         
      K::RandomIterator *RndIt;  
      K ProbeElm;             
      K GenElm;    
      int StaticDeg;
      K Ind[BBDENSE_MAXCOEFFS];
      K Reciprocal[BBDENSE_MAXCOEFFS];
      P Basis[BBDENSE_MAXCOEFFS];
      P FactorImages[BBFACTOR_MAXFACTORS]; 
      unsigned int NumFactors; 
      P FactorProducts[BBFACTOR_MAXFACTORS*BBFACTOR_MAXDEG]; 
      unsigned int Exps[BBFACTOR_MAXFACTORS];
      unsigned int SumExps;
      int Sign;
      K Cont;
      unsigned int DegF;
      P pgsem1; 
      unsigned int ParFracR;
      unsigned int ParFracC;
      P ParFrac[BBFACTOR_MAXDEG][BBFACTOR_MAXFACTORS];
      K ValueOfFactors[BBFACTOR_MAXFACTORS];
   public:
      inline friend ofstream& operator<<(ofstream&, 
         BlackBoxFactors< K, P, B >& );
      inline friend ifstream& operator>>(ifstream&, 
         BlackBoxFactors< K, P, B >& );

      BlackBoxFactors( ) { }
      BlackBoxFactors( B& );
      BlackBoxFactors( B&, double, unsigned int, MP_INT* );
      void Construct( B&, double, unsigned int, MP_INT* );
      ~BlackBoxFactors( );

      inline K* operator () ( K* );

      // static information access functions
      inline int GetStaticDeg()            { return StaticDeg;         }
      inline int GetNumFactors()           { return NumFactors;        }
      inline int GetSumExps()              { return SumExps;           }
      inline int GetSign()                 { return Sign;              }
      inline int GetDegF()                 { return DegF;              }
      inline int GetParFracR()             { return ParFracR;          }
      inline int GetParFracC()             { return ParFracC;          }
      inline int GetExps( int i )          { return Exps[i];           }
      inline K GetRandA( int i )           { return RandA[i];          }
      inline K GetRandB( int i )           { return RandB[i];          }
      inline K GetRandC( int i )           { return RandC[i];          }
      inline K GetProbeElm()               { return ProbeElm;          }
      inline K GetGenElm()                 { return GenElm;            }
      inline K GetInd( int i )             { return Ind[i];            }
      inline K GetReciprocal( int i )      { return Reciprocal[i];     }
      inline P GetBasis( int i )           { return Basis[i];          }
      inline K GetCont()                   { return Cont;              }
      inline P GetFactorImages( int i )    { return FactorImages[i];   }
      inline P GetFactorProducts( int i)   { return FactorProducts[i]; }
      inline P Getpgsem1()                 { return pgsem1;            }
      inline P GetParFrac( int i, int j )  { return ParFrac[i][j];     }
};

template < class K, class P, class B > 
BlackBoxFactors< K, P, B >::BlackBoxFactors(
   B& BB_i 
   )
{
   BB = &BB_i;
}

template < class K, class P, class B > 
BlackBoxFactors< K, P, B >::BlackBoxFactors(
   B& BB_i, 
   double   Prob_i = 1, 
   unsigned int Seed_i = 103069,
   MP_INT*  Card_i
   )
{
   Construct( BB_i, Prob_i, Seed_i, Card_i );
}

template < class K, class P, class B > 
void BlackBoxFactors< K, P, B >::Construct(
   B& BB_i, 
   double   Prob_i = 1, 
   unsigned int Seed_i = 103069,
   MP_INT*  Card_i
   )
{
   unsigned int i,j;  
   K YVal;
   P F1;
   P Factors;
   MP_INT MPZero, MPOne,MPSign;
   mpz_init_set_ui( &MPZero , 0 );
   mpz_init_set_ui( &MPOne , 1 );

   BB           = &BB_i;
   Degree       = -1;
   Probability  = Prob_i;
   Seed         = Seed_i; 
   NumVariables = BB_i.NumVars();
   mpz_init_set( &Card, Card_i );
   RndIt        = new K::RandomIterator( Seed, &Card );
   ProbeElm     = (*RndIt)();
   GenElm       = K( (MP_INT*)&MPOne );
   StaticDeg    = -1;
   YVal         = K( (MP_INT*)&MPZero );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace1(Probability, Seed, NumVariables, StaticDeg, ProbeElm, GenElm);
#endif

   // a2, ... , an    --> a0, ... , an-1
   // b1, ... , bn    --> b0, ... , bn-1
   // c1,c3, ... , cn --> c0, ... , cn-1
   for( i=0; i<NumVariables-1; i++ ) {
      RandA[i] = (*RndIt)();
      RandB[i] = (*RndIt)();
      RandC[i] = (*RndIt)();
   }
   RandB[NumVariables-1] = (*RndIt)();

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace2( RandA, RandB, RandC, NumVariables );
#endif

   MapFtoF1< K, P, B > BBF1( BB, RandA, RandB, NumVariables );
   UniDenseInterp( &BBF1, 0, YVal, BBDENSE_YVALUE, ProbeElm, GenElm,
                   StaticDeg, Ind, Reciprocal, Basis, F1 );

   F1.factors( F1, Sign, Cont, NumFactors, Exps, FactorImages );

   DegF = F1.dg();
   mpz_init_set_si( &MPSign , Sign ); 
   FactorImages[0] = FactorImages[0]*Cont * K( (MP_INT*) &MPSign );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace3( F1, Cont, FactorImages, Exps, Sign, NumFactors );
#endif

/* removed by EK 1/30/98; redundant check
   // check for unlucky random points
   for( i=0; i<NumFactors; i++ ) {
      for( j=0; j<NumFactors; j++ ) {
         if ( i < j ) {
            Factors.pgcd( FactorImages[i], FactorImages[j], Factors );
            if ( Factors != P( (MP_INT*)&MPOne ) ) 
               OutputErrorExit( BBERROR_FACTORS, "Unluky random points" );
         }
      }
   }
*/

   // compute the product of the factor images (with multiplicities)
   FactorProducts[0] = FactorImages[0]; 
   Exps[0] = Exps[0] - 1;
   for( SumExps=1,i=0; i<NumFactors; i++ ) 
       for( j=0; j<Exps[i]; j++,SumExps++ ) 
           FactorProducts[SumExps] = FactorProducts[SumExps-1]*FactorImages[i];
   Exps[0] = Exps[0] + 1;

   pgsem1 = P( (MP_INT*)&MPOne );
   for( i=0; i<NumFactors; i++ ) 
        for( j=0; j<Exps[i]-1; j++ ) 
            pgsem1 = FactorImages[i]*pgsem1;

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace4( SumExps, FactorProducts, pgsem1 );
#endif

   FindParFracCoeffs( FactorImages, Exps, NumFactors, ParFrac, ParFracR,
                      ParFracC );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace5( ParFrac, ParFracR, ParFracC );
#endif

   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
   mpz_clear( &MPSign );
}

template < class K, class P, class B > 
BlackBoxFactors< K, P, B >::~BlackBoxFactors( )
{
   mpz_clear( &Card );
}

template < class K, class P, class B > 
inline K* BlackBoxFactors< K, P, B >::operator( ) (
   K* Pnts_i
   )
{
   int         i,j;
   BiPoly< P > FBar;
   BiPoly< P > BiFactors[BBFACTOR_MAXFACTORS]; 
   K           EvalPnt;

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace6( Pnts_i, NumVariables );
#endif

   MapFtoFBar<K,P,B> BBFBar( BB, RandA, RandB, Pnts_i, NumVariables );

/*
   // general purpose bi-variate interpolation (w/degree guessing)
   BiDenseInterp( &BBFBar, ProbeElm, GenElm, StaticDeg, Ind, 
                  Reciprocal, Basis, FBar );
*/

   // special pruning bi-variate interpolation
   SpecialBiDenseInterp( &BBFBar, DegF, GenElm, StaticDeg, Ind, Reciprocal, 
                         Basis, FBar );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace7( FBar, FBar.Deg() );
#endif

   Lift( FBar, FactorImages, NumFactors, FactorProducts, SumExps, Exps,
         ParFrac, pgsem1, BiFactors );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace8( BiFactors, NumFactors );
#endif

   EvalPnt = Pnts_i[0] - RandB[0];
   for( i=0; i<NumFactors; i++) {
      ValueOfFactors[i] = (BiFactors[i])[0].eval( &EvalPnt );
      for( j=1; j<(BiFactors[i]).Deg(); j++ ) 
          ValueOfFactors[i] = ValueOfFactors[i] +
             (BiFactors[i])[j].eval( &EvalPnt );
      //ValueOfFactors[i] = ValueOfFactors[i] * (long)Sign;
   }

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace9( ValueOfFactors, NumFactors );
#endif

   return( ValueOfFactors );
}

template < class K, class P, class B > 
inline ofstream& operator<<(
   ofstream& of_i,
   BlackBoxFactors< K, P, B >& BB_i
   )
{
   int  i,j,k;
   char ZPElm[BBDEFS_KSTRZ];

   of_i << BBDEFS_FACTOR_ID   << endl;
   of_i << BB_i.NumVariables  << endl;
   of_i << BB_i.Degree        << endl;
   of_i << BB_i.Probability   << endl;
   of_i << BB_i.NumFactors    << endl;
   of_i << BB_i.SumExps       << endl;
   of_i << BB_i.Sign          << endl;
   of_i << BB_i.DegF          << endl;
   of_i << BB_i.ParFracR      << endl;
   of_i << BB_i.ParFracC      << endl;
   for( i=0; i<BB_i.NumFactors; i++ ) {
      of_i << BB_i.Exps[i] << endl;;
   }
   BB_i.ProbeElm.toString( ZPElm );
   of_i << ZPElm          << endl;
   of_i << BB_i.StaticDeg << endl;
   BB_i.GenElm.toString( ZPElm );
   of_i << ZPElm          << endl;
   for( i=0; i<=BB_i.StaticDeg; i++ ) {
      BB_i.Ind[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   for( i=1; i<=BB_i.StaticDeg; i++ ) {
      BB_i.Reciprocal[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   for( j=0; j<=BB_i.StaticDeg; j++ ) {
      of_i << BB_i.Basis[j].dg() << endl;
      for(i=0; i<=BB_i.Basis[j].dg(); i++ ) {
         (BB_i.Basis[j])[i].toString( ZPElm );
         of_i << ZPElm << endl;
      }
   }
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      BB_i.RandA[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      BB_i.RandC[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   for( i=0; i<BB_i.NumVariables; i++ ) {
      BB_i.RandB[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   BB_i.Cont.toString( ZPElm );
   of_i << ZPElm << endl;
   of_i << BB_i.pgsem1.dg() << endl;  
   for(i=0; i<=BB_i.pgsem1.dg(); i++ ) {
      BB_i.pgsem1[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   // factor images
   for( j=0; j<BB_i.NumFactors; j++ ) {
      of_i << BB_i.FactorImages[j].dg() << endl;
      for(i=0; i<=BB_i.FactorImages[j].dg(); i++ ) {
         (BB_i.FactorImages[j])[i].toString( ZPElm );
         of_i << ZPElm << endl;
      }
   }
   // factor products
   for( j=0; j<BB_i.SumExps; j++ ) {
      of_i << BB_i.FactorProducts[j].dg() << endl;
      for(i=0; i<=BB_i.FactorProducts[j].dg(); i++ ) {
         (BB_i.FactorProducts[j])[i].toString( ZPElm );
         of_i << ZPElm << endl;
      }
   }
   // partial fractions
   for( i=0; i<=BB_i.ParFracC; i++ ) {
      for( j=0; j<=BB_i.ParFracR; j++ ) {
         of_i << BB_i.ParFrac[i][j].dg() << endl;
         for( k=0; k<=BB_i.ParFrac[i][j].dg(); k++ ) {
            (BB_i.ParFrac[i][j])[k].toString( ZPElm );
            of_i << ZPElm << endl;
         } 
      } 
   }
}

template < class K, class P, class B > 
inline ifstream& operator>>(
   ifstream& if_i,
   BlackBoxFactors< K, P, B >& BB_i
   )
{
   int    i,j,k;
   int    dg,type;
   char   ZPElm[BBDEFS_KSTRZ];
   K      Coeff;
   MP_INT MPZero;
   mpz_init_set_ui( &MPZero, 0 );
 
   if_i >> type;
   if ( type != BBDEFS_FACTOR_ID ) {
      OutputErrorExit( BBERROR_FACTORS, "Incorrect instance file type" );
   }
   if_i >> BB_i.NumVariables;
   if_i >> BB_i.Degree;
   if_i >> BB_i.Probability;
   if_i >> BB_i.NumFactors;
   if_i >> BB_i.SumExps;
   if_i >> BB_i.Sign;
   if_i >> BB_i.DegF;
   if_i >> BB_i.ParFracR;
   if_i >> BB_i.ParFracC;
   for( i=0; i<BB_i.NumFactors; i++ ) {
      if_i >> BB_i.Exps[i];
   }
   if_i >> ZPElm; BB_i.ProbeElm = K( ZPElm );
   if_i >> BB_i.StaticDeg;
   if_i >> ZPElm; BB_i.GenElm = K( ZPElm );
   for( i=0; i<=BB_i.StaticDeg; i++ ) {
       if_i >> ZPElm; BB_i.Ind[i] = K( ZPElm );
   }
   for( i=1; i<=BB_i.StaticDeg; i++ ) {
      if_i >> ZPElm; BB_i.Reciprocal[i] = K( ZPElm );
   }
   for( j=0; j<=BB_i.StaticDeg; j++ ) {
      if_i >> dg;
      BB_i.Basis[j] = K( (MP_INT*)&MPZero ); 
      for(i=0; i<=dg; i++ ) {
         if_i >> ZPElm; Coeff = K( ZPElm );
         BB_i.Basis[j] = BB_i.Basis[j] + P( Coeff, i );
      }
   }
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      if_i >> ZPElm; BB_i.RandA[i] = K( ZPElm );
   }
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      if_i >> ZPElm; BB_i.RandC[i] = K( ZPElm );
   }
   for( i=0; i<BB_i.NumVariables; i++ ) {
      if_i >> ZPElm; BB_i.RandB[i] = K( ZPElm );
   }
   if_i >> ZPElm; BB_i.Cont = K( ZPElm );
   if_i >> dg;
   BB_i.pgsem1 = K( (MP_INT*)&MPZero ); 
   for(i=0; i<=dg; i++ ) {
      if_i >> ZPElm; Coeff = K( ZPElm );
      BB_i.pgsem1 = BB_i.pgsem1 + P( Coeff, i );
   }
   // factor images
   for( j=0; j<BB_i.NumFactors; j++ ) {
      if_i >> dg;
      BB_i.FactorImages[j] = K( (MP_INT*)&MPZero ); 
      for(i=0; i<=dg; i++ ) {
         if_i >> ZPElm; Coeff = K( ZPElm );
         BB_i.FactorImages[j] = BB_i.FactorImages[j] + P( Coeff, i );
      }
   }
   // factor products
   for( j=0; j<BB_i.SumExps; j++ ) {
      if_i >> dg;
      BB_i.FactorProducts[j] = K( (MP_INT*)&MPZero ); 
      for(i=0; i<=dg; i++ ) {
         if_i >> ZPElm; Coeff = K( ZPElm );
         BB_i.FactorProducts[j] = BB_i.FactorProducts[j] + P( Coeff, i );
      }
   }
   // partial fractions
   for( i=0; i<=BB_i.ParFracC; i++ ) {
      for( j=0; j<=BB_i.ParFracR; j++ ) {
         if_i >> dg;
         BB_i.ParFrac[i][j] = K( (MP_INT*)&MPZero );
         for(k=0; k<=dg; k++ ) {
            if_i >> ZPElm; Coeff = K( ZPElm );
            BB_i.ParFrac[i][j] = BB_i.ParFrac[i][j] + P( Coeff, k );
         }
      }
   }
   mpz_clear( &MPZero );
}

/*-------------------------------------------------------------
  class: BlackBoxFactorsEx
  date: 10/24/96

  description:
     Provides for constructing an extended domain factors 
     black box object.

-------------------------------------------------------------*/

template < class K, class P, class L, class B >
class BlackBoxFactorsEx : public BlackBoxEx< K,L > {
   private:
      B *BB;                     
      K RandA[BBFACTOR_MAXVAR];  
      K RandB[BBFACTOR_MAXVAR];  
      K RandC[BBFACTOR_MAXVAR];  
      MP_INT Card;               
      unsigned int Seed;         
      K::RandomIterator *RndIt;  
      K ProbeElm;             
      K GenElm;    
      int StaticDeg;
      K Ind[BBDENSE_MAXCOEFFS];
      K Reciprocal[BBDENSE_MAXCOEFFS];
      P Basis[BBDENSE_MAXCOEFFS];
      P FactorImages[BBFACTOR_MAXFACTORS]; 
      unsigned int NumFactors; 
      P FactorProducts[BBFACTOR_MAXFACTORS*BBFACTOR_MAXDEG]; 
      unsigned int Exps[BBFACTOR_MAXFACTORS];
      unsigned int SumExps;
      int Sign;
      K Cont;
      unsigned int DegF;
      P pgsem1; 
      unsigned int ParFracR;
      unsigned int ParFracC;
      P ParFrac[BBFACTOR_MAXDEG][BBFACTOR_MAXFACTORS];
      K ValueOfFactors[BBFACTOR_MAXFACTORS];
   public:
      BlackBoxFactorsEx( B&, double, unsigned int, MP_INT* );
      ~BlackBoxFactorsEx( );

      inline K* operator () ( K* );
      inline L* operator () ( L* );

      // static information access functions
      inline int GetStaticDeg()            { return StaticDeg;         }
      inline int GetNumFactors()           { return NumFactors;        }
      inline int GetSumExps()              { return SumExps;           }
      inline int GetSign()                 { return Sign;              }
      inline int GetDegF()                 { return DegF;              }
      inline int GetParFracR()             { return ParFracR;          }
      inline int GetParFracC()             { return ParFracC;          }
      inline int GetExps( int i )          { return Exps[i];           }
      inline K GetRandA( int i )           { return RandA[i];          }
      inline K GetRandB( int i )           { return RandB[i];          }
      inline K GetRandC( int i )           { return RandC[i];          }
      inline K GetProbeElm()               { return ProbeElm;          }
      inline K GetGenElm()                 { return GenElm;            }
      inline K GetInd( int i )             { return Ind[i];            }
      inline K GetReciprocal( int i )      { return Reciprocal[i];     }
      inline P GetBasis( int i )           { return Basis[i];          }
      inline K GetCont()                   { return Cont;              }
      inline P GetFactorImages( int i )    { return FactorImages[i];   }
      inline P GetFactorProducts( int i)   { return FactorProducts[i]; }
      inline P Getpgsem1()                 { return pgsem1;            }
      inline P GetParFrac( int i, int j )  { return ParFrac[i][j];     }
};

template < class K, class P, class L, class B > 
BlackBoxFactorsEx< K, P, L, B >::BlackBoxFactorsEx(
   B&           BB_i, 
   double       Prob_i = 1, 
   unsigned int Seed_i = 103069,
   MP_INT*      Card_i
   )
{
   unsigned int i,j;  
   K YVal;
   P F1;
   P Factors;
   MP_INT MPZero, MPOne,MPSign;
   mpz_init_set_ui( &MPZero , 0 );
   mpz_init_set_ui( &MPOne , 1 );

   BB           = &BB_i;
   Degree       = -1;
   Probability  = Prob_i;
   Seed         = Seed_i; 
   NumVariables = BB_i.NumVars();
   mpz_init_set( &Card, Card_i );
   RndIt        = new K::RandomIterator( Seed, &Card );
   ProbeElm     = (*RndIt)();
   GenElm       = K( (MP_INT*)&MPOne );
   StaticDeg    = -1;
   YVal         = K( (MP_INT*)&MPZero );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace1(Probability, Seed, NumVariables, StaticDeg, ProbeElm, GenElm);
#endif

   // a2, ... , an    --> a0, ... , an-1
   // b1, ... , bn    --> b0, ... , bn-1
   // c1,c3, ... , cn --> c0, ... , cn-1
   for( i=0; i<NumVariables-1; i++ ) {
      RandA[i] = (*RndIt)();
      RandB[i] = (*RndIt)();
      RandC[i] = (*RndIt)();
   }
   RandB[NumVariables-1] = (*RndIt)();

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace2( RandA, RandB, RandC, NumVariables );
#endif

   MapFtoF1Ex< K, P, L, B > BBF1( BB, RandA, RandB, NumVariables );
   UniDenseInterp( &BBF1, 0, YVal, BBDENSE_YVALUE, ProbeElm, GenElm,
                   StaticDeg, Ind, Reciprocal, Basis, F1 );
   F1.factors( F1, Sign, Cont, NumFactors, Exps, FactorImages );
   DegF = F1.dg();
   mpz_init_set_si( &MPSign , Sign ); 
   FactorImages[0] = FactorImages[0]*Cont * K( (MP_INT*) &MPSign );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace3( F1, Cont, FactorImages, Exps, Sign, NumFactors );
#endif

   // check for unlucky random points
   for( i=0; i<NumFactors; i++ ) {
      for( j=0; j<NumFactors; j++ ) {
         if ( i < j ) {
            Factors.pgcd( FactorImages[i], FactorImages[j], Factors );
            if ( Factors != P( (MP_INT*)&MPOne ) ) 
               OutputErrorExit( BBERROR_FACTORS, "Unluky random points" );
         }
      }
   }

   // compute the product of the factor images (with multiplicities)
   FactorProducts[0] = FactorImages[0]; 
   Exps[0] = Exps[0] - 1;
   for( SumExps=1,i=0; i<NumFactors; i++ ) 
       for( j=0; j<Exps[i]; j++,SumExps++ ) 
           FactorProducts[SumExps] = FactorProducts[SumExps-1]*FactorImages[i];
   Exps[0] = Exps[0] + 1;

   pgsem1 = P( (MP_INT*)&MPOne );
   for( i=0; i<NumFactors; i++ ) 
        for( j=0; j<Exps[i]-1; j++ ) 
            pgsem1 = FactorImages[i]*pgsem1;

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace4( SumExps, FactorProducts, pgsem1 );
#endif

   FindParFracCoeffs( FactorImages, Exps, NumFactors, ParFrac, ParFracR,
                      ParFracC );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace5( ParFrac, ParFracR, ParFracC );
#endif

   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
   mpz_clear( &MPSign );
}

template < class K, class P, class L, class B > 
BlackBoxFactorsEx< K, P, L, B >::~BlackBoxFactorsEx( ) {
   mpz_clear( &Card );
}

template < class K, class P, class L, class B > 
inline K* BlackBoxFactorsEx< K, P, L, B >::operator( ) (
   K* Pnts_i
   )
{
   int i,j;
   BiPoly< P > FBar;
   BiPoly< P > BiFactors[BBFACTOR_MAXFACTORS]; 
   L Pnts[BBFACTOR_MAXVAR];  // array of points projected to linear forms
   L FBarY[BBFACTOR_MAXDEG]; // array of FBar polynomials at different X values
   K Dep[BBFACTOR_MAXDEG];
   L      Y;
   K      EvalPnt;
   MP_INT MPOne;
   mpz_init_set_si( &MPOne, 1 );
   Y = L( (MP_INT*) &MPOne , 1 );

   // computer FBar(X,Y)

   // evalulate at distinct X values
   for( j=0; j<=DegF; j++ ) {
      Pnts[0] = Ind[j] + RandB[0];           
      // Y*(pn - an*(p1 - b1) - bn) + an*Ind[j] + bn
      for( i=1; i<NumVariables; i++ ) 
         Pnts[i] = Y*(Pnts_i[i]-(Pnts_i[0]-RandB[0])*RandA[i-1] - RandB[i]) +
                   RandA[i-1]*Ind[j] + RandB[i];

      FBarY[j] = (*(*BB)( Pnts ));
   }

   // interpolate to retrive the X coefficients of FBar(X,Y)
   for( i=0; i<=DegF; i++ ) {
      for( j=0; j<=DegF; j++ ) {
         Dep[j] = (FBarY[j])[i]; 
      }
      UniDenseInterp( Ind, Dep, DegF, StaticDeg, Reciprocal, Basis, FBar[i]); 
   }
   FBar.SetDeg( DegF );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace7( FBar, FBar.Deg() );
#endif

   // lift to univariate image to bi-variate projection of the factors 
   Lift( FBar, FactorImages, NumFactors, FactorProducts, SumExps, Exps,
         ParFrac, pgsem1, BiFactors );

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace8( BiFactors, NumFactors );
#endif

   // retrieve evaluation from lifted factors
   EvalPnt = Pnts_i[0] - RandB[0];
   for( i=0; i<NumFactors; i++) {
      ValueOfFactors[i] = (BiFactors[i])[0].eval( &EvalPnt );
      for( j=1; j<(BiFactors[i]).Deg(); j++ ) 
          ValueOfFactors[i] = ValueOfFactors[i] +
              (BiFactors[i])[j].eval( &EvalPnt );
   }

#ifdef BBDEFS_BBFACTOR_TRC
   BBFACSTrace9( ValueOfFactors, NumFactors );
#endif

   mpz_clear( &MPOne );
   return( ValueOfFactors );
}

template < class K, class P, class L, class B > 
inline L* BlackBoxFactorsEx< K, P, L, B >::operator( ) (
   L* Pnts_i
   )
{
}

#endif

