
/*-------------------------------------------------------------
  module: bbgcd.h
  author: Angel L. Diaz

  date: 10/30/96

  description: 
     The class in this module realizes the black box GCD algorithm. 
     In the FoxBox library, a GCD black box object is a 
     function object which when supplied with values (a pointer to an 
     array of field elements) for the symbols, computes the value of
     the greatest common divisor for each of the the input black 
     box polynomials.

  class:
     BlackBoxGCD

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBGCD
#define _BBGCD

#include "BlackBox/Algorithms/bbgcdaux.h"

/*-------------------------------------------------------------
  class: BlackBoxGCD
  date: 10/30/96

  reference: 
     A. Diaz and E. Kaltofen, "On computing greatest 
     common divisors with polynomials given by black boxes for 
     their evaluation", Proc. 1995 Internat. Symp. Symbolic 
     Algebraic Comput. ISSAC '95, ACM Press, A. H. M. Levelt ed., 
     pg. 232--239, 1995

  description: 
     Provides for constructing a GCD black box object.

-------------------------------------------------------------*/

template < class K, class P, class B > 
class BlackBoxGCD : public BlackBox< K > {
   private:
      B** BB;
      unsigned int NumBlackBoxes;
      unsigned int Seed;
      MP_INT Card;   
      unsigned int PolyDegs[BBGCD_MAXPOLYS];
      unsigned int Delta;
      K RandA[BBGCD_MAXVAR];  
      K RandB[BBGCD_MAXVAR];  
      K RandC[BBGCD_MAXBB];  
      P F0Bar;       
      P F1Bar;         
      P G0;    
      K LG0;    
      K::RandomIterator *RndIt; 
      K ProbeElm;   
      K GenElm;       
      int StaticDeg;
      K Ind[BBDENSE_MAXCOEFFS];
      K Reciprocal[BBDENSE_MAXCOEFFS];
      P Basis[BBDENSE_MAXCOEFFS];
      K GCDValue;
   public:
      BlackBoxGCD( );
      BlackBoxGCD( B**, unsigned int, MP_INT*, double, unsigned int );
      ~BlackBoxGCD( );
      void Construct( B**, unsigned int, MP_INT*, double, unsigned int );

      inline K* operator () ( K* );
};

template < class K, class P, class B > 
BlackBoxGCD< K, P,B >::BlackBoxGCD( ) {
}

template < class K, class P, class B > 
BlackBoxGCD< K, P, B >::BlackBoxGCD(
   B**          BB_i,
   unsigned int NumBlackBoxes_i, 
   MP_INT*      Card_i,
   double       Prob_i = 1, 
   unsigned int Seed_i = 103069
   )
{
   Construct( BB_i, NumBlackBoxes_i, Card_i, Prob_i, Seed_i );
}

template < class K, class P, class B > 
void BlackBoxGCD< K, P, B >::Construct(
   B**          BB_i,
   unsigned int NumBlackBoxes_i, 
   MP_INT*      Card_i,
   double       Prob_i = 1, 
   unsigned int Seed_i = 103069
   )
{
   K                    YVal;
   K                    Pnts[BBGCD_MAXVAR];  
   MP_INT               MPOne, MPZero;
   P                    BarPoly;
   MapFtoGCDFBar<K,P,B> BBBar;
   int                  i;
 
   BB            = BB_i;
   NumBlackBoxes = NumBlackBoxes_i;
   mpz_init_set( &Card, Card_i );
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   Seed         = Seed_i;
   Probability  = Prob_i;
   NumVariables = (*(BB_i[0])).NumVars();
   RndIt        = new K::RandomIterator( Seed, &Card );
   ProbeElm     = (*RndIt)();
   GenElm       = K( (MP_INT*)&MPZero );
   StaticDeg    = -1;
   YVal         = K( (MP_INT*)&MPZero );

   // pick random field elements
   // a2, ... , an
   // b2, ... , bn
   for( i=0; i<NumVariables-1; i++ ) {
      RandA[i] = (*RndIt)();
      RandB[i] = (*RndIt)();
   }
   // c3, ... , cr
   for( i=0; i<NumBlackBoxes-2; i++) 
      RandC[i] = (*RndIt)();

   // p1, ... , pn
   for( i=0; i<NumVariables; i++)
      Pnts[i] = K( (MP_INT*)&MPZero );

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace1( Probability, Seed, NumVariables, ProbeElm, GenElm, StaticDeg, 
                RandA, RandB, NumVariables-1, RandC, NumBlackBoxes-2 );
#endif

   // interpolate F1Bar( X, 0 )
   BBBar.SetF( BB[0], RandA, RandB, Pnts, NumVariables );
   UniDenseInterp( &BBBar, 0, YVal, BBDENSE_YVALUE, ProbeElm, GenElm, 
                   StaticDeg, Ind, Reciprocal, Basis, F1Bar ); 
   PolyDegs[0] = F1Bar.dg();

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace2( F1Bar, GenElm, StaticDeg );
#endif

   // compute F0Bar( X, 0 )
   BBBar.SetF( BB[1], RandA, RandB, Pnts, NumVariables ); 
   UniDenseInterp( &BBBar, 0, YVal, BBDENSE_YVALUE, ProbeElm, GenElm,
                   StaticDeg, Ind, Reciprocal, Basis, F0Bar ); 
   PolyDegs[1] = F0Bar.dg();

   for( i=2; i<NumBlackBoxes; i++ ) {
      BBBar.SetF( BB[i], RandA, RandB, Pnts, NumVariables );
      UniDenseInterp( &BBBar, 0, YVal, BBDENSE_YVALUE, ProbeElm, GenElm,
                      StaticDeg, Ind, Reciprocal, Basis, BarPoly ); 
     PolyDegs[i] = F0Bar.dg();
     F0Bar = F0Bar + BarPoly * RandC[i-2];
   }

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace3( F0Bar, GenElm, StaticDeg );
#endif

   // gcd( F0Bar(X,0), F1Bar(X,0) )
   G0.pgcd( F0Bar, F1Bar, G0 );
   Delta = G0.dg();
   LG0   = G0.lcoeff();

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace4( G0, LG0, Delta );
#endif
 
   mpz_clear( &MPOne );
   mpz_clear( &MPZero );
}

template < class K, class P, class B > 
BlackBoxGCD< K, P, B >::~BlackBoxGCD( ) {
   mpz_clear( &Card );
}

template < class K, class P, class B > 
inline K* BlackBoxGCD< K, P, B >::operator () (
   K* Pnts_i
   )
{
   MP_INT MPOne, MPZero;
   K                    Zero;
   K                    One;
   K                    Indep[BBGCD_MAXDELTA]; 
   K                    Dep[BBGCD_MAXDELTA]; 
   P                    TmpPoly;
   K                    TmpElm;
   unsigned int         NumValidGCDs = 1; 
   MapFtoGCDFBar<K,P,B> BBBar;
   P                    BarPoly;
   P                    TmpF0Bar;
   P                    TmpF1Bar;
   P                    TmpGCD;
   unsigned int         TmpDeg;
   K                    YVal;
   struct GCDPair {
      P GCD;
      K Y;
   }                    ValidGCDImages[BBGCD_MAXDELTA];
   int i,j;

   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   One  = K( (MP_INT*)&MPOne );
   Zero = K( (MP_INT*)&MPZero );
   YVal = One;
   GCDValue = One;
   TmpElm = One;

   // gcd( f0Bar(X,0), f1Bar(X,0) ) is our first valid GCD
   ValidGCDImages[0].GCD = G0;
   ValidGCDImages[0].Y   = Zero;

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace5( Pnts_i, NumVariables, Delta );
#endif

   // compute Delta many valid GCD's
   while ( NumValidGCDs <= Delta ) {
      // interpolate F1Bar( X, YVal )
      BBBar.SetF( BB[0], RandA, RandB, Pnts_i, NumVariables );
      UniDenseInterp( &BBBar, PolyDegs[0], YVal, BBDENSE_YVALUE, ProbeElm, 
                      GenElm, StaticDeg, Ind, Reciprocal, Basis, TmpF1Bar ); 

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace6( TmpF1Bar, TmpF1Bar.dg(), GenElm, StaticDeg );
#endif

      // compute F0Bar( X, YVal )
      BBBar.SetF( BB[1], RandA, RandB, Pnts_i, NumVariables ); 
      UniDenseInterp( &BBBar, PolyDegs[1], YVal, BBDENSE_YVALUE, ProbeElm, 
                      GenElm, StaticDeg, Ind, Reciprocal, Basis, TmpF0Bar ); 

      for( i=2; i<NumBlackBoxes; i++ ) {
         BBBar.SetF( BB[i], RandA, RandB, Pnts_i, NumVariables );
         UniDenseInterp( &BBBar, PolyDegs[i], YVal, BBDENSE_YVALUE, ProbeElm, 
                         GenElm, StaticDeg, Ind, Reciprocal, Basis, BarPoly ); 
         TmpF0Bar = TmpF0Bar + BarPoly * RandC[i-2];
      }

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace7( TmpF0Bar, TmpF0Bar.dg(), GenElm, StaticDeg, NumValidGCDs );
#endif

      TmpGCD.pgcd( TmpF0Bar, TmpF1Bar, TmpGCD );
      TmpDeg = TmpGCD.dg();
      TmpGCD = ( TmpGCD / TmpGCD.lcoeff() ) * LG0;

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace8( TmpGCD, TmpGCD.dg() );
#endif

      // checking for early termination
      if ( ( YVal == One ) && ( TmpDeg == Delta ) ) {

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace9( );
#endif

         GCDValue = TmpGCD.eval( Pnts_i );
         goto XDONE;
      }
      // do we have an unlucky black box
      if ( TmpDeg < Delta ) {
         OutputErrorExit( BBERROR_GCD, "Unlucky black box GCD" );
      }
      else 
         // we have a valid GCD
         if (  TmpDeg = Delta ) {
#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace10( );
#endif
            ValidGCDImages[NumValidGCDs].GCD = TmpGCD;
            ValidGCDImages[NumValidGCDs].Y   = YVal;
            NumValidGCDs++;
         }
         // discard point
         else {
#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace11( );
#endif
         }

      YVal.DistinctSelect(); 
   }

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace12( ValidGCDImages, NumValidGCDs );
#endif

   for( j=1; j<NumValidGCDs; j++ ) 
      GCDValue = GCDValue * Pnts_i[0];

   for( j=0; j<NumValidGCDs-1; j++ ) {
      for( i=0; i<NumValidGCDs; i++ ) {
         Indep[i] = ValidGCDImages[i].Y;
         Dep[i]   = ValidGCDImages[i].GCD[j];
      }
      TmpPoly = Zero;
      UniDenseInterp( Indep, Dep, Delta, StaticDeg, Reciprocal,
                      Basis, TmpPoly ); 
      GCDValue = GCDValue + TmpElm * TmpPoly.eval( &One );
      TmpElm = TmpElm * Pnts_i[0];
   }

   XDONE:

#ifdef BBDEFS_BBGCD_TRC
   BBGCDTrace13( GCDValue );
#endif

   mpz_clear( &MPOne );
   mpz_clear( &MPZero );
   return( &GCDValue );
}

#endif

