
/*-------------------------------------------------------------
  module: bbnumden.h
  author: Angel L. Diaz

  date: 11/1/96

  description: 
     The class in this module realizes the black box numerator/denominator
     separation algorithm. In the FoxBox library, a numerator/denominator 
     black box object is a function object which when supplied with values 
     (a pointer to an array of field elements) for the symbols, computes the 
     value for the numerator and denominator the input black box rational
     function.

  class:
     BlackBoxNumDen

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBNUMDEN
#define _BBNUMDEN

#include "BlackBox/Algorithms/bbnumdenaux.h"

/*-------------------------------------------------------------
  class: BlackBoxNumDen
  date: 11/1/96

  reference: 
     E. Kaltofen and B. Trager, "On Computing with Polynomials 
     Given By Black Boxes for Their Evaluations: Greatest Common 
     Divisors, Factorization, Separation of Numerator and 
     Denominators", J. Sym. Comp. (1990), 9/3, pg. 312--315. 

  description: 
     Provides for constructing a numerator/denominator black box object

-------------------------------------------------------------*/

template < class K, class P, class B > 
class BlackBoxNumDen : public BlackBox< K > {
   private:
      B* BB;
      unsigned int Seed;
      MP_INT Card;   
      K RandA[BBNUMDENAUX_MAXVARS];  
      K RandB[BBNUMDENAUX_MAXVARS];  
      P LinForms[BBNUMDENAUX_MAXVARS];  
      int D;
      int E;
      int EHat;                  // bound on deg( g(x1,...,xn) )
      K LC;
      P FHat;
      P GHat;
      K::RandomIterator *RndIt; 
      K GenElm; 
      int StaticDeg;
      K Ind[BBNUMDENAUX_MAXCOEFFS];
      K Reciprocal[BBNUMDENAUX_MAXCOEFFS];
      P Basis[BBNUMDENAUX_MAXCOEFFS];
      K ValueOfNumDen[2];
   public:
      BlackBoxNumDen( );
      BlackBoxNumDen( B&, MP_INT*,  unsigned int, double, unsigned int );
      ~BlackBoxNumDen( );

      inline K* operator () ( K* );
      // static information access functions
      inline int GetStaticDeg()            { return StaticDeg;         }
      inline int GetD()                    { return D;                 }
      inline int GetE()                    { return E;                 }
      inline int GetEHat()                 { return EHat;              }
      inline K GetRandA( int i )           { return RandA[i];          }
      inline K GetRandB( int i )           { return RandB[i];          }
      inline K GetGenElm()                 { return GenElm;            }
      inline K GetLC()                     { return LC;                }
      inline K GetInd( int i )             { return Ind[i];            }
      inline K GetReciprocal( int i )      { return Reciprocal[i];     }
      inline P GetBasis( int i )           { return Basis[i];          }
      inline P GetLinForms( int i )        { return LinForms[i];       }
      inline P GetFHat()                   { return FHat;              }
      inline P GetGHat()                   { return GHat;              }
};

template < class K, class P, class B > 
BlackBoxNumDen< K, P, B >::BlackBoxNumDen( ) {
}

template < class K, class P, class B > 
BlackBoxNumDen< K, P, B >::BlackBoxNumDen(
   B&           BB_i,
   MP_INT*      Card_i,
   unsigned int EHat_i,              // bound on deg( g )
   double       Prob_i = 1, 
   unsigned int Seed_i = 103069 
   )
{
   int    i;
   int    Deg;
   MP_INT MPZero,MPOne;
   K      RndPnts[BBNUMDENAUX_MAXVARS];  
   K      FHatCoeffs[BBNUMDENAUX_MAXDEG];
   K      ValRatFunc;
   K      ValAprox;

   BB = &BB_i;
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   mpz_init_set( &Card, Card_i );
   EHat         = EHat_i;
   Seed         = Seed_i;
   Probability  = Prob_i;
   NumVariables = BB_i.NumVars();
   RndIt        = new K::RandomIterator( Seed, &Card );
   GenElm       = K( (MP_INT*)&MPOne );
   LC           = K( (MP_INT*)&MPOne );
   StaticDeg    = -1;
   P X( (MP_INT*)&MPOne, 1 );
   K Zero( (MP_INT*)&MPZero );
 
   // pick random field elements
   // a2, ... , an
   // b2, ... , bn
   for( i=0; i<NumVariables-1; i++ ) {
      RandA[i] = (*RndIt)();
      RandB[i] = (*RndIt)();
   }

   // compute the n linear forms X, a2*X + b2, ..., an*X + bn
   LinForms[0] = X;
   for( i=1; i<NumVariables; i++ )
      LinForms[i] = X*RandA[i-1] + RandB[i-1];

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace1( Probability, Seed, NumVariables, GenElm, StaticDeg, RandA,
                 RandB, NumVariables-1, LinForms, NumVariables );
#endif

   // probabilistically compute the degrees of f and g
   Deg = 0;
   while ( BBDEFS_TRUE ) {

      D = Deg;
      E = Deg; 

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace7( D, D, E );
#endif

      Pade( BB, D, E, LinForms, LC, GenElm, StaticDeg, Ind, Reciprocal,
            Basis, FHat, GHat );
      if ( !(GHat == Zero) ) {
         // pick another set of random points
         RndPnts[0] = (*RndIt)();
         for( i=1; i<NumVariables; i++ )
            RndPnts[i] = RandA[i-1]*RndPnts[0] + RandB[i-1];
         // MUST ADD A CHECK FOR PNT THAT MAPS g(x1,...,xn) TO ZERO
         ValRatFunc = *((*BB)( RndPnts ));
         ValAprox   = FHat.eval(&RndPnts[0])/GHat.eval(&RndPnts[0]);

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace9( NumVariables, RndPnts, FHat, GHat, ValRatFunc, ValAprox );
#endif

         if ( ValRatFunc == ValAprox ) {
            D = FHat.dg();
            E = GHat.dg();
            break;
         }
      }
      Deg++;
      if ( Deg > EHat )
         OutputErrorExit( BBERROR_NUMDEN, 
                          "Construction : Unluky random points" ); 
   }

   FHatCoeffs[0] = FHat[0].Denom(); 
   for( i=1; i <=D; i++ )
      FHatCoeffs[i] = FHat[i].Denom();

   LC = LC.LCM( FHatCoeffs, D+1 );

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace2( FHat, GHat, FHatCoeffs, D, E, LC );
#endif

   FHat = FHat * LC;
   GHat = GHat * LC;

   LC = GHat.lcoeff();

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace3( FHat, GHat );
#endif

   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
}

template < class K, class P, class B > 
BlackBoxNumDen< K, P, B >::~BlackBoxNumDen( ) {
   mpz_clear( &Card );
}

template < class K, class P, class B > 
inline K* BlackBoxNumDen< K, P, B >::operator()(
   K* Pnts_i
   )
{
   int i,j,k;
   unsigned int NumYVals, NumYCandidates, MaxCandidates;
   MP_INT MPZero,MPOne;
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   P TmpFHat,TmpGHat,TmpPoly;
   K TmpElm; 
   P FBarArray[BBNUMDENAUX_MAXDEG];
   P GBarArray[BBNUMDENAUX_MAXDEG];
   K ConstPnts[BBNUMDENAUX_MAXVARS];   // pi - ai*p1 - bi, i=1,...,n
   P FullLinForms[BBNUMDENAUX_MAXVARS];  
   K YArray[BBNUMDENAUX_MAXDEG];
   K Dep[BBNUMDENAUX_MAXDEG];
   K I;
   K Zero,One;
 
   // compute Y multipliers
   for( i=0; i<NumVariables-1; i++ ) 
      ConstPnts[i] = Pnts_i[i+1] - RandA[i]*Pnts_i[0] - RandB[i]; 

   FBarArray[0]  = FHat;
   GBarArray[0]  = GHat;
   Zero          = K( (MP_INT*)&MPZero );
   One           = K( (MP_INT*)&MPOne );
   YArray[0]     = Zero;
   I             = One;
   NumYVals      = BBDEFS_MAX( D, E ) + 1;
   MaxCandidates = 2*D*E;
   StaticDeg     = -1;
 
#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace4( Pnts_i, NumVariables, ConstPnts, NumYVals, MaxCandidates );
#endif

   // compute the univariate images of FBar and GBar
   FullLinForms[0] = LinForms[0];
   j = 0; NumYCandidates = 1;
   while ( BBDEFS_TRUE ) {

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace8( NumYCandidates );
#endif

      for( i=1; i<NumVariables; i++ ) 
         FullLinForms[i] = LinForms[i] + I*ConstPnts[i-1];
      Pade( BB, D, E, FullLinForms, LC, GenElm, StaticDeg, Ind, Reciprocal,
            Basis, TmpFHat, TmpGHat );
      if ( TmpFHat.dg() >= D && TmpGHat.dg() >= E ) {
         YArray[j+1]    = I;
         FBarArray[j+1] = TmpFHat;
         GBarArray[j+1] = TmpGHat;
         if ( I == One ) {
            ValueOfNumDen[0] = FBarArray[j+1].eval( &Pnts_i[0] );
            ValueOfNumDen[1] = GBarArray[j+1].eval( &Pnts_i[0] );
#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace10();
#endif 
            goto XDONE;
         }
         j++;
      }
      NumYCandidates++;
      I.DistinctSelect();
      if ( j == NumYVals ) 
         break;
      if ( NumYCandidates == MaxCandidates )
         OutputErrorExit( BBERROR_NUMDEN, 
                          "Evaluation : Unluky random points" ); 
   }

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace5( FBarArray, GBarArray, YArray, j+1 );
#endif 

   StaticDeg = -1;
   // interpolate FBar 
   ValueOfNumDen[0] = Zero;
   TmpElm = One;
   for( i=0; i<=j; i++ ) {
      for( k=0; k<=j; k++ ) {
         Dep[k] = (FBarArray[k])[i];
      }
      UniDenseInterp( YArray, Dep, j, StaticDeg, Reciprocal,
                      Basis, TmpPoly ); 
      ValueOfNumDen[0] = ValueOfNumDen[0] + TmpElm * TmpPoly.eval( &One );
      TmpElm = TmpElm * Pnts_i[0];
   }  

   // interpolate GBar
   ValueOfNumDen[1] = Zero;
   TmpElm = One;
   for( i=0; i<=j; i++ ) {
      for( k=0; k<=j; k++ ) {
         Dep[k] = (GBarArray[k])[i];
      }
      UniDenseInterp( YArray, Dep, j, StaticDeg, Reciprocal,
                      Basis, TmpPoly ); 
      ValueOfNumDen[1] = ValueOfNumDen[1] + TmpElm * TmpPoly.eval( &One );
      TmpElm = TmpElm * Pnts_i[0];
   }  
 
  XDONE:

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace6( ValueOfNumDen[0], ValueOfNumDen[1] );
#endif

   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
 
   return( ValueOfNumDen );
}

#endif

