/*-------------------------------------------------------------
  module: bbsparse.h
  author: Angel L. Diaz
  date: 8/14/96
-------------------------------------------------------------*/

#ifndef BBSPARSE
#define BBSPARSE

#include "BlackBox/Algorithms/bbsparseaux.h"

/*-------------------------------------------------------------
  function: SparseInterp
  date: 8/14/96

  reference: the varibale names mirror those utilized in  the sparse 
     conversion algorithm detailed in "Factorization of polynomials 
     given by straight-line programs", E. Kaltofen, Randomness and 
     Computation Vol 5, Advances in Computing Research, S. Micali ed., 
     Greenwhich, Connecticut, 1989, pg 387--394

  description: This is an improved version of E. Kaltofen's sparse 
     conversion algorithm. This improved version is due to A. Diaz
     and E. Kaltofen. It takes advantage of what we call a 
     "pruning" of monomials and a method for rapidly determining the 
     degree of such monomials. A formal referece is TBA.

  assumptions: The variable number BBSPARSEAUX_RESERVEDVAR is not
     utilized by the black box polynomial
-------------------------------------------------------------*/

template< class K, class P, class B > void SparseInterp(
   B&                     bb_i,         // black box polynomial 
   BBSPARSEAUX_DegVector& VarVector_i,  // vector of interpolation var. nums. 
   BBSPARSEAUX_DegVector& DegVector_i,  // vector of exact degress of each var
   unsigned int           DegPoly_i,    // total deg of the poly rep. by the 
                                        // black box 
   MP_INT*                Card_i,       // card of the set from which to pick 
                                        // random elements
   BBSPARSEAUX_VecDegVector& PruneDegs, // exponent vector of pruned polynomial
   bbvector< K >&            PruneMons, // monomial vector of pruned polynomial
   P*                     P_i,
   char*                  ProbName_i,
   int&                   IsRestart_m,
   int                    NumProc_i = 1 //
)
{
   unsigned int ii,e,d;                // loop variables 
   unsigned int i;                     // loop over each variable 
   unsigned int k;                     // loop over augmented Ji vector
   unsigned int d0;                    // bound on all of the degs of each var 
   unsigned int ji;                    // size of linear system (ji x ji) 
   unsigned int T;                     // number of terms 
   unsigned int PruneDeg;              //
   int          IsDoneEarly;
   K            PruneVal;              // pruned polynomial evaluated at
                                       // (p0)^k,...,(pi)^k,ai+1,...,an
   BBSPARSEAUX_DegVector VarVector;    // augmented vector of variables
   unsigned int NumVars; 
   K::RandomIterator *RndIt;           // random field elment generator 
   K A[BBSPARSEAUX_MAXVARS];           // initial random elements 
   K P[BBSPARSEAUX_MAXVARS];           // random evaluation points p0,...,pi
   K InputPnts[BBSPARSEAUX_MAXVARS];   // (p0)^k,...,(pi)^k,ai+1,...,an 
   K BBVal;                            // black box evaluated at
                                       // (p0)^k,...,(pi)^k,ai+1,...,an
   unsigned int BBCnt;                 // number of black box calls
   bbvector< K > gki;                  // adjusted black box evaluation
   bbvector< K > X;                    // solution vector 
   bbvector< K > V;                    // describes a trans. vandermonde matrix
   BBSPARSEAUX_VecDegVector Ji;        // vector of exponent vectors 
   BBSPARSEAUX_VecDegVector nJi;       // full vector of exponent vectors 
   bbvector< unsigned int > Ell;       // vector of of "remaining degree" 
                                       // for each monomial 
#ifdef BBDEFS_MPI_ON
   K PruneVals[BBSPARSEAUX_MAXJI];
   K BBVals[BBSPARSEAUX_MAXJI];
   int IsDist[BBSPARSEAUX_MAXJI];
   int Status;
   int RemainingEvals;
   int NumProc;
   int j,kk;

   for(j=0;j<BBSPARSEAUX_MAXJI; j++)
      IsDist[j] = BBDEFS_FALSE;
#endif

   T           = 1;
   BBCnt       = 0;
   IsDoneEarly = BBDEFS_FALSE;
   d0          = MaxDegree( DegVector_i );
   VarVector.push_back( BBSPARSEAUX_RESERVEDVAR );
   VarVector.insert( VarVector.begin() + 1, VarVector_i.begin(), 
                     VarVector_i.end() );
   NumVars = VarVector.size();
   RndIt   = new K::RandomIterator( 100, Card_i );

   for(i=1; i<NumVars; i++ ) {
      A[i]         = (*RndIt)(); 
      InputPnts[i] = A[i]; 
   }
   
#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace1( VarVector, NumVars, DegVector_i, d0 , A, InputPnts );
#endif
   if ( T > BBVECTOR_MAXSIZE ) {
      OutputErrorExit( BBERROR_SPARSE, 
                       "number of terms (T) > BBVECTOR_MAXSIZE" );
   }

   // interpolate one more variable
   for( i=0; i<NumVars; i++ ) {

      if ( IsRestart_m != BBDEFS_FALSE ) {
         // IsRestart indicates VarID
         i = IsRestart_m + 1;
         LoadState( BBCnt, T, InputPnts, NumVars, Ji, X, PruneDegs, PruneMons,
           ProbName_i, IsRestart_m ); 
      }

      while ( Ell.size() > 0 ) Ell.pop_back();
      ji = 0;
      for( ii=0; ii<T; ii++ ) {
         if ( i == 0 ) {  
            Ell.push_back( DegPoly_i );
            ji = DegPoly_i + 1;
         }
         else {
            PruneDeg = CompPruneDeg( Ji[ ii ] );
            Ell.push_back( BBDEFS_MIN( DegVector_i[ i - 1 ], PruneDeg ) );
            ji = ji + BBDEFS_MIN( DegVector_i[ i - 1 ], PruneDeg ) + 1 ;            
         }   
      }

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace2( VarVector[i], T, Ell, ji, Ji );
#endif
     
      Prune( VarVector, Ell, Ji, X, PruneDegs, PruneMons );

      T = Ji.size();
      if ( T == 0 ) T = 1;
      while ( Ell.size() > 0 ) Ell.pop_back();
      ji = 0;
      for( ii=0; ii<T; ii++ ) {
         if ( i == 0 ) {  
            Ell.push_back( DegPoly_i );
            ji = DegPoly_i + 1;
         }
         else {
            PruneDeg = CompPruneDeg( Ji[ ii ] );
            Ell.push_back( BBDEFS_MIN( DegVector_i[ i - 1 ], PruneDeg ) );
            ji = ji + BBDEFS_MIN( DegVector_i[ i - 1 ], PruneDeg ) + 1 ;        
         }   
      }
 
      // select random points p0,...,pi
      for( ii=0; ii<=i; ii++ )
         P[ii] = (*RndIt)();

      // interpolation complete 
      if ( ji == 1 ) {
         IsDoneEarly = BBDEFS_TRUE;
         break;
      }

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace3( Ell, PruneMons, PruneDegs, Ji, X, P, i );
#endif

      while ( gki.size() > 0 ) gki.pop_back();
#ifdef BBDEFS_MPI_ON
      // compute gki in parallel
      k = 0; kk = 0;
      while ( k < ji ) {
         RemainingEvals = ji - k;
         // RemainingEvals <= NumProc_i - 1 (we don't dist to driver)
         if ( RemainingEvals < NumProc_i ) {
            NumProc = RemainingEvals;
         }
         else {
            NumProc = NumProc_i-1;
         }
         // distribute evaluations
         for( j=1; j<=NumProc; j++ ) {
            for( ii=0; ii<=i; ii++ ) 
               InputPnts[ii] = P[ii].Exp( k ); 
            PruneVals[k] = HomoEvalPoly( InputPnts, NumVars, PruneDegs, 
                                         PruneMons ); 
            // sets IsDist to BBDEFS_TRUE after first round
            HomoDistBB( j, NumVars, bb_i, InputPnts, IsDist[j], Status );
            k++; 
         }
         // recieve values
         for ( j = 1; j <=NumProc; j++ ) {
             BBVals[kk] = bb_i.Wait( j, Status );
             kk++;
             BBCnt++;
#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace7( BBCnt );
#endif
         }
      }
      for( k=0; k<ji; k++ ) {
         gki.push_back( BBVals[k] - PruneVals[k] );
      } 
#elif !defined(BBDEFS_MPI_ON)
      // compute gki in serial
      for( k=0; k<ji; k++ ) {
         for( ii=0; ii<=i; ii++ ) { 
            InputPnts[ii] = P[ii].Exp( k ); 
         }
         PruneVal  = HomoEvalPoly( InputPnts, NumVars, PruneDegs, PruneMons ); 
         BBVal     = HomoEvalBB ( bb_i, InputPnts, NumVars );
         gki.push_back( BBVal - PruneVal );
         BBCnt++; 

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace7( BBCnt );
#endif

      }
#endif

      k = 0;
      while ( nJi.size() > 0 ) nJi.pop_back();
      while ( V.size() > 0 ) V.pop_back();
      for ( e=0; e<T; e++ ) {
         for( d=0; d<=Ell[e]; d++ ) {
            if ( i == 0 )
               V.push_back( P[0].Exp( d ) );
            else
               V.push_back( ProdRaisedPnts( P, Ji[e], d, i+1 ) ); 
            k++;
            if ( i == 0 ) 
               nJi.push_back( BBSPARSEAUX_DegVector( 1, d ) );
            else {
               BBSPARSEAUX_DegVector tmp;
               tmp.insert( tmp.begin(), Ji[e].begin(), Ji[e].end() );
               tmp.push_back( d );
               nJi.push_back( tmp );
            }
         }
      } 

     VandermondeSolveKL( V, gki, X, P_i );
     //VandermondeSolveZippel( V, gki, X, P_i );

     if ( ProbName_i != NULL ) {
        GetCPUTime( ProbName_i );
     }

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace4( V, gki, X, Ji, nJi );
#endif

     RefineJi( X, X.size(), nJi, Ji );
     T = Ji.size();

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace5( Ji );
#endif

     if ( ProbName_i != NULL ) {
        SaveState( BBCnt, T, InputPnts, NumVars, Ji, X, PruneDegs, PruneMons,
           ProbName_i, i ); 
     }

   }

#ifdef BBDEFS_MPI_ON
   // kill subtasks
   for ( j=1; j < NumProc_i; j++ ) {
       bb_i.Kill( j, Status );
   }
#endif
   
   if ( IsDoneEarly == BBDEFS_FALSE ) {
      while ( Ell.size() > 0 ) Ell.pop_back();
      ji = 0;
      for( ii=0; ii<T; ii++ ) {
         if ( i == 0 ) {  
            Ell.push_back( DegPoly_i );
            ji = DegPoly_i + 1;
         }
         else {
            PruneDeg = CompPruneDeg( Ji[ ii ] );
            Ell.push_back( BBDEFS_MIN( DegVector_i[ i - 1 ], PruneDeg ) );
            ji = ji + BBDEFS_MIN( DegVector_i[ i - 1 ], PruneDeg ) + 1 ;        
         }   
      }

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace2( VarVector[i-1], T, Ell, ji, Ji );
#endif
     
         Prune( VarVector, Ell, Ji, X, PruneDegs, PruneMons );

   }

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace6( X, Ji, PruneMons, PruneDegs, BBCnt );
#endif

}

/*-------------------------------------------------------------
  function: SparseInterp
  date: 12/24/96

  reference: the varibale names mirror those utilized in  the sparse 
     conversion algorithm detailed in "Factorization of polynomials 
     given by straight-line programs", E. Kaltofen, Randomness and 
     Computation Vol 5, Advances in Computing Research, S. Micali ed., 
     Greenwhich, Connecticut, 1989, pg 387--394

  description: This is E. Kaltofen's sparse conversion algorithm. 
      It is intented for comparison purposes ONLY.

-------------------------------------------------------------*/

template< class K, class P, class B > void SparseInterpOld(
   B&                     bb_i,         // black box polynomial 
   BBSPARSEAUX_DegVector& VarVector_i,  // vector of interpolation var. nums. 
   BBSPARSEAUX_DegVector& DegVector_i,  // vector of exact degress of each var
   unsigned int           DegPoly_i,    // total deg of the poly rep. by the 
                                        // black box 
   MP_INT*                Card_i,       // card of the set from which to pick 
                                        // random elements
   BBSPARSEAUX_VecDegVector& PruneDegs, // exponent vector of pruned polynomial
   bbvector< K >&            PruneMons, // monomial vector of pruned polynomial
   P*                     P_i,
   char*                  ProbName_i,
   int&                   IsRestart_m,
   int                    NumProc_i = 1 //
)
{
   unsigned int ii,e,d;                // loop variables 
   unsigned int i;                     // loop over each variable 
   unsigned int k;                     // loop over augmented Ji vector
   unsigned int d0;                    // bound on all of the degs of each var 
   unsigned int ji;                    // size of linear system (ji x ji) 
   unsigned int T;                     // number of terms 
   int          IsDoneEarly;
   BBSPARSEAUX_DegVector VarVector;    // augmented vector of variables
   unsigned int NumVars; 
   K::RandomIterator *RndIt;           // random field elment generator 
   K A[BBSPARSEAUX_MAXVARS];           // initial random elements 
   K P[BBSPARSEAUX_MAXVARS];           // random evaluation points p0,...,pi
   K InputPnts[BBSPARSEAUX_MAXVARS];   // (p0)^k,...,(pi)^k,ai+1,...,an 
   K BBVal;                            // black box evaluated at
                                       // (p0)^k,...,(pi)^k,ai+1,...,an
   unsigned int BBCnt;                 // number of black box calls
   bbvector< K > gki;                  // adjusted black box evaluation
   bbvector< K > X;                    // solution vector 
   bbvector< K > V;                    // describes a trans. vandermonde matrix
   BBSPARSEAUX_VecDegVector Ji;        // vector of exponent vectors 
   BBSPARSEAUX_VecDegVector nJi;       // full vector of exponent vectors 
   bbvector< unsigned int > Ell;       // vector of of "remaining degree" 
                                       // for each monomial 
   int kk;
   T           = 1;
   BBCnt       = 0;
   IsDoneEarly = BBDEFS_FALSE;
   d0          = MaxDegree( DegVector_i );
   VarVector.insert( VarVector.begin(), VarVector_i.begin(), 
                     VarVector_i.end() );
   NumVars = VarVector.size();
   RndIt   = new K::RandomIterator( 100, Card_i );

   for(i=1; i<NumVars; i++ ) {
      A[i]         = (*RndIt)(); 
      InputPnts[i] = A[i]; 
   }
   
#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace1( VarVector, NumVars, DegVector_i, d0 , A, InputPnts );
#endif
   if ( T > BBVECTOR_MAXSIZE ) {
      OutputErrorExit( BBERROR_SPARSE, 
                       "number of terms (T) > BBVECTOR_MAXSIZE" );
   }

   // interpolate one more variable
   for( i=0; i<NumVars; i++ ) {

      while ( Ell.size() > 0 ) Ell.pop_back();
      ji = 0;
      for( ii=0; ii<T; ii++ ) {
         if ( i == 0 ) {  
            Ell.push_back( d0 );
            ji = d0 + 1;
         }
         else {
            int tmpd = 0;
            for( kk=0; kk<Ji[ ii ].size(); kk++ )
               tmpd = tmpd + (Ji[ ii ])[kk];
            Ell.push_back( BBDEFS_MIN( d0, DegPoly_i - tmpd ) );
            ji = ji + BBDEFS_MIN( d0, DegPoly_i - tmpd ) + 1 ;        
         }   
      }

      // select random points p0,...,pi
      for( ii=0; ii<=i; ii++ )
         P[ii] = (*RndIt)();

      cout << "Size = " << ji << endl;
      while ( gki.size() > 0 ) gki.pop_back();
      // compute gki in serial
      for( k=0; k<ji; k++ ) {
         for( ii=0; ii<=i; ii++ ) { 
            InputPnts[ii] = P[ii].Exp( k ); 
         }
         BBVal     = (*bb_i( InputPnts ));
         gki.push_back( BBVal );
         BBCnt++; 

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace7( BBCnt );
#endif

      }

      k = 0;
      while ( nJi.size() > 0 ) nJi.pop_back();
      while ( V.size() > 0 ) V.pop_back();
      for ( e=0; e<T; e++ ) {
         for( d=0; d<=Ell[e]; d++ ) {
            if ( i == 0 )
               V.push_back( P[0].Exp( d ) );
            else
               V.push_back( ProdRaisedPnts( P, Ji[e], d, i+1 ) ); 
            k++;
            if ( i == 0 ) 
               nJi.push_back( BBSPARSEAUX_DegVector( 1, d ) );
            else {
               BBSPARSEAUX_DegVector tmp;
               tmp.insert( tmp.begin(), Ji[e].begin(), Ji[e].end() );
               tmp.push_back( d );
               nJi.push_back( tmp );
            }
         }
      } 

     VandermondeSolveKL( V, gki, X, P_i );
     //VandermondeSolveZippel( V, gki, X, P_i );

     if ( ProbName_i != NULL ) {
        GetCPUTime( ProbName_i );
     }

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace4( V, gki, X, Ji, nJi );
#endif

     RefineJi( X, X.size(), nJi, Ji );
     T = Ji.size();

#ifdef BBDEFS_BBSPARSE_TRC
   BBSPARSETrace5( Ji );
#endif

   }

}
#endif

