
/*-------------------------------------------------------------
  module: bbfactoraux.h
  author: Angel L. Diaz

  date: 10/24/96

  description: 
     The functions and classes in this module provide auxiliary 
     functionality utilized by the black box factorization algorithm. 

  class:
     MapFtoF1
     MapFtoFBar
     MapFtoF1Ex
     MapFtoFBarEx

  functions:
     AddBiPoly
     TruncMultBiPoly
     AddMonomial
     TrucMultMonomial
     NextProdOfFactors
     FindParFracCoeffs
     Lift
    
  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBFACTORAUX
#define _BBFACTORAUX

#define BBFACTOR_MAXDEG     100
#define BBFACTOR_MAXVAR     100
#define BBFACTOR_MAXFACTORS 100

#include "BlackBox/CommonObjects/blackbox.h"
#include "BlackBox/Aux/bbdense.h"
#include "BlackBox/Algorithms/bbfactortrace.h"

/*-------------------------------------------------------------
  function: AddBiPoly
  date: 8/9/96

  description: 
     Sum the input bivariate polynomials

  return:  
     BiPolyA_i + BiPolyB_i

  assumptions: 
     It is assumed that the degree of each input bivariate 
     polynomial has been correctly set. 

-------------------------------------------------------------*/

template< class P > 
inline BiPoly<P> AddBiPoly( 
    BiPoly<P>& BiPolyA_i,     
    BiPoly<P>& BiPolyB_i 
    ) 
{
   unsigned int i;
   unsigned int TermsA, TermsB, TermsMax, TermsMin;
   BiPoly<P>    Sum;
   
   TermsA   = BiPolyA_i.Deg() + 1;
   TermsB   = BiPolyB_i.Deg() + 1;
   TermsMax = BBDEFS_MAX( TermsA, TermsB ); 
   TermsMin = BBDEFS_MIN( TermsA, TermsB );
   
   Sum.SetDeg( TermsMax - 1 );

   // sum up to common number of terms
   for(i=0;i<=TermsMin;i++)
     Sum[i] = BiPolyA_i[i] + BiPolyB_i[i];

   // determine which input polynomial has the maximum number of terms
   // and insert the remaining coefficients
   if ( TermsA == TermsMax )
      for(i=TermsMin+1;i<=TermsMax;i++)
          Sum[i] = BiPolyA_i[i];
   else
      for(i=TermsMin+1;i<=TermsMax;i++)
          Sum[i] = BiPolyB_i[i];

   return Sum;
}

/*-------------------------------------------------------------
  function: TruncMultBiPoly
  date: 8/9/96

  description: 
     Compute the product of the input bivariate polynomials 
     tuncated at Y^d

  return: 
     BiPolyA_i * BiPolyB_i mod Y^d

  assumptions: 
     It is assumed that the degree of each input bivariate
     polynomial has been correctly set. 

-------------------------------------------------------------*/

template< class P > 
inline BiPoly<P> TruncMultBiPoly( 
   BiPoly<P>&   BiPolyA_i, 
   BiPoly<P>&   BiPolyB_i, 
   unsigned int TruncDegree_i   // degree to truncate product
   ) 
{
   MP_INT       MPZero;
   unsigned int i, k;
   unsigned int TermsA, TermsB, TermsProd; 
   BiPoly<P>    Prod;
   P            PartialProd;
   mpz_init_set_ui( &MPZero, 0 );

   TermsA = BiPolyA_i.Deg() + 1;
   TermsB = BiPolyB_i.Deg() + 1;

   // determine the degree of the resulting product
   TermsProd = BBDEFS_MIN( TruncDegree_i, TermsA + TermsB - 1 ); 
   Prod.SetDeg(TermsProd - 1);

   // perform multiplication up to the truncation degree
   for(k=0;k<=TermsProd;k++) {
      PartialProd = P( (MP_INT*)&MPZero, 0 );
      for(i=0;i<=k;i++) {
         if ( ( i <= TermsA ) && ( ( k - i ) <= TermsB ) )
            PartialProd = PartialProd + BiPolyA_i[i] * BiPolyB_i[k-i];  
      }
      Prod[k] = PartialProd; 
   }

   mpz_clear( &MPZero );
   return Prod; 
}

/*-------------------------------------------------------------
  function: AddMonomial
  date: 8/9/96

  description: 
     The sum of the input bivariate polynomial and a monomial 
     at a specified Y degree for the univairate polynomial coefficient.

  return: 
     BiPoly_i + (Coeff_i)Y^d

  assumptions: 
     It is assumed that the degree of each input bivariate polynomial 
     has been correctly set. 

-------------------------------------------------------------*/

template< class P > 
inline BiPoly<P> AddMonomial( 
   BiPoly<P>&   BiPoly_i, 
   P&           Coeff_i, 
   unsigned int YDegree_i 
   ) 
{
   unsigned int i;
   unsigned int TermsBiPoly, TermsUniPoly;
   BiPoly<P>    Prod;
   MP_INT       Zero;
   mpz_init_set_si( &Zero, 0 );

   TermsBiPoly  = BiPoly_i.Deg() + 1;
   TermsUniPoly = YDegree_i + 1;

   // populate the result with the input bivariate polynomial
   for(i=0;i<=TermsBiPoly;i++) 
      Prod[i] = BiPoly_i[i];
    
   // if the degree of the monomial is below the bivariate polynomial
   if ( TermsUniPoly <= TermsBiPoly ) {
      Prod[YDegree_i] = BiPoly_i[YDegree_i] + Coeff_i;
      Prod.SetDeg( BiPoly_i.Deg() ); 
   }
   else { 
      // pad with zeros
      for(i=TermsBiPoly+1;i<TermsUniPoly;i++)
         Prod[i] = P( &Zero ); 
      // insert the coeffecient
      Prod[YDegree_i] = Coeff_i;
      Prod.SetDeg( YDegree_i );
   }
      
   mpz_clear( &Zero );
   return Prod;
}

/*-------------------------------------------------------------
  function: TrucMultMonomial
  date: 8/9/96

  description: 
     Compute the product of the input bivariate 
     polynomial and a monomial at a specified Y degree for the 
     univairate polynomial coefficient. This product is performed 
     modulo a truncation degree.

  return: 
     BiPoly_i * (Coeff_i)Y^d mod Y^n

  assumptions: 
     It is assumed that the degree of each input bivariate polynomial 
     has been correctly set. 

-------------------------------------------------------------*/

template< class P> 
inline BiPoly<P> TrucMultMonomial( 
   BiPoly<P>&   BiPoly_i, 
   P&           Coeff_i, 
   unsigned int YDegree_i, 
   unsigned int TruncDegree_i 
   ) 
{
   unsigned int i;
   unsigned int TermsBiPoly, TermsUniPoly;
   unsigned int RemainingTerms;
   BiPoly<P>    Prod;
   MP_INT       Zero;
   mpz_init_set_si( &Zero, 0 );
      
   TermsBiPoly  = BiPoly_i.Deg() + 1;
   TermsUniPoly = YDegree_i + 1;

   // if the degree of the monomial is larger than the truncation degree
   // return zero 
   if ( YDegree_i >= TruncDegree_i ) {
      for(i=0;i<TruncDegree_i-1;i++)
         Prod[i] = P( &Zero );
      Prod.SetDeg( TruncDegree_i - 1 );
   }
   else {

      // pad with zeros
      for(i=0;i<YDegree_i;i++)
         Prod[i] = P( &Zero );

      // multiply the remaining terms
      RemainingTerms = BBDEFS_MIN( TruncDegree_i - YDegree_i  , TermsBiPoly );
      for(i=0;i<RemainingTerms;i++)
         Prod[YDegree_i+i] = BiPoly_i[i]*Coeff_i;

      Prod.SetDeg( YDegree_i + RemainingTerms - 1 );
   }

   mpz_clear( &Zero );
   return Prod;
}

/*-------------------------------------------------------------
  function: NextProdOfFactors
  date: 8/9/96

  description: 
     Compute the incremental product (with multiplicities) of the 
     factors, given a previous incremental product of the factors 
     and the next term of each factor.

  reference: 
    the varibale names mirror those utilized in 
    "Factorization of polynomials given by straight-line programs",
    E. Kaltofen, Randomness and Computation Vol 5, Advances in 
    Computing Research, S. Micali ed., Greenwhich, Connecticut,
    1989, pg 401--403

  assumptions: 
     Also, it is assumed that the bivariate arithmetic constructor 
     initializes to the (0)*Y^0 polynomial.

-------------------------------------------------------------*/

template< class P > 
inline void NextProdOfFactors(
   BiPoly<P>*  PrevProds_i,// array which holds the incremental product (with
                           // multiplicities) of the factors (each factor is 
                           // at the kth step of the Hensle loop) 
   BiPoly<P>* gik_i,       // array of the ith factors at kth degree of the
                           // Hensle loop 
   P* gikp1_i,             // array of monomials corresp. to each of the ith
                           // factors 
   unsigned int YDeg_i,    // degree of each of the monomials gikp1_i 
   unsigned int r_i,       // number of factors 
   unsigned int *Exps_i,   // exponents of each factor ( e1, ..., er ) 
   unsigned int ExpsSum_i, // sum of the exponents ( e1 + ... + er ) 
   unsigned int TruncDeg_i,
   BiPoly<P>* NextProds_o  // array which holds the new incremental product 
                           // (with multiplicities) of the factors 
   )
{
   unsigned int i, j;
   unsigned int ProdIndex = 0;
   BiPoly<P>    a[BBFACTOR_MAXFACTORS];
   BiPoly<P>    ah[BBFACTOR_MAXFACTORS];
   BiPoly<P>    t1,t2,t3,t4;

   a[0]           = PrevProds_i[0];
   ah[0]          = AddMonomial( ah[0], gikp1_i[0], YDeg_i );
   NextProds_o[0] = AddBiPoly( a[0], ah[0] );
 
   ProdIndex = 1;
   for(i=1;i<Exps_i[0];i++){
      a[i]  = PrevProds_i[i];
      t1    = TruncMultBiPoly( ah[i-1], gik_i[0], TruncDeg_i ); 
      t2    = TrucMultMonomial( a[i-1], gikp1_i[0], YDeg_i, TruncDeg_i ); 
      t3    = TrucMultMonomial( ah[i-1], gikp1_i[0], YDeg_i, TruncDeg_i );
      t4    = AddBiPoly( t1, t2 );
      ah[i] = AddBiPoly( t4, t3 );
      NextProds_o[i] = AddBiPoly( a[i], ah[i] );
      ProdIndex      = ProdIndex + 1; 
   }
   
   for(j=1;j<r_i;j++) {
      for(i=0;i<Exps_i[j];i++) {
         a[ProdIndex]  = PrevProds_i[ProdIndex];
         t1 = TruncMultBiPoly( ah[ProdIndex-1], gik_i[j], TruncDeg_i ); 
         t2 = TrucMultMonomial( a[ProdIndex-1], gikp1_i[j], YDeg_i, 
              TruncDeg_i ); 
         t3 = TrucMultMonomial( ah[ProdIndex-1], gikp1_i[j], YDeg_i, 
              TruncDeg_i );
         t4 = AddBiPoly( t1, t2 );
         ah[ProdIndex]          = AddBiPoly( t4, t3 );
         NextProds_o[ProdIndex] = AddBiPoly( a[ProdIndex], ah[ProdIndex] );
         ProdIndex              = ProdIndex + 1; 
      }
   }
}

/*-------------------------------------------------------------
  function: FindParFracCoeffs
  date: 8/20/96

  description: 

  reference: 
    the varibale names mirror those utilized in 
    "Factorization of polynomials given by straight-line programs",
    E. Kaltofen, Randomness and Computation Vol 5, Advances in 
    Computing Research, S. Micali ed., Greenwhich, Connecticut,
    1989, pg 401--403

-------------------------------------------------------------*/

template< class P > 
inline void FindParFracCoeffs(
   P* UFactors_i, 
   unsigned int* Exps_i,
   unsigned int NumFactors_i,
   P ParFrac_o[BBFACTOR_MAXDEG][BBFACTOR_MAXFACTORS],
   unsigned int& r_o,
   unsigned int& c_o
   )
{
   unsigned int i,m;
   unsigned int d;
   unsigned int r;
   P h0s[BBFACTOR_MAXFACTORS];  // 1 ... r
   P si0[BBFACTOR_MAXFACTORS];  // 0 ... r-1
   P ti0[BBFACTOR_MAXFACTORS];  // 0 ... r-1
   P gi0,hi0,g,gcd;
   P Numeri;
   P hat_gi0m,hat_hi0m;
   P FProd;              
   MP_INT MPOne;
   MP_INT MPExp;
   mpz_init( &MPExp );
   mpz_init_set_si( &MPOne, 1 );

   r = NumFactors_i - 1;
   FProd           = UFactors_i[ r ];
   h0s[r]          = UFactors_i[ r ];

   for( i=r-1; i>0; i-- ) {
      FProd = FProd * UFactors_i[i];
      h0s[i] = FProd;
   }

   d = (UFactors_i[0]).dg() + (h0s[1]).dg();

   for( i=0; i<=r-1; i++ ) {
      gi0 = UFactors_i[i];
      hi0 = h0s[i+1];
      g.peea( gi0, hi0, gcd, si0[i], ti0[i] );
      // this turns out to be division by a field elment
      // MUST ADD CHECK FOR CONSTANT GCD, ELSE FACTORS ARE NOT REL PRIME
      si0[i] = si0[i]/gcd;
      ti0[i] = ti0[i]/gcd;
   }

   for( m=0; m<=d-1; m++ ) {
      Numeri = P( &MPOne , m );
      for( i=0; i<=r-1; i++ ) {
         gi0 = UFactors_i[i];
         hi0 = h0s[i+1];
         g.prem( ti0[i] * Numeri, gi0, hat_gi0m );
         g.prem( si0[i] * Numeri, hi0, hat_hi0m );
         mpz_set_ui( &MPExp, Exps_i[i] );
         ParFrac_o[m][i] = hat_gi0m / P( &MPExp );
         if ( i == r-1 ) {
            mpz_set_ui( &MPExp, Exps_i[r] );
            ParFrac_o[m][r] = hat_hi0m / P( &MPExp );
         }
         else
           Numeri = hat_hi0m;
      }
   }

   c_o = d-1;
   r_o = r;
 
   mpz_clear( &MPOne );
   mpz_clear( &MPExp ); 
}

/*-------------------------------------------------------------
  function: Lift
  date: 8/21/96

  description: 

  reference: 
    the varibale names mirror those utilized in 
    "Factorization of polynomials given by straight-line programs",
    E. Kaltofen, Randomness and Computation Vol 5, Advances in 
    Computing Research, S. Micali ed., Greenwhich, Connecticut,
    1989, pg 401--403

-------------------------------------------------------------*/

template< class P > 
inline void Lift(
   BiPoly< P >&  fbar_i,   // poly in X and Y, lc[Y] in K
   P*            g0_i,     // array of true factor images of fbar(X,0)
   unsigned int  NumFactors_i,
   P*            g0Prod_i, // array of products of factors (with multiplicities)
   unsigned int  SumExps_i,
   unsigned int* Exps_i,   // array of corresponding exponents
   P ParFrac_i[BBFACTOR_MAXDEG][BBFACTOR_MAXFACTORS], // diophantine solutions
   P pgsem1_i,             // Prod of factors (with multiplicities - 1)
   BiPoly< P >* BiFactors_o
   )
{
   unsigned int i,k,j,m;
   unsigned int MaxDeg = 0;
   unsigned int SumFactorDegs = 0;
   unsigned int FactorDegs[BBFACTOR_MAXFACTORS]; 
   BiPoly< P > gp[BBFACTOR_MAXDEG];    // array of inc. prod. of the factors
   BiPoly< P > ngp[BBFACTOR_MAXDEG];
   BiPoly< P > g[BBFACTOR_MAXFACTORS]; // array of lifted factors
   BiPoly< P > gikm1[BBFACTOR_MAXDEG]; // array of previous lifted factors
   P gik[BBFACTOR_MAXDEG];             // array of next coeff. for each factor
   P TmpCoeff;
   P ttk,uk,Rem;
   MP_INT MPZero;
   mpz_init_set_si( &MPZero, 0 );

   for( i=0; i<NumFactors_i; i++ ) {
      FactorDegs[i] = g0_i[i].dg();
      SumFactorDegs = SumFactorDegs + FactorDegs[i];
   }

   MaxDeg = fbar_i.Deg();

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
   BBFACLIFTTrace1(FactorDegs, NumFactors_i, MaxDeg, SumFactorDegs);
#endif

   MaxDeg = fbar_i.Deg();

   for( i=0; i<SumExps_i; i++ ) {
      (gp[i])[0] = g0Prod_i[i]; 
      (gp[i]).SetDeg( 0 );
   }

   for( i=0; i<NumFactors_i; i++ ) {
      (g[i])[0] = g0_i[i];
      (g[i]).SetDeg( 0 );
   }

   // Lift g0[i] from Y^(k-1) to Y^k
   for( k=0; k<MaxDeg; k++ ) {

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
   BBFACLIFTTrace2( k );
#endif

      if ( k == 0 ) 
         TmpCoeff = GetCoeff( gp[SumExps_i-1], k );
      else {

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
   BBFACLIFTTrace3( gp, SumExps_i);
#endif

         NextProdOfFactors( gp, gikm1, gik, k, NumFactors_i, Exps_i, 
            SumExps_i, MaxDeg+1, ngp );

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
   BBFACLIFTTrace4( ngp, SumExps_i);
#endif

         TmpCoeff = GetCoeff( ngp[SumExps_i-1], k );
         for( j=0; j<BBFACTOR_MAXDEG; j++ )
            gp[j] = ngp[j];
      }
      ttk = GetCoeff( fbar_i, k ) - TmpCoeff; 
      uk.pquorem(ttk, pgsem1_i, uk, Rem);
      if ( Rem != P( (MP_INT*)&MPZero ) ) {
         cout << "ttk = " << ttk << endl; 
         cout << "pgsem1_i = " << pgsem1_i << endl;
         cout << "uk = " << uk << endl;
         cout << "Rem = " << Rem << endl;
         OutputErrorExit( BBERROR_FACTORS, "Untrue univariate factorization" );
      }

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
   BBFACLIFTTrace5(TmpCoeff,ttk,uk);
#endif

      // for each factor
      for( i=0; i<NumFactors_i; i++ ) {
         gik[i] = P( &MPZero );
         for( m=0; m<SumFactorDegs; m++ ) {

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
    BBFACLIFTTrace6(m,i,ParFrac_i[m][i]);
#endif
            if ( ( m > uk.dg() ) || ( uk == P( &MPZero ) ) )
               TmpCoeff = P( &MPZero );
            else
               TmpCoeff = uk[m];
            gik[i] = gik[i] + ParFrac_i[m][i] * TmpCoeff;
          
         }
         gikm1[i] = g[i]; 
         g[i] = AddMonomial( gikm1[i], gik[i], k+1 );

#ifdef BBDEFS_BBFACTOR_LIFT_TRC
   BBFACLIFTTrace7(gik[i],gikm1[i],g[i],i);
#endif

      }
   }

   for( i=0; i<NumFactors_i; i++ )
       BiFactors_o[i] = g[i];

   mpz_clear( &MPZero );

}

template< class P > inline P GetCoeff(
   BiPoly< P >& BiPoly_i,
   unsigned int index_i
   )
{
   P Coeff;
   // this assumes that every coeff of the BiPoly is initialized to zero
   Coeff = BiPoly_i[index_i+1];

   return( Coeff );
}

/*-------------------------------------------------------------
  class: MapFtoF1
  date: 8/23/96

  reference:

  description:

  assumptions: 
     The input black box polynomial is in at least two varibales.

-------------------------------------------------------------*/

template < class K, class P, class B > 
class MapFtoF1 : public BlackBox< K > {
   private:
      B* BB;             // black box polynomial
      K* RandA;          // elements for multivariate polynomial projection
      K* RandB;          // elements for multivariate polynomial projection
      K Pnts[BBFACTOR_MAXVAR]; // array of projected points
   public:
      MapFtoF1( B*, K*, K*, unsigned int );
      ~MapFtoF1() {}
      inline K operator () ( K* );
};

template < class K, class P, class B > 
MapFtoF1< K, P, B>::MapFtoF1(
   B* BB_i,
   K* RandA_i,
   K* RandB_i,
   unsigned int NumPnts_i
   )
{
   BB            = BB_i;
   RandA         = RandA_i;
   RandB         = RandB_i;
   NumVariables  = NumPnts_i;
}

template < class K, class P, class B > 
inline K MapFtoF1< K, P,B >::operator () (
   K* Pnts_i
   )
{
   unsigned int i;
   K X;
   X = Pnts_i[0];

   Pnts[0] = X + RandB[0];           // X    + b1
   Pnts[1] = RandA[0]*X + RandB[1];  // a1*X + b2
   for( i=2; i<NumVariables; i++ )
     Pnts[i] = RandA[i-1]*X + RandB[i];  // an*X + bn

   return ( (*BB)( Pnts ) );
}

/*-------------------------------------------------------------
  class: MapFtoFBar
  date: 8/22/96

  description:

  assumptions: 
     The input black box polynomial is in at least two varibales.
-------------------------------------------------------------*/

template < class K, class P, class B > 
class MapFtoFBar : public BlackBox< K > {
   private:
      B* BB;             // black box polynomial
      K* RandA;          // elements for multivariate polynomial projection
      K* RandB;          // elements for multivariate polynomial projection
      K ConstPnts[BBFACTOR_MAXVAR]; // pn - an*(p1 - b1) - bn
      K Pnts[BBFACTOR_MAXVAR];      // array of projected points
   public:
      MapFtoFBar( B*, K*, K*, K*, unsigned int );
      ~MapFtoFBar() {}
      inline K operator () ( K* );
};

template < class K, class P, class B > 
MapFtoFBar< K, P,B >::MapFtoFBar(
   B* BB_i,
   K* RandA_i,
   K* RandB_i,
   K* Pnts_i,
   unsigned int NumPnts_i
   )
{
   unsigned int i;

   BB            = BB_i;
   RandA         = RandA_i;
   RandB         = RandB_i;
   NumVariables  = NumPnts_i;

   // pn - an*(p1 - b1) - bn
   for( i=0; i<NumVariables-1; i++ )
      ConstPnts[i] = Pnts_i[i+1] - RandA[i]*(Pnts_i[0]-RandB[0]) - RandB[i+1];
}

template < class K, class P, class B > 
inline K MapFtoFBar< K, P, B >::operator () (
   K* Pnts_i
   )
{
   unsigned int i;
   K X;
   K Y;
   X = Pnts_i[0];
   Y = Pnts_i[1];

   Pnts[0] = X + RandB[0];           // X + b1
   // Y*(pn - an*(p1 - b1) - bn) + an*X + bn
   for( i=1; i<NumVariables; i++ )
      Pnts[i] = Y*ConstPnts[i-1] + RandA[i-1]*X + RandB[i];

   return ( (*BB)( Pnts ) );
}

/*-------------------------------------------------------------
  class: MapFtoF1Ex
  date: 10/24/96

  description: 

  assumptions: 
     The input black box polynomial is in at least two varibales.
-------------------------------------------------------------*/

template < class K, class P, class L, class B > 
class MapFtoF1Ex : public BlackBoxEx< K, L > {
   private:
      B* BB;             // black box polynomial
      K* RandA;          // elements for multivariate polynomial projection
      K* RandB;          // elements for multivariate polynomial projection
      K Pnts[BBFACTOR_MAXVAR]; // array of projected points
   public:
      MapFtoF1Ex( B*, K*, K*, unsigned int );
      ~MapFtoF1Ex() {}
      inline K operator () ( K* );
};

template < class K, class P, class L, class B > 
MapFtoF1Ex< K, P, L, B >::MapFtoF1Ex( 
   B* BB_i, 
   K* RandA_i, 
   K* RandB_i, 
   unsigned int NumPnts_i
   )
{
   BB            = BB_i;
   RandA         = RandA_i;
   RandB         = RandB_i;
   NumVariables  = NumPnts_i;
}

template < class K, class P, class L, class B > 
inline K MapFtoF1Ex< K, P, L, B >::operator () ( 
   K* Pnts_i
   )
{ 
   unsigned int i;
   K X;
   X = Pnts_i[0];

   Pnts[0] = X + RandB[0];           // X    + b1
   Pnts[1] = RandA[0]*X + RandB[1];  // a1*X + b2
   for( i=2; i<NumVariables; i++ )
     Pnts[i] = RandA[i-1]*X + RandB[i];  // an*X + bn

   return ( (*BB)( Pnts ) );
}

/*-------------------------------------------------------------
  class: MapFtoFBarEx
  date: 10/24/96

  description: 

  assumptions: 
     The input black box polynomial is in at least two varibales.

-------------------------------------------------------------*/

template < class K, class P, class L, class B > 
class MapFtoFBarEx : public BlackBoxEx< K, L > {
   private:
      B* BB;             // black box polynomial
      K* RandA;          // elements for multivariate polynomial projection
      K* RandB;          // elements for multivariate polynomial projection
      K ConstPnts[BBFACTOR_MAXVAR]; // pn - an*(p1 - b1) - bn
      K Pnts[BBFACTOR_MAXVAR];      // array of projected points
   public:
      MapFtoFBarEx( B*, K*, K*, K*, unsigned int );
      ~MapFtoFBarEx() {}
      inline K operator () ( K* );
};

template < class K, class P, class L, class B > 
MapFtoFBarEx< K, P, L, B >::MapFtoFBarEx( 
   B* BB_i, 
   K* RandA_i, 
   K* RandB_i, 
   K* Pnts_i, 
   unsigned int NumPnts_i
   )
{
   unsigned int i;

   BB            = BB_i;
   RandA         = RandA_i;
   RandB         = RandB_i;
   NumVariables  = NumPnts_i;

   // pn - an*(p1 - b1) - bn
   for( i=0; i<NumVariables-1; i++ ) 
      ConstPnts[i] = Pnts_i[i+1] - RandA[i]*(Pnts_i[0]-RandB[0]) - RandB[i+1];
}

template < class K, class P, class L, class B > 
inline K MapFtoFBarEx< K, P, L, B >::operator () ( 
   K* Pnts_i
   )
{ 
   unsigned int i;
   K X;
   K Y;
   X = Pnts_i[0];
   Y = Pnts_i[1];

   Pnts[0] = X + RandB[0];           // X + b1
   // Y*(pn - an*(p1 - b1) - bn) + an*X + bn
   for( i=1; i<NumVariables; i++ )
      Pnts[i] = Y*ConstPnts[i-1] + RandA[i-1]*X + RandB[i];  

   return ( (*BB)( Pnts ) );
}

#endif
