
/*-------------------------------------------------------------
  module: bbcauchy.h
  author: Angel L. Diaz

  date: 10/22/96

  description: 
     The classes in this module realizes the black box representation 
     of the determinant of a cauchy matrix common object.
     In the FoxBox library, a cauchy determinant black box object 
     is a function object which when supplied with values (a pointer to an 
     array of field elements) for the symbols, computes the value for the 
     cauchy determinant.

  classes:
     BlackBoxCauchyDet

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBCAUCHYDET
#define _BBCAUCHYDET

#include "BlackBox/CommonObjects/blackbox.h"

#define BBVAND_MAXPNTS 65

/*-------------------------------------------------------------
  class: BlackBoxCauchyDet
  date: 10/22/96

  description:
     Provides for constructing a determinant of a cauchy
     matrix common black box object.

-------------------------------------------------------------*/

template< class K > 
class BlackBoxCauchyDet : public BlackBox < K > {
   private:
      int n; // dimension of matrix
      K Value;
   public:
      BlackBoxCauchyDet( );
      BlackBoxCauchyDet( const BlackBoxCauchyDet< K >& );
      BlackBoxCauchyDet( unsigned int NumVars_i, 
                         unsigned int Deg_i, 
                         double       Prob_i
                    ) : BlackBox< K >() 
      { 
         NumVariables     = NumVars_i; 
         Degree           = Deg_i; 
         Probability      = Prob_i;
         n                = NumVariables/2;
      }
      ~BlackBoxCauchyDet() { }

      inline K* operator () ( K* );
      inline BlackBoxCauchyDet< K >& 
             operator=( const BlackBoxCauchyDet< K >& );
};

template< class K > 
BlackBoxCauchyDet< K >::BlackBoxCauchyDet( ) : BlackBox< K >( ) {
}

template< class K >
BlackBoxCauchyDet< K >::BlackBoxCauchyDet( 
   const BlackBoxCauchyDet< K >& Det_i 
   ) : BlackBox< K >( ) 
{ 
   *this = Det_i; 
}

template< class K > 
inline K* BlackBoxCauchyDet< K >::operator () (
   K *Pnts_i
   ) 
{ 
   K      N,D;
   int    i,j;
   MP_INT MPZero;
   mpz_init_set_ui( &MPZero, 0 );

   // numerator
   N = (long)1;
   for( i=0; i<n ; i++ ) 
      for( j=i+1; j<n; j++ ) 
          N = N * (Pnts_i[j] - Pnts_i[i] ) * 
              ( Pnts_i[j+n] - Pnts_i[i+n] );

   // denominator
   D = (long)1;
   for( i=0; i<n; i++ )
      for( j=0; j<n; j++ ) {
         D = D * ( Pnts_i[i] + Pnts_i[j+n ] );
      }

   if ( D == K( (MP_INT*)&MPZero ) ) 
      OutputErrorExit( BBERROR_CAUCHYDET, "Division by zero" );

   Value = N/D;
   mpz_clear( &MPZero );

   return( &Value );
}

template< class K > 
inline BlackBoxCauchyDet< K >& BlackBoxCauchyDet< K >::operator=(
   const BlackBoxCauchyDet< K >& Det_i
   )
{ 
   NumVariables     = Det_i.NumVariables;
   Degree           = Det_i.Degree; 
   Probability      = Det_i.Probability;

   return *this; 
}

#endif
