
/*-------------------------------------------------------------
  module: bbdet.h
  author: Angel L. Diaz

  date: 10/21/96

  description: 
     The class in this module realizes the black box representation 
     of the determinant of a matrix common object. In the FoxBox library, a
     determinant black box object is a function object which 
     when supplied with values (a pointer to an array of field elements) 
     for the symbols, computes the value for the determinant.

  class:
     BlackBoxDet 
     BlackBoxDetEx

  functions:
     FracFreeElm

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBDET
#define _BBDET

#define BBDET_MAXN 65
#define BBDET_MAXM 65

#include "BlackBox/CommonObjects/bbpoly.h"

/*-------------------------------------------------------------
  class: BlackBoxDet
  date: 10/21/96

  description:
     Provides for constructing a determinant of a matrix common 
     object from a pointer to a function. It is expected 
     that the function takes as input 
        (1) an array of coefficient domain elements (K) 
        (2) a two dimensional array of domain elements 
     and populates the array with the corresponding input elements.

-------------------------------------------------------------*/

template< class K > 
class BlackBoxDet : public BlackBox < K > {
   private:
      K A[BBDET_MAXN][BBDET_MAXM];
      typedef void ( *FPTR )( K*, K [BBDET_MAXN][BBDET_MAXM] );
      FPTR FillMatrixFunc;
      int n,m;
      K Value;
   public:
      BlackBoxDet( );
      BlackBoxDet( const BlackBoxDet< K >& );
      BlackBoxDet( FPTR         FillMatrixFunc_i, 
                   unsigned int NumVars_i, 
                   unsigned int Deg_i, 
                   double       Prob_i,
                   int          n_i,
                   int          m_i
                  ) : BlackBox< K >() 
      { 
         FillMatrixFunc   = FillMatrixFunc_i; 
         NumVariables     = NumVars_i; 
         Degree           = Deg_i; 
         Probability      = Prob_i;
         n                = n_i;
         m                = m_i;
      }

      ~BlackBoxDet() { }

      inline K* operator () ( K* );
      inline BlackBoxDet< K >& operator=( const BlackBoxDet< K >& );
};

template< class K > 
BlackBoxDet< K >::BlackBoxDet( ) : BlackBox< K >( ) {
}

template< class K >
BlackBoxDet< K >::BlackBoxDet( 
   const BlackBoxDet< K >& Det_i 
   ) : BlackBox< K >( ) 
{ 
   *this = Det_i; 
}

template< class K > 
inline K* BlackBoxDet< K >::operator () (
   K *Pnts_i
   ) 
{ 
   ( *FillMatrixFunc )( Pnts_i , A ); 
   Value = FracFreeElm( A, n, m );
   return ( &Value ); 
}

template< class K > 
inline BlackBoxDet< K >& BlackBoxDet< K >::operator=(
   const BlackBoxDet< K >& Det_i
   )
{ 
   FillMatrixFunc   = Det_i.FillMatrixFunc; 
   NumVariables     = Det_i.NumVariables;
   Degree           = Det_i.Degree; 
   Probability      = Det_i.Probability;
   n                = Det_i.n;
   m                = Det_i.m;

   return *this; 
}

/*-------------------------------------------------------------
  class: BlackBoxDetEx
  date: 10/21/96

  description:
     Provides for constructing a determinant of a matrix common 
     object from two pointers to functions. It is expected 
     that two "unknown" functions are provided:

      (1) takes as input an array of coefficient domain elements (K),
          a two dimensional array of domain elements and populates 
          the array with the corresponding input elements. 

      (2) takes as input an array of coefficient extension domain 
          elements (L), a two dimensional array of extended domain 
          elements and populates the array with the corresponding 
          input elements.

-------------------------------------------------------------*/

template< class K, class L > 
class BlackBoxDetEx : public BlackBoxEx < K, L > {
   private:
      K AK[BBDET_MAXN][BBDET_MAXM];
      L AL[BBDET_MAXN][BBDET_MAXM];
      typedef void ( *FPTRK )( K*, K [BBDET_MAXN][BBDET_MAXM] );
      typedef void ( *FPTRL )( L*, L [BBDET_MAXN][BBDET_MAXM] );
      FPTRK FillMatrixFuncK;
      FPTRL FillMatrixFuncL;
      int n,m;
      K KValue;
      L LValue;
   public:
      BlackBoxDetEx( );
      BlackBoxDetEx( const BlackBoxDetEx< K, L >& );
      BlackBoxDetEx( FPTRK         FillMatrixFuncK_i, 
                     FPTRL         FillMatrixFuncL_i,
                     unsigned int NumVars_i, 
                     unsigned int Deg_i, 
                     double       Prob_i,
                     int          n_i,
                     int          m_i
                    ) : BlackBoxEx< K, L >() 
      { 
         FillMatrixFuncK   = FillMatrixFuncK_i; 
         FillMatrixFuncL   = FillMatrixFuncL_i;
         NumVariables      = NumVars_i; 
         Degree            = Deg_i; 
         Probability       = Prob_i;
         n                 = n_i;
         m                 = m_i;
      }

      ~BlackBoxDetEx() { }

      inline K* operator () ( K* );
      inline L* operator () ( L* );
      inline BlackBoxDetEx< K,L >& operator=( const BlackBoxDetEx< K,L >& );
};

template< class K, class L > 
BlackBoxDetEx< K, L >::BlackBoxDetEx( ) : BlackBoxEx< K, L >( ) {
}

template< class K, class L >
BlackBoxDetEx< K, L >::BlackBoxDetEx( 
   const BlackBoxDetEx< K, L >& Det_i 
   ) : BlackBoxEx< K, L >( ) 
{ 
   *this = Det_i; 
}

template< class K, class L > 
inline K* BlackBoxDetEx< K, L >::operator () (
   K *Pnts_i
   ) 
{ 
   ( *FillMatrixFuncK )( Pnts_i , AK ); 
   KValue = FracFreeElm( AK, n, m );
   return ( &KValue ); 
}

template< class K, class L >
inline L* BlackBoxDetEx< K, L >::operator () (
   L *Pnts_i
   ) 
{ 
   ( *FillMatrixFuncL )( Pnts_i , AL );
   LValue = FracFreeElm( AL, n, m );
   return ( &LValue );
}

template< class K, class L > 
inline BlackBoxDetEx< K, L >& BlackBoxDetEx< K, L >::operator=(
   const BlackBoxDetEx< K, L >& Det_i
   )
{ 
   FillMatrixFuncK  = Det_i.FillMatrixFuncK; 
   FillMatrixFuncL  = Det_i.FillMatrixFuncL;
   NumVariables     = Det_i.NumVariables;
   Degree           = Det_i.Degree; 
   Probability      = Det_i.Probability;
   n                = Det_i.n;
   m                = Det_i.m;

   return *this; 
}

/*-------------------------------------------------------------
  function: FracFreeElm
  date: 10/21/96

  reference:
     K. O. Geddes, S. R. Czapor, G. Labahn, "Algorithms for 
     Computer Algebra", Kluwer Academic Publ., pp 406, 1992

  description:
     Fraction Free Gaussian Elimination

-------------------------------------------------------------*/

template <class K > 
K FracFreeElm( 
   K A[BBDET_MAXN][BBDET_MAXM],
   int n,
   int m 
)
{
   MP_INT MPOne, MPZero, MPSign;
   K One,Zero;
   K Divisor;
   K SwapElm;
   K Det;
   int k,i,j,p;
   int Sign, r;

   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   Zero    = K( (MP_INT*)&MPZero );
   One     = K( (MP_INT*)&MPOne );
   Sign    = 1;
   Divisor = One;
   r       = 0;
   
   // eliminate below row r, with pivot in column k
   for( k=0; k<n; k++ ) {
      if ( r < m ) {
         // find a nonzero pivot
         for( p=r; p<m; p++ ) {
            if ( A[p][k] != Zero ) 
               break;
         }
         if ( p < m ) {
            // pivot is in row p, so switch rows p and r
            for( j=k; j<n; j++ ) {
               SwapElm = A[p][j];
               A[p][j] = A[r][j]; 
               A[r][j] = SwapElm;
            } 
            // keep track of Sign changes due to row exchange
            if ( r != p ) 
               Sign = -1 * Sign;
            for( i=r+1; i<m; i++ ) {
               for( j=k+1; j<n; j++) {
                 A[i][j] = (A[r][k]*A[i][j] - A[r][j]*A[i][k])/
                             Divisor;
               }
               A[i][k] = Zero;
            }
            Divisor = A[r][k]; 
            r++;
         } 
      }
   }
   mpz_init_set_si( &MPSign, Sign );
   if ( r == m  ) 
      Det = K( (MP_INT*)&MPSign ) * Divisor;
   else
      Det = Zero;

   mpz_clear( &MPOne  );
   mpz_clear( &MPZero );
   mpz_clear( &MPSign );
   return( Det );
}

#endif
