
/*-------------------------------------------------------------
  module: bbratfunc.h
  author: Angel L. Diaz

  date: 10/21/96

  description: 
     The class in this module realizes the black box representation 
     of the common object rational function. In the FoxBox library, a
     rational function black box object is a function object which 
     when supplied with values (a pointer to an array of field elements) 
     for the symbols, computes the value for the rational function.

  class:
     BlackBoxURatFunc 

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.

-------------------------------------------------------------*/

#ifndef _BBRATFUNC
#define _BBRATFUNC

#include "BlackBox/CommonObjects/blackbox.h"

/*-------------------------------------------------------------
  class: BlackBoxURatFunc 
  date: 10/21/96

  description:
     Provides for constructing a common object black box 
     rational function from a pointer to a function. It is expected 
     that the "unknown" function takes as input an array of
     coefficient domain elements (K) and returns a coefficient 
     domain element.

  assumptions:
    * evaluates f(x1,...,xn)/g(x1,...,xn) in K(x1,...,xn)
    * GCD(f,g) = 1.
    * the evaluation function returns NULL on a root of g(x1,...,xn)

-------------------------------------------------------------*/

template< class K > 
class BlackBoxURatFunc : public BlackBox< K > {
   private:
      typedef K ( *FPTR )( K* );
      FPTR RationalFunction;
      K Value;
   public:
      BlackBoxURatFunc( );
      BlackBoxURatFunc( const BlackBoxURatFunc< K >& );
      BlackBoxURatFunc( FPTR         RatFunc_i, 
                        unsigned int NumVars_i, 
                        unsigned int Deg_i,    // bound on deg( g(x1,...,xn) )
                        double       Prob_i
                      ) : BlackBox< K >() 
      { 
         RationalFunction = RatFunc_i; 
         NumVariables     = NumVars_i; 
         Degree           = Deg_i; 
         Probability      = Prob_i;
      }

      ~BlackBoxURatFunc() { }

      inline K* operator () ( K* );
      inline BlackBoxURatFunc< K >& operator=( const BlackBoxURatFunc< K >& );
};

template< class K > 
BlackBoxURatFunc< K >::BlackBoxURatFunc( ) : BlackBox< K >( ) {
}

template< class K >
BlackBoxURatFunc< K >::BlackBoxURatFunc( 
   const BlackBoxURatFunc< K >& RatFunc_i 
   ) : BlackBox< K >( ) 
{ 
   *this = RatFunc_i; 
}

template< class K > 
inline K* BlackBoxURatFunc< K >::operator () (
   K *Pnts_i
   ) 
{ 
   Value = ( *RationalFunction )( Pnts_i ); 
   return &Value;
}

template< class K > 
inline BlackBoxURatFunc< K >& BlackBoxURatFunc< K >::operator=(
   const BlackBoxURatFunc< K >& RatFunc_i
   )
{ 
   RationalFunction = RatFunc_i.RationalFunction; 
   NumVariables     = RatFunc_i.NumVariables;
   Degree           = RatFunc_i.Degree; 
   Probability      = RatFunc_i.Probability;

   return *this; 
}

#endif
