
/*-------------------------------------------------------------
  module: bbtoeplitz.h
  author: Angel L. Diaz

  date: 10/22/96

  description: 
     The classes in this module realizes the black box representation 
     of the determinant of a symmetric toeplitz matrix common object.
     In the FoxBox library, a symmetric toeplitz determinant black box object 
     is a function object which when supplied with values (a pointer to an 
     array of field elements) for the symbols, computes the value for the 
     symmetric toeplitz determinant.

  classes:
     BlackBoxSymToeDet

  functions:
     SubRes
     ToeDet

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBSYMTOEDET
#define _BBSYMTOEDET

#include "BlackBox/CommonObjects/blackbox.h"

#define BBTOE_MAXN 65
#define BBTOE_MAXM 65

/*-------------------------------------------------------------
  class: BlackBoxSymToeDet
  date: 10/22/96

  description:
     Provides for constructing a determinant of a symmetric 
     toeplitz matrix common black box object.

-------------------------------------------------------------*/

template< class K, class KPoly > 
class BlackBoxSymToeDet : public BlackBox < K > {
   private: 
      K Value;
   public:
      BlackBoxSymToeDet( );
      BlackBoxSymToeDet( const BlackBoxSymToeDet< K, KPoly >& );
      BlackBoxSymToeDet( unsigned int NumVars_i, 
                      unsigned int Deg_i, 
                      double       Prob_i
                    ) : BlackBox< K >() 
      { 
         NumVariables     = NumVars_i; 
         Degree           = Deg_i; 
         Probability      = Prob_i;
      }
      ~BlackBoxSymToeDet() { }

      inline K* operator () ( K* );
      inline BlackBoxSymToeDet< K, KPoly >& 
             operator=( const BlackBoxSymToeDet< K, KPoly >& );
};

template< class K, class KPoly > 
BlackBoxSymToeDet< K, KPoly >::BlackBoxSymToeDet( ) : BlackBox< K >( ) {
}

template< class K, class KPoly >
BlackBoxSymToeDet< K, KPoly >::BlackBoxSymToeDet( 
   const BlackBoxSymToeDet< K, KPoly >& Det_i 
   ) : BlackBox< K >( ) 
{ 
   *this = Det_i; 
}

template< class K, class KPoly > 
inline K* BlackBoxSymToeDet< K, KPoly >::operator () (
   K *Pnts_i
   ) 
{ 
   KPoly dummy;
   Value = ToeDet( Pnts_i, Pnts_i, NumVariables, dummy );
   return ( &Value );
}

template< class K, class KPoly > 
inline BlackBoxSymToeDet< K, KPoly >& BlackBoxSymToeDet< K, KPoly >::operator=(
   const BlackBoxSymToeDet< K, KPoly >& Det_i
   )
{ 
   NumVariables     = Det_i.NumVariables;
   Degree           = Det_i.Degree; 
   Probability      = Det_i.Probability;

   return *this; 
}

/*-------------------------------------------------------------
  function: SubRes
  date: 10/22/96

  reference:
     Algorithm is Brown's subresultant PRS in  D. E. Knuth,
     "The Art of Computer Programming, Vol. 2, Seminumerical 
      Algorithms," Ed. 2, Addison Wesley, Reading, MA, 1981

  description:
     compute the j-th subresultant of F and G w.r.t the variable x

-------------------------------------------------------------*/

template < class K, class KPoly >
KPoly SubRes(
   KPoly& F_i,
   KPoly& G_i,
   int    j_i,
   K&     dummy_i
        )
{
   KPoly U,V,R;
   KPoly SR;
   K ell,g,h;
   int Delta;
   K Zero,One,MinusOne;
   MP_INT MPZero;
   MP_INT MPOne;
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
 
   Zero = K( (MP_INT*)&MPZero ); 
   One  = K( (MP_INT*)&MPOne );
   MinusOne = Zero - One;
   U    = F_i;
   V    = G_i;
   g    = One;
   h    = One;
   SR   = Zero;
  
   Delta = U.dg() - V.dg();
   while ( V.dg() > j_i ) {
      ell   = V.lcoeff();
      R.prem( U * ell.Exp( Delta + 1), V, R );
      if ( R == Zero ) {
         SR = Zero;
         break;
      }
      
      U = V; V =  R * ( MinusOne.Exp(Delta+1)  / ( g * h.Exp( Delta ) ) );

      g = U.lcoeff(); h = h.Exp( 1 - Delta ) * g.Exp( Delta );

      Delta = U.dg() - V.dg();
   }

   g = V.lcoeff(); h = h.Exp( 1 - Delta ) * g.Exp( Delta );

   if ( V.dg() == j_i ) {
      SR = V * ( h / g ) ;
   } 
   else if ( ( U.dg() - 1 ) == j_i ) {
      SR = V * One; 
   }

   mpz_clear( &MPOne );
   mpz_clear( &MPZero );

   return( SR ); 
}

/*-------------------------------------------------------------
  function: ToeDet
  date: 10/22/96

  reference:

  description:

-------------------------------------------------------------*/

template < class K, class KPoly >
K ToeDet(
   K*     Row1_i,
   K*     Col1_i,
   int    n_i, 
   KPoly& dummy_i
        )
{
   KPoly Tau,F,SR;
   K Zero,Det;
   int i;
   MP_INT MPZero;
   MP_INT MPOne;
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );

   Zero = K( (MP_INT*)&MPZero ); 
   Tau  = Zero;
   
   // compute Tau
   for( i=n_i; i>=1; i-- ) 
      Tau = Tau + KPoly( (MP_INT*)&MPOne, n_i - i ) * Col1_i[i-1] ;
 
   for( i=2; i<=n_i; i++ ) 
      Tau = Tau + KPoly( (MP_INT*)&MPOne, i + n_i - 2 ) * Row1_i[i-1];

   F  = KPoly( (MP_INT*)&MPOne, 2 * n_i - 1 );
   SR = SubRes( F, Tau, n_i - 1, Zero );

   if ( SR.dg() < n_i - 1 ) 
      Det = Zero;
   else
      Det = SR.lcoeff();

   mpz_clear( &MPZero );
   mpz_clear( &MPOne );

   return( Det );
}

#endif
