
/*-------------------------------------------------------------
  module: bbvandermonde.h
  author: Angel L. Diaz

  date: 10/22/96

  description: 
     The classes in this module realizes the black box representation 
     of the determinant of a vandermonde matrix common object.
     In the FoxBox library, a vandermonde determinant black box object 
     is a function object which when supplied with values (a pointer to an 
     array of field elements) for the symbols, computes the value for the 
     vandermonde determinant.

  classes:
     BlackBoxVandDet

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBVANDDET
#define _BBVANDDET

#include "BlackBox/CommonObjects/blackbox.h"

#define BBVAND_MAXPNTS 65

/*-------------------------------------------------------------
  class: BlackBoxVandDet
  date: 10/22/96

  description:
     Provides for constructing a determinant of a vandermonde
     matrix common black box object.

-------------------------------------------------------------*/

template< class K > 
class BlackBoxVandDet : public BlackBox < K > {
   private:
      int UsedVars;
      int StartPnt;
      int EndPnt;
      int IsShift;
      K   Value;
   public:
      BlackBoxVandDet( );
      BlackBoxVandDet( const BlackBoxVandDet< K >& );
      BlackBoxVandDet( unsigned int NumVars_i, 
                       unsigned int Deg_i, 
                       double       Prob_i,
                       int          UsedVars_i,
                       int          StartPnt_i = 0,
                       int          EndPnt_i = 0,
                       int          IsShift_i = BBDEFS_FALSE
                    ) : BlackBox< K >() 
      { 
         Construct( NumVars_i, Deg_i, Prob_i, UsedVars_i, StartPnt_i, 
                    EndPnt_i, IsShift_i );
      }
      ~BlackBoxVandDet() { }
      inline Construct( unsigned int NumVars_i, 
                        unsigned int Deg_i, 
                        double       Prob_i,
                        int          UsedVars_i,
                        int          StartPnt_i = 0,
                        int          EndPnt_i = 0,
                        int          IsShift_i = BBDEFS_FALSE
                      )
      {
         NumVariables     = NumVars_i; 
         Degree           = Deg_i; 
         Probability      = Prob_i;
         UsedVars         = UsedVars_i;
         StartPnt         = StartPnt_i;
         EndPnt           = EndPnt_i;
         IsShift          = IsShift_i;
      }

      inline K* operator () ( K* );
      inline BlackBoxVandDet< K >& 
             operator=( const BlackBoxVandDet< K >& );
};

template< class K > 
BlackBoxVandDet< K >::BlackBoxVandDet( ) : BlackBox< K >( ) {
}

template< class K >
BlackBoxVandDet< K >::BlackBoxVandDet( 
   const BlackBoxVandDet< K >& Det_i 
   ) : BlackBox< K >( ) 
{ 
   *this = Det_i; 
}

template< class K > 
inline K* BlackBoxVandDet< K >::operator () (
   K *Pnts_i
   ) 
{ 
   K Pnts[BBVAND_MAXPNTS];
   int i,j;
   MP_INT MPOne;
   mpz_init_set_ui( &MPOne, 1 );

   Value = K( (MP_INT*)&MPOne );

   if ( IsShift == BBDEFS_FALSE ) {
      for( j=0; j<UsedVars; j++ ) 
         for( i=0; i<j; i++ ) 
            Value = Value*(Pnts_i[i]-Pnts_i[j]);
   }
   else {
      Pnts[0] = Pnts_i[0]; Pnts[1] = Pnts_i[1];
      for(i = StartPnt+2;i<EndPnt;i++) 
         Pnts[i-StartPnt] = Pnts_i[i];

         for( j=0; j<UsedVars; j++ ) 
           for( i=0; i<j; i++ ) 
              Value = Value*(Pnts[i]-Pnts[j]);
   }

   mpz_clear( &MPOne );

   return &Value;
}

template< class K > 
inline BlackBoxVandDet< K >& BlackBoxVandDet< K >::operator=(
   const BlackBoxVandDet< K >& Det_i
   )
{ 
   NumVariables     = Det_i.NumVariables;
   Degree           = Det_i.Degree; 
   Probability      = Det_i.Probability;

   return *this; 
}

#endif
