
/*-------------------------------------------------------------
  module: bbfachmapmpi.h
  author: Angel L. Diaz

  date: 10/28/96

  description: 
     The class in this module extends the homomorphic map of a black 
     box factors black box by realizing a MPI compliant parallel 
     interface. 

  class:
     BlackBoxFactorsHMapMPI

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#include "BlackBox/HomomorphicMaps/bbfachmap.h"
#include "BlackBox/Aux/bbdefsmpi.h"

/*-------------------------------------------------------------
  class: BlackBoxFactorsHMapMPI
  date: 10/28/96

  description:
     Provides for the remote evaluation of a homomorphic map 
     of a factors black box object.

-------------------------------------------------------------*/

template < class K, class KP, class F, class FP, class B, class FB, class H > 
class BlackBoxFactorsHMapMPI : public BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H> {
   public:
      inline friend ofstream& operator<<(ofstream&, 
         BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>& );
      inline friend ifstream& operator>>(ifstream&, 
         BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>& );

      BlackBoxFactorsHMapMPI( FB& );
      BlackBoxFactorsHMapMPI( FB&, BlackBoxFactors<K,KP,B>&, H&);
      ~BlackBoxFactorsHMapMPI( ); 

      inline void Distribute( int,  F*, int&, int& );
      inline F*   Wait( int, int& );
      inline void Kill( int, int& );
};

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline F* BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>::Wait(
   int  Source_i,
   int& Status_m
   )
{
   int i;
   MPI_Status status;                 // Return status for receive 
   char ZPElmStr[BBMPI_KSTRZ];

   for( i=0; i<NumFactors; i++ ) {
      MPI_Recv( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Source_i, 
                BBMPI_FACTORS_ANS, MPI_COMM_WORLD, &status);
      ValueOfFactors[i] = F( ZPElmStr );
   }

   return ValueOfFactors;
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
inline void BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>::Kill(
   int  Dest_i,
   int& Status_m
   )
{
   int Request = BBMPI_KILL;  
   MPI_Send( &Request, 1, MPI_INT, Dest_i, BBMPI_REQUEST, MPI_COMM_WORLD);
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
inline void BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>::Distribute(
   int  Dest_i,
   F*   Pnts_i,
   int& IsDist_m,
   int& Status_m
   )
{
   int i,j,k,deg;
   int Request;       
   char ZPElmStr[BBMPI_KSTRZ];

   if ( !(IsDist_m) ) {
   // request a Factorization construction
   Request = BBMPI_FACTORS;
   MPI_Send( &Request, 1, MPI_INT, Dest_i, BBMPI_REQUEST, MPI_COMM_WORLD);
   MPI_Send( &StaticDeg, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC, 
             MPI_COMM_WORLD);
   MPI_Send( &NumVariables, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   MPI_Send( &NumFactors, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   MPI_Send( &SumExps, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   MPI_Send( &Sign, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   MPI_Send( &DegF, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   MPI_Send( &ParFracR, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   MPI_Send( &ParFracC, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   for( i=0; i<NumFactors; i++ )
      MPI_Send( &Exps[i], 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
                MPI_COMM_WORLD);
   for( i=0; i<NumVariables-1; i++ )  {
      RandA[i].toString( ZPElmStr );
      MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC, 
                MPI_COMM_WORLD);
   }
   for( i=0; i<NumVariables; i++ ) {
      RandB[i].toString( ZPElmStr );
      MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC, 
                MPI_COMM_WORLD);
   }
   for( i=0; i<NumVariables-1; i++ )  {
      RandC[i].toString( ZPElmStr );
      MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC, 
                MPI_COMM_WORLD);
   }
   ProbeElm.toString( ZPElmStr );
   MPI_Send( &ZPElmStr , BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC, 
             MPI_COMM_WORLD);
   GenElm.toString( ZPElmStr );
   MPI_Send( &ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   Cont.toString( ZPElmStr );
   MPI_Send( &ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
   for( i=0; i<=StaticDeg; i++ ) {
      Ind[i].toString( ZPElmStr );
      MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC,
                MPI_COMM_WORLD);
   }
   for( i=1; i<=StaticDeg; i++ ) {
      Reciprocal[i].toString( ZPElmStr );
      MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, BBMPI_FACTORS_STATIC,
                MPI_COMM_WORLD);
   }
   for( j=0; j<=StaticDeg; j++ ) {
      deg = Basis[j].dg();
      MPI_Send( &deg, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
      for( i=0; i<=deg; i++ ) {
         (Basis[j])[i].toString( ZPElmStr );
         MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, 
                   BBMPI_FACTORS_STATIC, MPI_COMM_WORLD);
      }
   }
   for( j=0; j<NumFactors; j++ ) {
      deg = FactorImages[j].dg();
      MPI_Send( &deg, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
      for( i=0; i<=deg; i++ ) {
         (FactorImages[j])[i].toString( ZPElmStr );
         MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, 
                   BBMPI_FACTORS_STATIC, MPI_COMM_WORLD);
      }
   }
   for( j=0; j<SumExps; j++ ) {
      deg = FactorProducts[j].dg();
      MPI_Send( &deg, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
             MPI_COMM_WORLD);
      for( i=0; i<=deg; i++ ) {
         (FactorProducts[j])[i].toString( ZPElmStr );
         MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, 
                   BBMPI_FACTORS_STATIC, MPI_COMM_WORLD);
      }
   }
   deg = pgsem1.dg();
   MPI_Send( &deg, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
            MPI_COMM_WORLD);
   for( i=0; i<=deg; i++ ) {
       pgsem1[i].toString( ZPElmStr );
       MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, 
                 BBMPI_FACTORS_STATIC, MPI_COMM_WORLD);
   }
   for( i=0; i<=ParFracC; i++ ) {
      for( j=0; j<=ParFracR; j++ ) {
         deg = ParFrac[i][j].dg();
         MPI_Send( &deg, 1, MPI_INT, Dest_i, BBMPI_FACTORS_STATIC,
                   MPI_COMM_WORLD);
         for( k=0; k<=deg; k++ ) {
            (ParFrac[i][j])[k].toString( ZPElmStr );
            MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i, 
                      BBMPI_FACTORS_STATIC, MPI_COMM_WORLD);
         }
      }
   }
   IsDist_m = BBDEFS_TRUE;
   }

   // request a Factors evaluation
   Request = BBMPI_FACTORS_EVAL;       
   MPI_Send( &Request, 1, MPI_INT, Dest_i, BBMPI_REQUEST, MPI_COMM_WORLD);

   // send points
   for( i=0; i<NumVariables; i++ ) {
      Pnts_i[i].toString( ZPElmStr );
      MPI_Send( ZPElmStr, BBMPI_KSTRZ, MPI_CHAR, Dest_i,
                BBMPI_FACTORS_STATIC, MPI_COMM_WORLD);
   }

}

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline ofstream& operator<<(
   ofstream& of_i,
   BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>& BB_i
   )
{
   BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>* BB;
   BB = &BB_i;    // derived* to base*
   of_i << *BB;
}

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline ifstream& operator>>(
   ifstream& if_i,
   BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>& BB_i
   )
{
   BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>* BB;
   BB = &BB_i;    // derived* to base*
   if_i >> *BB;
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>::BlackBoxFactorsHMapMPI( 
   FB& FB_i
   ) : BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>( FB_i )
{
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>::BlackBoxFactorsHMapMPI(
   FB& FB_i,
   BlackBoxFactors<K,KP,B>& BB_i,
   H& h_i
   ) : BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>( FB_i, BB_i, h_i )
{
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxFactorsHMapMPI<K,KP,F,FP,B,FB,H>::~BlackBoxFactorsHMapMPI( )
{
}

