/*-------------------------------------------------------------
  module: ntl2.0.h

  author: Angel Diaz and Erich Kaltofen
  date: 01/23/98

  Realizes a black box compliant univariate polynomial and
  field arithmetic plug-in to V. Shoup's package NTL 2.0;
  Modulus must be a prime < 2^30 (zz_p and zz_pX classes)
-------------------------------------------------------------*/

#ifndef NTL_PLUGIN
#define NTL_PLUGIN

#define NTL_SHMaxRand 2147483647

#include "BlackBox/Aux/bbdefs.h"
#include "PlugIns/NTL2.0/zz_pX.h"
#include "PlugIns/NTL2.0/zz_pXFactoring.h"
#include "PlugIns/gmp.h"

class NTLzz_p;         
class NTLzz_pX; 

MP_INT NTL_MPCard; 
long   NTLzz_pCard;

void NTLInitEnv( MP_INT* );
void NTLCleanUpEnv( ); 

void NTLInitEnv(
   MP_INT* MPCard_i
   ) 
{ 
  mpz_init_set( &NTL_MPCard, MPCard_i ); 
  NTLzz_pCard = mpz_get_si( MPCard_i );

  // friend void zz_pInit(long NewP, long maxroot=FFTMaxRoot);
  zz_pInit( NTLzz_pCard );
}

void NTLCleanUpEnv( ) 
{ 
   mpz_clear( &NTL_MPCard );
}

/*-------------------------------------------------------------
  class: NTLzz_p

  date: 01/23/98

  Black box compliant field arithmetic plug-in to NTL 2.0
-------------------------------------------------------------*/

class  NTLzz_p {
   private:
      zz_p NTLzz_pElm; 
   public:
      friend NTLzz_pX;
      inline friend ostream& operator<<( ostream&, const NTLzz_p& );

      NTLzz_p();
      NTLzz_p( const long& );
      NTLzz_p( const zz_p& );
      NTLzz_p( const MP_INT* );
      NTLzz_p( const char* );
      NTLzz_p( const NTLzz_p& );
      ~NTLzz_p();
  
      inline NTLzz_p& operator= ( const NTLzz_p&  );
      inline friend int operator==( const NTLzz_p&, const NTLzz_p& );
      inline friend int operator!=( const NTLzz_p&, const NTLzz_p& );

      inline NTLzz_p operator*( const NTLzz_p& );
      inline NTLzz_p operator/( const NTLzz_p& );
      inline NTLzz_p operator+( const NTLzz_p& );
      inline NTLzz_p operator-( const NTLzz_p& );

      inline NTLzz_p DistinctSelect( );
      inline MP_INT card( )               { return NTL_MPCard; }
      inline NTLzz_p Exp( int );
      inline NTLzz_p Denom( );
      inline NTLzz_p LCM( NTLzz_p*, unsigned int );
      inline void toString( char* ); 

      class RandomIterator {
         public:
            RandomIterator(const unsigned int, const MP_INT*);
            ~RandomIterator()        { mpz_clear(&MPCard); }
            NTLzz_p operator()();
         private:
            long RandStateArray[32];
            unsigned int RandSeed;
            MP_INT MPCard;
            long   NTLzz_pCard;
      };

};

inline ostream& operator<<( 
   ostream& os, 
   const NTLzz_p& FielElm_i
  )
{
   os << FielElm_i.NTLzz_pElm;
   return os;
}

NTLzz_p::NTLzz_p( ) { 
}

NTLzz_p::NTLzz_p( 
   const long& FielElm_i 
   )   
{ 
   NTLzz_pElm <<  FielElm_i; 
}

NTLzz_p::NTLzz_p( 
   const zz_p& FielElm_i 
   )   
{ 
   NTLzz_pElm = FielElm_i; 
} 

NTLzz_p::NTLzz_p( const MP_INT *MPInt_i ) 
{
   MP_INT MPRem;
   mpz_init( &MPRem );
 
   mpz_mod( &MPRem, MPInt_i, &NTL_MPCard );

   NTLzz_pElm <<  mpz_get_si( &MPRem );
   mpz_clear( &MPRem );
}

NTLzz_p::NTLzz_p( const char* Pntr_i ) 
{
   long a;
   sscanf(Pntr_i, "%i", &a);

   NTLzz_pElm << a;
}

NTLzz_p::NTLzz_p(
   const NTLzz_p& FielElm_i
   )    
{ 
   NTLzz_pElm = FielElm_i.NTLzz_pElm; 
}

NTLzz_p::~NTLzz_p( ) 
{ 
}

inline NTLzz_p& NTLzz_p::operator= ( 
   const NTLzz_p& FielElm_i
   )
{ 
   NTLzz_pElm = FielElm_i.NTLzz_pElm; 
   return *this; 
} 

inline int operator==(
   const NTLzz_p& FielElmA_i, 
   const NTLzz_p& FielElmB_i
   )
{ 
   return( FielElmA_i.NTLzz_pElm == FielElmB_i.NTLzz_pElm ); 
}

inline int operator!=(
   const NTLzz_p& FielElmA_i, 
   const NTLzz_p& FielElmB_i
   )
{ 
   return( FielElmA_i.NTLzz_pElm != FielElmB_i.NTLzz_pElm ); 
}

inline NTLzz_p NTLzz_p::operator*(
   const NTLzz_p& FielElm_i
   )
{ 
   NTLzz_p K ; 
   mul( K.NTLzz_pElm, this->NTLzz_pElm, FielElm_i.NTLzz_pElm ); 
   return K; 
}

inline NTLzz_p NTLzz_p::operator/(
   const NTLzz_p& FielElm_i
   )
{  
   NTLzz_p K; 
   inv( K.NTLzz_pElm, FielElm_i.NTLzz_pElm ); 
   if ( *this == (NTLzz_p) 1 ) 
      return K;
   return ( (*this) * K ); 
}

inline NTLzz_p NTLzz_p::operator+( 
   const NTLzz_p& FielElm_i 
   )
{ 
   NTLzz_p K; 
   add( K.NTLzz_pElm, this->NTLzz_pElm, FielElm_i.NTLzz_pElm ); 
   return K; 
}

inline NTLzz_p NTLzz_p::operator-(
   const NTLzz_p& FielElm_i
   )
{ 
   NTLzz_p K; 
   sub( K.NTLzz_pElm, this->NTLzz_pElm, FielElm_i.NTLzz_pElm ); 
   return K; 
}

inline NTLzz_p NTLzz_p::DistinctSelect( ) {
   NTLzz_p K;
   K = *this;

   add( this->NTLzz_pElm, this->NTLzz_pElm, zz_p( 1 ) ); 
   return K;
}

inline NTLzz_p NTLzz_p::Exp( 
   int n_i
   ) 
{
   unsigned int i;
   int Sign;
   zz_p Prod;
   zz_p InvProd;
 
   if ( n_i == 0 ) {
      Prod <<  1;
      return NTLzz_p( Prod );
   }
   else
      Prod = NTLzz_pElm;

    if ( n_i < 0 ) {
       Sign = -1;
       n_i = n_i * -1;
    }
    else
       Sign = 1;

   for(i=1;i<n_i;i++) 
      mul( Prod, Prod, NTLzz_pElm );

   if ( Sign == 1 )
       return NTLzz_p( Prod );
    else {
       inv( InvProd, Prod ); 
       return NTLzz_p( InvProd );
    }

}

inline NTLzz_p NTLzz_p::Denom( ) 
{
   return NTLzz_p( (long)1 );
}

inline NTLzz_p NTLzz_p::LCM(
   NTLzz_p* NTLzz_pElmArray_i, 
   unsigned int NumElms_i 
   ) 
{
   return NTLzz_p( (long)1);
}

inline void NTLzz_p::toString(
   char* Pntr_m
   ) 
{ 
   sprintf(Pntr_m,"%i", rep( NTLzz_pElm) );
}

NTLzz_p::RandomIterator::RandomIterator(
   const unsigned int Seed_i, 
   const MP_INT* Card_i
   ) : RandSeed(Seed_i) 
{
   if ( mpz_cmp_ui( Card_i, 0 ) < 0 ) {
       cout << "NTLzz_p::RandomIterator : input card < 0" << endl;
       exit(0);
   }

   mpz_init_set( &MPCard, Card_i );
   NTLzz_pCard = mpz_get_ui( &MPCard );

   for(int i=0;i<32;i++)
      RandStateArray[i] = (long)i;
   initstate( Seed_i, (char*) RandStateArray, sizeof(RandStateArray) );
}

NTLzz_p NTLzz_p::RandomIterator::operator()( ) 
{
   char *PrevState;
   long r1;

   PrevState = (char*)setstate((char*)RandStateArray);

   r1 = random();

   setstate(PrevState);

   return( NTLzz_p( r1 ) );
}

/*-------------------------------------------------------------
  class: NTLzz_pX

  date: 8/6/96

  Black box compliant univariate polynomial arithmetic plug-in 
  to v. shoup.
-------------------------------------------------------------*/

class NTLzz_pX {
   //private:
   public:
      zz_pX PolyElm;
   public: 
      friend ostream& operator<<( ostream&, const NTLzz_pX& );

      NTLzz_pX( );
      NTLzz_pX( const long& );
      NTLzz_pX( const NTLzz_p& );
      NTLzz_pX( const NTLzz_pX& );
      NTLzz_pX( const zz_pX& );
      NTLzz_pX( const char* );
      NTLzz_pX( const NTLzz_p&, unsigned int ); 
      ~NTLzz_pX( );

      inline NTLzz_pX& operator= (const NTLzz_pX& );
      inline friend int operator==(const NTLzz_pX&, const NTLzz_pX& );
      inline friend int operator!=(const NTLzz_pX&, const NTLzz_pX& );

      inline NTLzz_pX operator*( const NTLzz_pX& );
      inline NTLzz_pX operator/( const NTLzz_pX& );
      inline NTLzz_pX operator+( const NTLzz_pX& );
      inline NTLzz_pX operator-( const NTLzz_pX& );

      inline NTLzz_p operator[] ( unsigned int i );
      inline unsigned int dg( );
      inline NTLzz_p eval( const NTLzz_p* );
      inline NTLzz_p lcoeff( ); 
      inline void pgcd(const NTLzz_pX&, const NTLzz_pX&, NTLzz_pX& ); 
      inline void prem(const NTLzz_pX&, const NTLzz_pX&, NTLzz_pX& );
      inline void pquorem(const NTLzz_pX&, const NTLzz_pX&, NTLzz_pX&, NTLzz_pX&);
      inline void peea( const NTLzz_pX&, const NTLzz_pX&, NTLzz_pX&,
                        NTLzz_pX&, NTLzz_pX& );
      inline void factors( const NTLzz_pX&, int&, NTLzz_p&, unsigned int&,
                           unsigned int*, NTLzz_pX* );

      inline NTLzz_pX Diff( );

      inline void toString( char* ); 
};

ostream& operator<<(
   ostream& os, 
   const NTLzz_pX& NTLzz_pX_i
   )
{
   os << NTLzz_pX_i.PolyElm;
   return os;
}

NTLzz_pX::NTLzz_pX( ) 
{
}

NTLzz_pX::NTLzz_pX( 
   const long& PolyElm_i 
   )       
{ 
   PolyElm << PolyElm_i;
}

NTLzz_pX::NTLzz_pX(
   const NTLzz_p& NTLzz_pElm_i
   )        
{ 
   PolyElm <<  NTLzz_pElm_i.NTLzz_pElm;
}

NTLzz_pX::NTLzz_pX(
   const NTLzz_pX& NTLzz_pX_i
   )       
{ 
   PolyElm = NTLzz_pX_i.PolyElm;  
}

NTLzz_pX::NTLzz_pX(
   const zz_pX& zz_pX_i
   )
{
   PolyElm = zz_pX_i;
}


NTLzz_pX::NTLzz_pX( const char* Pntr_i ) 
{
   cout << "NTLzz_pX::NTLzz_pX( const char* Pntr_i ) not implemented!" << endl;
   exit(0);
}

NTLzz_pX::NTLzz_pX(
   const NTLzz_p& NTLzz_pElm_i, 
   unsigned int Deg_i
   ) 
{ 
   long deg = Deg_i;
   PolyElm = zz_pX( deg, NTLzz_pElm_i.NTLzz_pElm );  
}

NTLzz_pX::~NTLzz_pX( ) 
{ 
}

inline NTLzz_pX& NTLzz_pX::operator= (
   const NTLzz_pX& NTLzz_pX_i
   )
{ 
   PolyElm = NTLzz_pX_i.PolyElm; 
   return *this; 
}

inline int operator==(
   const NTLzz_pX& NTLzz_pXA_i, 
   const NTLzz_pX& NTLzz_pXB_i
   )
{ 
   return( NTLzz_pXA_i.PolyElm == NTLzz_pXB_i.PolyElm ); 
}

inline int operator!=(
   const NTLzz_pX& NTLzz_pXA_i, 
   const NTLzz_pX& NTLzz_pXB_i
   )
{ 
   return( NTLzz_pXA_i.PolyElm != NTLzz_pXB_i.PolyElm ); 
}

inline NTLzz_pX NTLzz_pX::operator*(
   const NTLzz_pX& NTLzz_pXA_i
   )
{ 
   NTLzz_pX P; 
   mul( P.PolyElm, this->PolyElm, NTLzz_pXA_i.PolyElm ); 
   return P; 
}

inline NTLzz_pX NTLzz_pX::operator/(
   const NTLzz_pX& NTLzz_pXA_i
   )
{ 
   NTLzz_pX q; 
   div( q.PolyElm, this->PolyElm, NTLzz_pXA_i.PolyElm );
   return q; 
}

inline NTLzz_pX NTLzz_pX::operator+(
   const NTLzz_pX& NTLzz_pXA_i
   )
{ 
   NTLzz_pX P; 
   add( P.PolyElm, this->PolyElm, NTLzz_pXA_i.PolyElm ); 
   return P; 
}

inline NTLzz_pX NTLzz_pX::operator-( 
   const NTLzz_pX& NTLzz_pXA_i
   )
{ 
   NTLzz_pX P; 
   sub( P.PolyElm, this->PolyElm, NTLzz_pXA_i.PolyElm ); 
   return P; 
}

inline NTLzz_p NTLzz_pX::operator[] (
   unsigned int Deg_i
   ) 
{ 
   NTLzz_p K; 
   K.NTLzz_pElm = coeff( PolyElm, Deg_i); 
   return K;
}

inline unsigned int NTLzz_pX::dg( ) 
{ 
   long nDeg = deg( PolyElm );

   if ( nDeg < 0 ) nDeg = 0;

   return nDeg; 
}

inline NTLzz_p NTLzz_pX::eval(
   const NTLzz_p pnts_i[]
   )
{
   NTLzz_p res;

   ::eval( res.NTLzz_pElm, PolyElm, pnts_i[0].NTLzz_pElm );

   return( res );
}

inline NTLzz_p NTLzz_pX::lcoeff( ) 
{ 
   NTLzz_p K; 
   K.NTLzz_pElm = LeadCoeff( PolyElm );
   return K;
}

inline void NTLzz_pX::pgcd(
   const NTLzz_pX& NTLzz_pXA_i, 
   const NTLzz_pX& NTLzz_pXB_i,
   NTLzz_pX& NTLzz_pX_o
   ) 
{ 
   GCD( NTLzz_pX_o.PolyElm, NTLzz_pXA_i.PolyElm, NTLzz_pXB_i.PolyElm ); 
}

inline void NTLzz_pX::pquorem( 
   const NTLzz_pX& NTLzz_pXA_i, 
   const NTLzz_pX& NTLzz_pXB_i,
   NTLzz_pX& NTLzz_pXQ_o,
   NTLzz_pX& NTLzz_pXR_o
   )
{
   DivRem( NTLzz_pXQ_o.PolyElm, NTLzz_pXR_o.PolyElm, NTLzz_pXA_i.PolyElm, 
           NTLzz_pXB_i.PolyElm); 

}

inline void NTLzz_pX::prem(
   const NTLzz_pX& NTLzz_pXA_i, 
   const NTLzz_pX& NTLzz_pXB_i,
   NTLzz_pX& NTLzz_pXR_o
   )
{
   rem( NTLzz_pXR_o.PolyElm, NTLzz_pXA_i.PolyElm, NTLzz_pXB_i.PolyElm); 
}

inline NTLzz_pX NTLzz_pX::Diff( ) 
{
   NTLzz_pX P;
   
   diff( P.PolyElm, PolyElm );

   return P;
}

inline void NTLzz_pX::toString( 
   char* Pntr_m
   ) 
{
   cout << "void NTLzz_pX::toString: not implemented!" << endl;
   exit(0);
}

inline void NTLzz_pX::peea(
   const NTLzz_pX& NTLzz_pXA_i,
   const NTLzz_pX& NTLzz_pXB_i,
   NTLzz_pX& NTLzz_pXG_m,
   NTLzz_pX& NTLzz_pXS_m,
   NTLzz_pX& NTLzz_pXT_m
   )
{
   // d = gcd(a,b), a s + b t = d
   XGCD( NTLzz_pXG_m.PolyElm, NTLzz_pXS_m.PolyElm, NTLzz_pXT_m.PolyElm,
         NTLzz_pXA_i.PolyElm, NTLzz_pXB_i.PolyElm );
}

inline void NTLzz_pX::factors(
   const NTLzz_pX& NTLzz_pX_i,
   int& Sign_o,
   NTLzz_p& Cont_o,
   unsigned int& NumFactors_o,
   unsigned int *Exps_o,
   NTLzz_pX *Factors_o
   )
{
   int i;
   zz_p  LC, LCInv;
   zz_pX MonicPoly;
   Vector(pair(zz_pX,long)) Facs;

   GetCoeff( LC, NTLzz_pX_i.PolyElm, deg( NTLzz_pX_i.PolyElm ) ); 
   inv( LCInv, LC); 
   mul( MonicPoly, NTLzz_pX_i.PolyElm, LCInv );
   berlekamp( Facs, MonicPoly ); 

   Sign_o            = 1;
   Cont_o.NTLzz_pElm = LC;
   NumFactors_o      = Facs.length();
   for(i=0; i<NumFactors_o; i++) {
      Factors_o[i] = Facs[i].a;
      Exps_o[i]    = Facs[i].b;
   }
}

#endif
