/*-------------------------------------------------------------
  module: saclib.h

  author: Angel L. Diaz
  date: 8/6/96

  Realizes a black box compliant univariate polynomial and
  field arithmetic plug-in to SACLIB.
-------------------------------------------------------------*/

#ifndef SACLIB_PLUGIN
#define SACLIB_PLUGIN
#define SACLIB_MAXFACTORS 10

#include "BlackBox/Aux/bbdefs.h"
#include "PlugIns/Saclib/saclib.h"
#include "PlugIns/gmp.h"
#include <ctype.h>

class SaclibQ;         
class SaclibQX; 
class SaclibQXT;
class SaclibZP;
class SaclibZPX;

MP_INT SACLIB_MPCard;   
Word   SACLIB_SLCard       = NIL; 
Word   SACLIB_SLMaxRand    = NIL;
Word   SACLIB_VariableName = NIL;
int    SACLIB_TruncDeg     = 0;

void SaclibInitEnv( long , Word& ); 
void SaclibSetPrime( const MP_INT* );
void SaclibCleanUpEnv( ); 
inline Word SaclibBoundMPInt( const MP_INT *, MP_INT * );
inline Word SaclibBinarySearch( const MP_INT * );
inline Word ConvertMPToSacInt( const MP_INT * );
inline Word ConvertLongToSacInt( long );
inline int RPEQL( Word, Word ); 
inline int IPEQL( Word, Word );

/*-------------------------------------------------------------
  Description: Auxiliary functions and procedures

  date: 8/6/96

  Utilized by the black box compliant field and polynomial 
  arithmetic plug-in to SACLIB, as well as black box applications.
-------------------------------------------------------------*/

inline Word SaclibBoundMPInt( const MP_INT *Key_i, MP_INT *Bound_o ) {
   Word Bound = 1;
   mpz_set_ui( Bound_o, 1 );

   while ( mpz_cmp( Key_i, Bound_o ) > 0 ) {
      mpz_mul_ui( Bound_o, Bound_o, 2);
      Bound = IPROD( Bound, 2 );
   }

   return Bound;
}

inline Word SaclibBinarySearch(const MP_INT *Key) {
   MP_INT MPPivot; 
   mpz_init_set_ui( &MPPivot, 0 );
   Word SLPivot = 0;

   MP_INT MPRightEnd; 
   mpz_init( &MPRightEnd );
   Word SLRightEnd = SaclibBoundMPInt( Key, &MPRightEnd );

   MP_INT MPLeftEnd; 
   mpz_init_set_ui( &MPLeftEnd, 1);
   Word SLLeftEnd = 1; 

   while ( mpz_cmp( &MPRightEnd, &MPLeftEnd ) >= 0 ) {
      mpz_add( &MPPivot, &MPLeftEnd, &MPRightEnd ); 
      SLPivot = ISUM( SLLeftEnd, SLRightEnd );

      mpz_div_ui( &MPPivot, &MPPivot, 2 );  
      SLPivot = IQ( SLPivot, 2 );

      if ( mpz_cmp( Key, &MPPivot ) == 0 ) break;
      if ( mpz_cmp( Key, &MPPivot ) < 0 ) { 
         mpz_sub_ui( &MPRightEnd, &MPPivot, 1 ); 
         SLRightEnd = IDIF( SLPivot, 1 ); 
      }
      else { 
         mpz_add_ui( &MPLeftEnd, &MPPivot, 1 ); 
         SLLeftEnd = ISUM( SLPivot, 1 ); 
      }
   }

   mpz_clear( &MPRightEnd );
   mpz_clear( &MPPivot );
   mpz_clear( &MPLeftEnd );

   return SLPivot;
}

inline Word ConvertMPToSacInt(const MP_INT *MPInt_i) {
   Word SLInt = 0;
   MP_INT MPInt;
   int IsPositive;

   mpz_init_set( &MPInt, MPInt_i );
   if ( mpz_cmp_si( &MPInt, 0 ) != 0 ) {

      IsPositive = Positive( &MPInt );
      if ( IsPositive == BBDEFS_FALSE ) 
         mpz_abs( &MPInt, &MPInt);

      SLInt = SaclibBinarySearch( &MPInt ); 

      if ( IsPositive == BBDEFS_FALSE ) 
         SLInt = INEG(SLInt);
   }
   mpz_clear( &MPInt ); 
   return SLInt;
}

inline Word ConvertLongToSacInt( long long_i ) {
   MP_INT MPInt;
   Word SLInt;

   mpz_init_set_ui( &MPInt, long_i );
   SLInt = ConvertMPToSacInt( &MPInt );
  
   mpz_clear( &MPInt ); 
   return SLInt;
}

inline void SaclibSetTruncDeg( const int d_i ) {
   SACLIB_TruncDeg = d_i;
}

void SaclibInitEnv(long a, Word& stack) { 
   char *VariableName = new char[2];
   strcpy(VariableName,"x");
   NU = a; 
 
   BEGINSACLIB(&stack); 
   GCM = 1; 
   STATSACLIB(); 

   GCGLOBAL(&SACLIB_SLCard);
   GCGLOBAL(&SACLIB_SLMaxRand);
   GCGLOBAL(&SACLIB_VariableName);

   SACLIB_VariableName = LIST1(LFS(VariableName));
   SACLIB_SLCard = -1; 
   SACLIB_SLMaxRand = ConvertLongToSacInt(2147483647); // 2^31-1
}  

void SaclibSetPrime( const MP_INT* MPCard_i ) {
   SACLIB_SLCard = ConvertMPToSacInt( MPCard_i );
}

void SaclibCleanUpEnv() { 
   STATSACLIB(); 
   ENDSACLIB(SAC_FREEMEM); 
   mpz_clear(&SACLIB_MPCard);
}

int AtomToString(
  Word A,            // Atom
  int i,             // insert at i
  char* AtomString_m // String rep. of Atom
  )
{
   Word Ap,D[21],N,Q;

   // Write minus sign
   if (A < 0) {
      Ap = -A;
      AtomString_m[i] = '-';
      i++;
   }
   else
      Ap = A;

   // Convert to decimal. 
   N = 0;
   do {
      Q = Ap / 10;
      N = N + 1;
      D[N] = Ap - 10 * Q;
      Ap = Q;
   } while ( !(Ap == 0) );

   // Write digits. */
   do {
      AtomString_m[i] = D[N] + '0';
      i++; 
      N = N - 1;
   } while (!(N == 0));
   goto Return;

Return: 
   AtomString_m[i] = '\0';
   return( i );  // location of \0
}

int IntegerToString(
  Word A,         // Integer
  int j,          // Insert at j
  char* IntString_m // String rep. of integer
  )
{
   Word Ap,B[21],H,h,i,q;

   // A single-precision
   if (A < BETA) {
      j = AtomToString(A,j,IntString_m);
      goto Return;
   }

   // Compute list of THETA-digits
   Ap = A;
   H = NIL;
   do {
      IDQR(Ap,THETA,&Ap,&h);
      H = COMP(h,H);
   } while (!(Ap == 0));

   // Write first THETA-digit
   ADV(H,&h,&H);
   j = AtomToString(h,j,IntString_m);

   // Write remaining digits
   while (H != NIL) {
      ADV(H,&h,&H);
      h = ABS(h);
      for (i = 1; i <= ETA; i++) {
          q = h / 10;
          B[i] = h - 10 * q;
          h = q;
      }
      for (i = ETA; i >= 1; i--) {
          IntString_m[j] = B[i] + '0';
          j++;
      }
   }
   IntString_m[j] = '\0';
  
Return: 
   return( j ); // location of \0
}

int RatNumToString(
  Word R,               // integer
  int i,                // insert at i
  char* RatNumString_m  // string rep. of rational number
  )
{
   Word R1,R2;

   // Write
   if (R == 0)
      i = AtomToString( R, i, RatNumString_m );
   else {
     FIRST2(R,&R1,&R2);
     i = IntegerToString( R1, i, RatNumString_m );
     if (R2 != 1) {
        RatNumString_m[i] = '/';
        i++;
        i = IntegerToString( R2, i, RatNumString_m );
     }
   }

Return: 
   RatNumString_m[i] = '\0';
   return( i ); // location of \0
}

void RatPolyToString(
  Word A,
  char* RatPolyString_m  // string rep. of rational polynomial
  )
{
  Word Ap,e,a,n1,p1;
  int i = 0;

  // A = 0 
  if (A == 0) {
     RatPolyString_m[0] = '0';
     RatPolyString_m[1] = '\0';
     return;
  }

  // Initialize. 
  Ap = A;
  n1 = RNINT(-1);
  p1 = RNINT(1);
  RatPolyString_m[i] = '(';
  i++;

Step3:
  // Write coefficient 
  FIRST2(Ap,&e,&a);
  if (EQUAL(a,n1) && e > 0) {
     RatPolyString_m[i] = '-';
     i++;
     goto Step4;
  }
  if (EQUAL(a,p1) && e > 0) {
     if (A != Ap) {
        RatPolyString_m[i] = '+';
        i++;
     }
     goto Step4;
  }
  if (RNSIGN(a) > 0 && A != Ap) {
     RatPolyString_m[i] = '+';
     i++;
  }
  i = RatNumToString( a, i, RatPolyString_m);

Step4: 
  // Write variable and exponent 
  if (e > 0) {
     RatPolyString_m[i] = 'x';
     i++;
  }
  if (e > 1) {
     RatPolyString_m[i] = '^';
     i++;
     i = AtomToString( e, i, RatPolyString_m );
  }
  Ap = RED2(Ap);
  if (Ap != NIL)
     goto Step3;

  // Close
  RatPolyString_m[i] = ')';
  i++;
  RatPolyString_m[i] = '\0';
}

Word CharToWord( 
   const char* String_i, 
   int& i_m 
   )
{
   Word C;

   // Get the current character 
   C = String_i[i_m];
   i_m++;
 
   return(C); // i_m location of char in string
}

Word NextCharToWord(
   const char* String_i, 
   int& i_m 
   )
{
   Word C;

   // Get the next non-white-space character
   do {
      C = CharToWord(String_i,i_m);
   } while (isspace(C));

   return(C);
}

Word LookAheadCharToWord(
   const char* String_i, 
   int& i_m 
   )
{
   Word C;
   // Get the next character. */
   C = NextCharToWord( String_i, i_m );
  
   i_m--;
   return( C );
}

Word StringToAtom(
   const char* String_i, 
   int& i_m 
   )
{
   Word A,C,S;

   // Skip blanks and read sign, if any. */
   S = 1;
   C = NextCharToWord( String_i, i_m );
   if (C == '+')
      C = CharToWord(String_i,i_m);
   else
   if (C == '-') {
      C = CharToWord(String_i,i_m);
      S = -1;
   }
   else
      if (DIGIT(C) == 0)
         goto Step3;

   // Read digits and convert
   A = 0;
   do {
      if (A > BETA / 10)
         goto Step3;
      A = 10 * A + C - '0';
      if (A >= BETA)
         goto Step3;
      C = CharToWord(String_i,i_m);
   } while (DIGIT(C) != 0);
   A = S * A;
   i_m--;
   goto Return;

Step3: // Error
   FAIL("StringToAtom","Unexpected character",C);

Return: 
   return(A);
}

Word StringToVar(
   const char* String_i, 
   int& i_m 
   )
{
   Word C,v;

   // Read
   C = NextCharToWord( String_i, i_m );
   if (LETTER(C) == 0)
      FAIL("StringToVar","Unexpected character",C);
   v = NIL;
   do {
      v = COMP(C,v);
      C = CharToWord( String_i, i_m );
   } while (LETTER(C) || DIGIT(C));
   v = INV(v);
   i_m--;
   return( v );
}

Word StringToInteger(
   const char* String_i,
   int& i_m
   )
{
   Word A,B,C,b,bp,i,s;

   // Read sign
   s = 1;
   C = NextCharToWord( String_i, i_m );
   if (C == '-')
      s = -1;
   if (C == '+' || C == '-')
      C = CharToWord( String_i, i_m );
   if (DIGIT(C) == 0)
      goto Step6;

   // Skip leading zeros
   while (C == 0)
         C = CharToWord( String_i, i_m );

   // Is A=0
   A = 0;
   if (DIGIT(C) == 0) {
      i_m--;
      goto Return;
   }

   // Compute THETA-digits
   B = NIL;
   b = 0;
   i = 0;
   do {
      b = 10 * b + C - '0';
      i = i + 1;
      if (i == ETA) {
         B = COMP(b,B);
         b = 0;
         i = 0;
      }
      C = CharToWord( String_i, i_m );
    } while (!(DIGIT(C) == 0));

    // Convert to base BETA
    B = INV(B);
    while (B != NIL) {
       A = IDPR(A,THETA);
       ADV(B,&bp,&B);
       A = ISUM(A,bp);
    }
    if (A != 0)
       A = IDPR(A,EXPF(10,i));
    A = ISUM(A,b);
    if (s < 0)
       A = INEG(A);
    i_m--;
    goto Return;

Step6: // Error
    FAIL("StringToInteger","Unexpected character",C);

Return: 
       return( A );
}

Word StringToRatNum(
   const char* String_i,
   int& i_m
   )
{
   Word C,R,R1,R2;
   
   // Read
   R1 = StringToInteger( String_i, i_m );
   C = CharToWord( String_i, i_m );
   if (C == '/')
      R2 = StringToInteger( String_i, i_m );
   else {
      R2 = 1;
      i_m--;
   }
   if (R1 == 0)
      R = 0;
   else
      R = LIST2(R1,R2);

  return(R);
}

void StringToRatPoly(
   const char* String_i,
   int& i_m,
   Word* A_i
   )
{
   Word A,C,Cp,a,e,r,rp,v;

   // r = 0, no variables
   C = NextCharToWord( String_i, i_m );
   if (C != '(') {
      i_m--;
      A = StringToRatNum( String_i, i_m );
      r = 0;
      goto Return;
   }

   // Initialize
   A = NIL;
   rp = 0;

   // Read first coefficient
   C = LookAheadCharToWord( String_i, i_m );
   if (LETTER(C) || DIGIT(C) || (C == '(')) {
      if (LETTER(C))
         a = RNINT(1);
      else 
         if (DIGIT(C))
            a = StringToRatNum( String_i, i_m );
         else
            if (C == '(')
               StringToRatPoly( String_i, i_m, &a );
      goto Step5;
    }

Step4: // Read coefficient
    C = NextCharToWord( String_i, i_m );
    if (C != '+' && C != '-')
       goto Step7;
    Cp = LookAheadCharToWord( String_i, i_m );
    if (LETTER(Cp) || DIGIT(Cp)) {
       if (LETTER(Cp))
          a = RNINT(1);
       else
          a = StringToRatNum( String_i, i_m );
       if (C == '-')
          a = RNNEG(a);
    }
    if (Cp == '(') {
       StringToRatPoly( String_i, i_m, &a );
       if (C == '-')
          a = RPNEG(rp,a);
    }
     
Step5: //  Read variable and exponent
    C = LookAheadCharToWord( String_i, i_m );
    if (LETTER(C)) {
       v = StringToVar( String_i, i_m );
       C = NextCharToWord( String_i, i_m );
       if (C != '^') {
          e = 1;
          i_m--;
       }
       else {
          C = LookAheadCharToWord( String_i, i_m );
          if (DIGIT(C))
             e = StringToAtom( String_i, i_m );
          else
             goto Step7;
       }
    }
    else 
       e = 0;

Step6: // Check for end
   A = COMP2(a,e,A);
   C = NextCharToWord( String_i, i_m );
   if (C != ')') {
      i_m--;
      goto Step4;
   }
   else
      r = rp + 1;
   A = INV(A);
   goto Return;
       
Step7: // Error
   FAIL("StringToRatPoly","Unexpected character",C);

Return: // Prepare for return
   *A_i = A;
   return;
}

/*-------------------------------------------------------------
  class: SaclibQ

  date: 8/6/96

  Black box compliant field arithmetic plug-in to SACLIB.
-------------------------------------------------------------*/

class SaclibQ {
   private:
      Word SaclibQElm;
   public:
      friend SaclibQX;
      friend SaclibQXT;
      inline friend ostream& operator<<(ostream& os, const SaclibQ& SaclibQ_i);

      SaclibQ()                        { } 
      SaclibQ(const Word& SaclibQElm_i)  { SaclibQElm = SaclibQElm_i; }
      SaclibQ(const long& SaclibQElm_i)  { SaclibQElm = RNINT(SaclibQElm_i); }
      SaclibQ(const MP_INT*);
      SaclibQ(const char* );
      SaclibQ(const SaclibQ& SaclibQ_i)    { SaclibQElm = SaclibQ_i.SaclibQElm; }
      ~SaclibQ()                       { } 

      inline SaclibQ& operator=( const SaclibQ& ); 
      inline friend int operator==( const SaclibQ&, const SaclibQ& );
      inline friend int operator!=( const SaclibQ&, const SaclibQ& );

      inline SaclibQ operator*(const SaclibQ&);
      inline SaclibQ operator/(const SaclibQ&);
      inline SaclibQ operator+(const SaclibQ&);
      inline SaclibQ operator-(const SaclibQ&);

      inline SaclibQ DistinctSelect( );
      inline MP_INT card( )          { return SACLIB_MPCard; }
      inline SaclibQ Exp( int );
      inline SaclibQ Denom( );
      inline SaclibQ LCM( SaclibQ*, unsigned int );
      inline void toString(char* Pntr_m) { RatNumToString(SaclibQElm,0,Pntr_m); }
      inline void NumtoString(char*); 
      inline void DenomtoString(char*);

      class RandomIterator {
         public:
            RandomIterator(const unsigned int, const MP_INT*);
            ~RandomIterator()        { mpz_clear(&MPCard); }
            SaclibQ operator()();
         private:
            long RandStateArray[32];
            unsigned int RandSeed;
            MP_INT MPCard;
            Word SLCard;
      };

      inline Word Native() { return SaclibQElm; }
};

SaclibQ::SaclibQ( const MP_INT *MPInt_i ) {
   SaclibQElm = RNINT( ConvertMPToSacInt( MPInt_i ) );
}

SaclibQ::SaclibQ(const char* String_i) {
   int i = 0;
   SaclibQElm = StringToRatNum( String_i, i );
}

inline ostream& operator<<(ostream& os, const SaclibQ& SaclibQ_i) { 
   RNWRITE( SaclibQ_i.SaclibQElm ); 
   return os; 
}

inline SaclibQ& SaclibQ::operator=(const SaclibQ& SaclibQ_i) { 
   this->SaclibQElm = SaclibQ_i.SaclibQElm; 
   return *this; 
}

inline int operator==(const SaclibQ& SaclibQA_i, const SaclibQ& SaclibQB_i) {
   if ( RNCOMP( SaclibQA_i.SaclibQElm, SaclibQB_i.SaclibQElm ) == 0 ) 
      return 1; 
   else 
      return 0; 
}

inline int operator!=(const SaclibQ& SaclibQA_i, const SaclibQ&  SaclibQB_i) { 
   if ( RNCOMP( SaclibQA_i.SaclibQElm, SaclibQB_i.SaclibQElm) == 0 ) 
      return 0; 
   else 
      return 1; 
}

inline SaclibQ SaclibQ::operator*(const SaclibQ& SaclibQ_i) { 
   return SaclibQ( RNPROD( this->SaclibQElm, SaclibQ_i.SaclibQElm ) ); 
}

inline SaclibQ SaclibQ::operator/(const SaclibQ& SaclibQ_i) { 
   return SaclibQ( RNQ( this->SaclibQElm, SaclibQ_i.SaclibQElm ) ); 
}

inline SaclibQ SaclibQ::operator+(const SaclibQ& SaclibQ_i) { 
   return SaclibQ( RNSUM( this->SaclibQElm, SaclibQ_i.SaclibQElm ) ); 
}

inline SaclibQ SaclibQ::operator-(const SaclibQ& SaclibQ_i) { 
   return SaclibQ( RNDIF( this->SaclibQElm, SaclibQ_i.SaclibQElm ) ); 
}

inline SaclibQ SaclibQ::DistinctSelect() {
   Word One        = RNINT(1); 
   SaclibQ NextSaclibQ = *this; 

   this->SaclibQElm  = RNSUM( this->SaclibQElm, One ); 
   return NextSaclibQ; 
} 

inline SaclibQ SaclibQ::Exp( int n_i ) {
    int i;
    int Sign;
    Word Prod;
 
    if ( n_i == 0 ) {
       Prod = RNINT(1);
       return SaclibQ( Prod );
    }
    else
       Prod = SaclibQElm;
  
    if ( n_i < 0 ) {
       Sign = -1;
       n_i = n_i * -1;
    }
    else
       Sign = 1;

    for(i=1;i<n_i;i++) 
       Prod = RNPROD( Prod, SaclibQElm );

    if ( Sign == 1 )
       return SaclibQ( Prod );
    else
       return SaclibQ( RNINV( Prod ) );
}

inline SaclibQ SaclibQ::Denom( ) {
   return SaclibQ( (Word)RNINT( RNDEN( SaclibQElm ) ) );
}

inline SaclibQ SaclibQ::LCM( 
   SaclibQ* SaclibQElmArray_i, 
   unsigned int NumElms_i 
   ) 
{
   // MUST IMPROVE THIS ALGORITHM 
   int i;
   Word LC;
   // it is asumed that the field elements are rationals with denominator 1
   LC = ILCM( RNNUM( SaclibQElmArray_i[0].SaclibQElm ) ,
              RNNUM( SaclibQElmArray_i[1].SaclibQElm ) );
   for( i=2; i< NumElms_i; i++ )
      LC = ILCM( LC, RNNUM( SaclibQElmArray_i[i].SaclibQElm ) );
   
   return SaclibQ( (Word)RNINT( LC ) );
}

inline void SaclibQ::NumtoString(
   char* Pntr_m
   ) 
{
 Word Num;
 Num = RNNUM( SaclibQElm );
 IntegerToString( Num, 0, Pntr_m );
}

inline void SaclibQ::DenomtoString(
   char* Pntr_m
   ) 
{
 Word Den;
 Den = RNDEN( SaclibQElm );
 IntegerToString( Den, 0, Pntr_m );
}

SaclibQ::RandomIterator::RandomIterator(const unsigned int Seed_i, 
const MP_INT* Card_i) : RandSeed(Seed_i) {
   if ( mpz_cmp_ui( Card_i, 0 ) < 0 ) 
      OutputErrorExit( BBERROR_SACLIB, 
         "SaclibQ::RandomIterator : input card < 0" ); 

   mpz_init_set( &MPCard, Card_i );
   SLCard = ConvertMPToSacInt( &MPCard );

   for(int i=0;i<32;i++) 
      RandStateArray[i] = (long)i;
   initstate( Seed_i, (char*) RandStateArray, sizeof(RandStateArray) );
}

SaclibQ SaclibQ::RandomIterator::operator()() {
   long r;
   Word rr;
   Word c,x;
   char *PrevState;

   x = 0; c = 1;

   PrevState = (char*)setstate((char*)RandStateArray);

   while ( ICOMP(c,SLCard) == -1 ) {   
      r = random()-1;           // random long [0,2147483647)
      rr = ConvertLongToSacInt(r); 
      x = IPROD(x,SACLIB_SLMaxRand);
      x = ISUM(x,rr);
      c = IPROD(c,SACLIB_SLMaxRand);
   }
  
   x = IREM(x,SLCard);
   setstate(PrevState);
   return(SaclibQ(RNINT(x)));
}

/*-------------------------------------------------------------
  class: SaclibQX

  date: 8/6/96

  Black box compliant univariate polynomial arithmetic plug-in 
  to SACLIB.
-------------------------------------------------------------*/

class SaclibQX {
   private:
      Word ElmQX;
   public:       
      friend SaclibQXT;
      inline friend ostream& operator<<( ostream&, const SaclibQX& );

      SaclibQX( ) { }
      SaclibQX( const long& ); 
      SaclibQX( const Word& );
      SaclibQX( const char* );
      SaclibQX( const Word&, const unsigned int& );
      SaclibQX( const SaclibQ&, const unsigned int&  );
      SaclibQX( const SaclibQ& );
      SaclibQX( const SaclibQX& );  
      SaclibQX( const long&, const int& );
      ~SaclibQX( ) { }

      inline SaclibQX& operator=( const SaclibQX& ); 
      inline friend int operator==( const SaclibQX&, const SaclibQX& );
      inline friend int operator!=( const SaclibQX&, const SaclibQX& );

      inline SaclibQX operator*( const SaclibQX& );
      inline SaclibQX operator/( const SaclibQX& );
      inline SaclibQX operator+( const SaclibQX& );
      inline SaclibQX operator-( const SaclibQX& );

      inline SaclibQ operator[] ( unsigned int ); 
      inline unsigned int dg( );
      inline SaclibQ eval( const SaclibQ* ); 
      inline SaclibQ lcoeff( );
      inline void prem( const SaclibQX&, const SaclibQX&, SaclibQX& );
      inline void pquorem( const SaclibQX&, const SaclibQX&, SaclibQX&, SaclibQX&);
      inline void pgcd( const SaclibQX&, const SaclibQX&, SaclibQX& ); 
      inline void peea( const SaclibQX&, const SaclibQX&, SaclibQX&,
                        SaclibQX&, SaclibQX& );
      inline void factors( const SaclibQX&, int&, SaclibQ&, unsigned int&,
         unsigned int*, SaclibQX* );
      inline SaclibQX Diff( );
      inline void toString( char* Pntr_m ) { RatPolyToString(ElmQX,Pntr_m); }
};

inline int RPEQL(
   Word RNPolyA_i, 
   Word RNPolyB_i 
   ) 
{
   int deg = PDEG( RNPolyA_i ); 

   if ( deg != PDEG( RNPolyB_i ) ) 
      return BBDEFS_FALSE;

   // handle the zero cases
   if ( ( RNPolyA_i == 0 ) && ( RNPolyB_i == 0 ) ) return  BBDEFS_TRUE; 
   if ( ( RNPolyA_i == 0 ) && ( RNPolyB_i != 0 ) ) return  BBDEFS_FALSE;
   if ( ( RNPolyA_i != 0 ) && ( RNPolyB_i == 0 ) ) return  BBDEFS_FALSE;
   
   Word c1,c2; 
   c1 = PCL( RNPolyA_i ); 
   c2 = PCL( RNPolyB_i );

   for(int i=1;i<=deg+1;i++) 
      if ( RNCOMP(LELTI(c1,i),LELTI(c2,i)) < 0 ) return  BBDEFS_FALSE;

   return  BBDEFS_TRUE; 
}

inline ostream& operator<<( ostream& os, const SaclibQX& SaclibQX_i ) { 
   RPWRITE( 1, SaclibQX_i.ElmQX, SACLIB_VariableName ); 
   return os; 
}

SaclibQX::SaclibQX( const long& SaclibQElm_i ) { 
   ElmQX = PMON( RNINT( SaclibQElm_i ) ,0 ); 
}

SaclibQX::SaclibQX( const Word& ElmQX_i ) { 
   ElmQX = ElmQX_i; 
}

SaclibQX::SaclibQX( const char* String_i ) { 
   int i = 0;
   StringToRatPoly( String_i, i, &ElmQX );
}

SaclibQX::SaclibQX( const Word& ElmQX_i, const unsigned int& Deg_i ) { 
   ElmQX = PMON( ElmQX_i, Deg_i ); 
}

SaclibQX::SaclibQX( const SaclibQ& SaclibQElm_i, const unsigned int& Deg_i ) { 
   ElmQX = PMON( SaclibQElm_i.SaclibQElm, Deg_i ); 
}

SaclibQX::SaclibQX( const SaclibQ& SaclibQElm_i ) { 
   ElmQX = PMON( SaclibQElm_i.SaclibQElm, 0 ); 
}

SaclibQX::SaclibQX( const SaclibQX& SaclibQX_i ) { 
   ElmQX = SaclibQX_i.ElmQX; 
}

SaclibQX::SaclibQX( const long& pe, const int& Deg_i ) { 
   ElmQX = PMON( RNINT( pe ), Deg_i ); 
}

inline SaclibQX& SaclibQX::operator=( const SaclibQX& SaclibQX_i ) { 
   this->ElmQX = SaclibQX_i.ElmQX; 
   return *this; 
}

inline int operator==( const SaclibQX& SaclibQXA_i, 
const SaclibQX& SaclibQXB_i ) { 
   return( RPEQL( SaclibQXA_i.ElmQX, SaclibQXB_i.ElmQX ) ); 
}

inline int operator!=( const SaclibQX& SaclibQXA_i, 
const SaclibQX& SaclibQXB_i ) { 
   return !( SaclibQXA_i == SaclibQXB_i );
}

inline SaclibQX SaclibQX::operator*( const SaclibQX& SaclibQX_i ) {  
   return SaclibQX( RPPROD( 1, ElmQX, SaclibQX_i.ElmQX ) ); 
}

inline SaclibQX SaclibQX::operator/( const SaclibQX& SaclibQX_i ) {  
   Word Q,R; 
   RPQR( 1, ElmQX, SaclibQX_i.ElmQX, &Q, &R ); 
   return SaclibQX( Q ); 
}    

inline SaclibQX SaclibQX::operator+( const SaclibQX& SaclibQX_i ) {  
   return SaclibQX( RPSUM( 1, ElmQX, SaclibQX_i.ElmQX ) ); 
}

inline SaclibQX SaclibQX::operator-( const SaclibQX& SaclibQX_i ) {  
   return SaclibQX( RPDIF( 1, ElmQX, SaclibQX_i.ElmQX ) ); 
}

inline SaclibQ SaclibQX::operator[] ( unsigned int index_i ) { 

   if ( index_i <= PDEG( ElmQX ) ) {
       return SaclibQ( LELTI( PCL( ElmQX ), PDEG( ElmQX ) - index_i + 1 ) );
   }
   else {
       MP_INT MPZero;
       mpz_init_set_ui( &MPZero,  0 );
       return SaclibQ( &MPZero ); 
   }
}

inline unsigned int SaclibQX::dg( ) { 
   return PDEG( ElmQX ); 
}

inline SaclibQ SaclibQX::eval( const SaclibQ* EvalPoint ) { 
   return SaclibQ( RPEMV( 1, ElmQX, EvalPoint->SaclibQElm ) ); 
}

inline SaclibQ SaclibQX::lcoeff( ) { 
   return SaclibQ( PLDCF( ElmQX ) ); 
}

inline void SaclibQX::prem( 
   const SaclibQX& SaclibQXA_i, 
   const SaclibQX& SaclibQXB_i,
   SaclibQX& SaclibQXR_o
   ) 
{  
   Word Q; 
   RPQR( 1, SaclibQXA_i.ElmQX , SaclibQXB_i.ElmQX, &Q, 
         &(SaclibQXR_o.ElmQX) ); 
}    

inline void SaclibQX::pquorem(
   const SaclibQX& SaclibQXA_i, 
   const SaclibQX& SaclibQXB_i,
   SaclibQX& SaclibQXQ_o,
   SaclibQX& SaclibQXR_o
   ) 
{  
   RPQR( 1, SaclibQXA_i.ElmQX , SaclibQXB_i.ElmQX, &(SaclibQXQ_o.ElmQX), 
         &(SaclibQXR_o.ElmQX) ); 
}

inline void SaclibQX::pgcd( const SaclibQX& SaclibQXA_i, 
const SaclibQX& SaclibQXB_i, SaclibQX& SaclibQXC_o ) { 
   Word IntPoly1, IntPoly2;
   Word Gcd, Coffactor1, Coffactor2;

   IntPoly1 = IPFRP( 1, SaclibQXA_i.ElmQX );
   IntPoly2 = IPFRP( 1, SaclibQXB_i.ElmQX );

   IPGCDC( 1, IntPoly1, IntPoly2, &Gcd, &Coffactor1, &Coffactor2 );

   SaclibQXC_o.ElmQX = RPFIP( 1, Gcd );
}

inline void SaclibQX::peea( 
   const SaclibQX& SaclibQXA_i, 
   const SaclibQX& SaclibQXB_i, 
   SaclibQX& SaclibQXG_i,
   SaclibQX& SaclibQXS_i, 
   SaclibQX& SaclibQXT_i
   )
{
   Word C,D,C1,C2,D1,D2;
   Word Z;
   Word Q,R,R1,R2;

   C = SaclibQXA_i.ElmQX;
   D = SaclibQXB_i.ElmQX;
   C1 = PMON( RNINT(1), 0 );  D1 = PMON( RNINT(0), 0);
   C2 = PMON( RNINT(0), 0 );  D2 = PMON( RNINT(1), 0);
   Z  = PMON( RNINT(0), 0 );

   while ( !(RPEQL( D, Z )) ) {
      RPQR( 1, C, D, &Q, &R ); 
      R1 = RPDIF(1,C1,RPPROD(1,Q,D1)); R2 = RPDIF(1,C2,RPPROD(1,Q,D2));
      C = D; C1 = D1; C2 = D2;
      D = R; D1 = R1; D2 = R2;
   }

   SaclibQXG_i.ElmQX = C;
   SaclibQXS_i.ElmQX = C1;
   SaclibQXT_i.ElmQX = C2;
}

inline void SaclibQX::factors( 
   const SaclibQX& SaclibQX_i, 
   int& Sign_o, 
   SaclibQ& Cont_o, 
   unsigned int& NumFactors_o, 
   unsigned int *Exps_o, 
   SaclibQX *Factors_o 
   ) 
{
   unsigned int i;
   Word IntPoly1;
   Word Sign, Cont, ListOfFactors;
  
   IntPoly1 = IPFRP( 1, SaclibQX_i.ElmQX );
   IUPFAC( IntPoly1, &Sign, &Cont, &ListOfFactors );
  
   Sign_o = Sign;
   Cont_o = SaclibQ( (long) Cont );
   NumFactors_o = LENGTH( ListOfFactors );
   for(i=1;i<=NumFactors_o;i++) {
      Exps_o[i-1] = FIRST( LELTI ( ListOfFactors, i ) );
      Factors_o[i-1].ElmQX = RPFIP( 1, SECOND( LELTI ( 
         ListOfFactors, i ) ) );
   }
}

inline SaclibQX SaclibQX::Diff( ) {
   return SaclibQX( RPDMV( 1, ElmQX ) );
}

/*-------------------------------------------------------------
  class: SaclibQXT
 
  date: 10/8/96
-------------------------------------------------------------*/
 
class SaclibQXT {
   private:
      Word ElmQXT;
   public:
      inline friend ostream& operator<<( ostream&, const SaclibQXT& );
 
      SaclibQXT( ) { }
      SaclibQXT( const long& );
      SaclibQXT( const Word& );
      SaclibQXT( const char* );
      SaclibQXT( const Word&, const unsigned int& );
      SaclibQXT( const SaclibQ&, const unsigned int&  );
      SaclibQXT( const SaclibQ& );
      SaclibQXT( const SaclibQX& );
      SaclibQXT( const SaclibQXT& );
      SaclibQXT( const long&, const int& );
      ~SaclibQXT( ) { }

      inline SaclibQXT& operator=( const SaclibQXT& );
      inline friend int operator==( const SaclibQXT&, const SaclibQXT& );
      inline friend int operator!=( const SaclibQXT&, const SaclibQXT& );
 
      inline SaclibQXT operator*( const SaclibQXT& );
      inline SaclibQXT operator/( const SaclibQXT& );
      inline SaclibQXT operator+( const SaclibQXT& );
      inline SaclibQXT operator-( const SaclibQXT& );
 
      inline SaclibQ operator[] ( unsigned int );
      inline unsigned int dg( );
      inline SaclibQ eval( const SaclibQ* );
      inline SaclibQXT Diff( );
};

inline ostream& operator<<( ostream& os, const SaclibQXT& SaclibQXT_i ) {
   RPWRITE( 1, SaclibQXT_i.ElmQXT, SACLIB_VariableName );
   return os;
}
 
SaclibQXT::SaclibQXT( const long& SaclibQElm_i ) {
   ElmQXT = PMON( RNINT( SaclibQElm_i ) ,0 );
}
 
SaclibQXT::SaclibQXT( 
   const Word& ElmQX_i // saclib rational polynomial
   ) 
{
   Word RatMult, ElmZX, TruncElmZX;
   if ( ElmQX_i == 0 ) {
      ElmQXT = ElmQX_i;
   }
   else 
   if ( PDEG( ElmQX_i < SACLIB_TruncDeg ) ) {
      ElmQXT = ElmQX_i;
   }
   else {
      IPSRP( 1, ElmQX_i, &RatMult, & ElmZX );
      TruncElmZX = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZX );
      ElmQXT = RPPROD( 1, RPFIP( 1, TruncElmZX ), PMON( RNINV(RatMult), 0 ) );
   }
}
 
SaclibQXT::SaclibQXT( const char* String_i ) {
   Word RatMult, ElmZX, TruncElmZX;
   int i = 0;
   StringToRatPoly( String_i, i, &ElmQXT );
   if ( PDEG( ElmQXT ) >= SACLIB_TruncDeg ) {
      IPSRP( 1, ElmQXT, &RatMult, & ElmZX );
      TruncElmZX = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZX ); 
      ElmQXT = RPPROD( 1, RPFIP( 1, TruncElmZX ), PMON( RNINV(RatMult), 0 ) );
   }
}
 
SaclibQXT::SaclibQXT( 
   const Word& ElmQ_i,         // saclib rational number
   const unsigned int& Deg_i 
   ) 
{
   if ( Deg_i < SACLIB_TruncDeg ) 
      ElmQXT = PMON( ElmQ_i, Deg_i );
   else
      ElmQXT = PMON( RNINT(0), 0 );
}
 
SaclibQXT::SaclibQXT( 
   const SaclibQ& SaclibQElm_i, 
   const unsigned int& Deg_i 
   ) 
{
   if ( Deg_i < SACLIB_TruncDeg )
      ElmQXT = PMON( SaclibQElm_i.SaclibQElm, Deg_i );
   else
      ElmQXT = PMON( RNINT(0), 0 );
}
 
SaclibQXT::SaclibQXT( const SaclibQ& SaclibQElm_i ) {
   ElmQXT = PMON( SaclibQElm_i.SaclibQElm, 0 );
}
 
SaclibQXT::SaclibQXT( const SaclibQX& SaclibQX_i ) {
   Word RatMult, ElmZX, TruncElmZX;
   if ( PDEG( SaclibQX_i.ElmQX ) < SACLIB_TruncDeg )
      ElmQXT = SaclibQX_i.ElmQX;
   else {
      IPSRP( 1, SaclibQX_i.ElmQX, &RatMult, &ElmZX );
      TruncElmZX = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZX );
      ElmQXT = RPPROD( 1, RPFIP( 1, TruncElmZX ), PMON( RNINV(RatMult), 0 ) );
   }
}
 
SaclibQXT::SaclibQXT( const SaclibQXT& SaclibQXT_i ) {
   Word RatMult, ElmZX, TruncElmZX;
   if ( PDEG( SaclibQXT_i.ElmQXT ) < SACLIB_TruncDeg )
      ElmQXT = SaclibQXT_i.ElmQXT;
   else {
     IPSRP( 1, SaclibQXT_i.ElmQXT, &RatMult, &ElmZX );
     TruncElmZX = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZX );
     ElmQXT = RPPROD( 1, RPFIP( 1, TruncElmZX ), PMON( RNINV(RatMult), 0 ) );
   } 
}

SaclibQXT::SaclibQXT( const long& pe, const int& Deg_i ) {
   if ( Deg_i < SACLIB_TruncDeg )
      ElmQXT = PMON( RNINT( pe ), Deg_i );
   else
      ElmQXT = PMON( RNINT(0), 0 );
}

inline SaclibQXT& SaclibQXT::operator=( const SaclibQXT& SaclibQXT_i ) {
   this->ElmQXT = SaclibQXT_i.ElmQXT;
   return *this;
}
 
inline int operator==( const SaclibQXT& SaclibQXTA_i,
const SaclibQXT& SaclibQXTB_i ) {
   return( RPEQL( SaclibQXTA_i.ElmQXT, SaclibQXTB_i.ElmQXT ) );
}
 
inline int operator!=( const SaclibQXT& SaclibQXTA_i,
const SaclibQXT& SaclibQXTB_i ) {
   return !( SaclibQXTA_i == SaclibQXTB_i );
}
 
inline SaclibQXT SaclibQXT::operator*( const SaclibQXT& SaclibQXT_i ) {
   Word RatMultA, ElmZXA, TruncElmZXA;  
   Word RatMultB, ElmZXB, TruncElmZXB;
   Word TruncProdZX, TruncProdQX, RatMult;

   // compute TruncElmZXA, TruncElmZXB
   IPSRP( 1, ElmQXT, &RatMultA, &ElmZXA );
   TruncElmZXA = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZXA );
   IPSRP( 1, SaclibQXT_i.ElmQXT, &RatMultB, &ElmZXB );
   TruncElmZXB = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZXB );

   // compute the truncated product
   TruncProdZX = IUPTPR( SACLIB_TruncDeg, TruncElmZXA, TruncElmZXB ); 
  
   // convert to rational form, and return
   if ( TruncProdZX == 0 ) { 
      TruncProdQX = PMON( RNINT( 0 ), 0 );
   }
   else {
      RatMult = RNINV( RNPROD( RatMultA, RatMultB ) );
      TruncProdQX = RPPROD( 1, RPFIP( 1, TruncProdZX) , PMON( RatMult, 0 ) );
   }
 
   return SaclibQXT( TruncProdQX );
}
 
inline SaclibQXT SaclibQXT::operator/( 
   const SaclibQXT& SaclibQXT_i 
   ) 
{
    int i,j;
    int deg,bindex;
    Word b;
    Word TProd;
    Word ConstCoeff;
    Word BCoeff;
    Word ConstPoly;
    Word CoeffList;
    Word Coeff1,Coeff2;
  
    deg        = PDEG( SaclibQXT_i.ElmQXT );
    CoeffList  = PCL( SaclibQXT_i.ElmQXT );
    ConstCoeff = LELTI( CoeffList, deg+1 );

    if ( ConstCoeff == 0 ) {
       cerr << endl;
       cerr << (*this) << " / " << SaclibQXT_i << endl;
       OutputErrorExit( BBERROR_SACLIB,
         "SaclibQXT::operator/ : Can not compute inverse" );
    }

    ConstCoeff = RNINV( ConstCoeff);
    b          = PMON( ConstCoeff, 0 );  

    for( i=1; i<SACLIB_TruncDeg; i++ ) {
       TProd = RNINT(0);
       for( j=1; j<=i; j++ ) {
          bindex = deg-j+1;
          if ( bindex > 0 ) 
             Coeff1 = LELTI( CoeffList, bindex );
          else
             Coeff1 = RNINT(0);
          bindex = PDEG(b) -( i-j) + 1;
          if ( bindex > 0 ) 
             Coeff2 = LELTI( PCL( b ), bindex);
          else
             Coeff2 = RNINT(0);
          if ( ( Coeff1 != 0 ) && ( Coeff2 != 0 ) ) 
             TProd  = RNSUM( TProd, RNPROD( Coeff1, Coeff2 ) ); 
       }
          if ( TProd != 0 )  
             BCoeff = RNPROD( RNINT(-1), RNPROD( ConstCoeff, TProd ));
          else
             BCoeff = RNINT(0);
          b      = RPSUM( 1, b, PMON( BCoeff , i ) );
    }
    return( (*this) * SaclibQXT( b ) );
}
 
inline SaclibQXT SaclibQXT::operator+( const SaclibQXT& SaclibQXT_i ) {
   Word RatMultA, ElmZXA, TruncElmZXA, TruncElmQXA;
   Word RatMultB, ElmZXB, TruncElmZXB, TruncElmQXB;
   Word TruncSumQX;
 
   if ( ( (*this).ElmQXT == 0 ) && ( SaclibQXT_i.ElmQXT == 0 ) ) {
      TruncSumQX = PMON( RNINT( 0 ), 0 );
   }
   else
   if ( (*this).ElmQXT == 0 ) {
      TruncSumQX = SaclibQXT_i.ElmQXT;
   }
   else
   if ( SaclibQXT_i.ElmQXT == 0 ) {
      TruncSumQX = (*this).ElmQXT;
   }
   else {
   // compute TruncElmZXA, TruncElmZXB, TruncElmQXA, TruncElmQXB
   IPSRP( 1, ElmQXT, &RatMultA, &ElmZXA );
   TruncElmZXA = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZXA );
   TruncElmQXA = RPPROD( 1, RPFIP( 1, TruncElmZXA) , 
                         PMON( RNINV( RatMultA ), 0 ) );
   IPSRP( 1, SaclibQXT_i.ElmQXT, &RatMultB, &ElmZXB );
   TruncElmZXB = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZXB );
   TruncElmQXB = RPPROD( 1, RPFIP( 1, TruncElmZXB) , 
                         PMON( RNINV( RatMultB ), 0 ) ); 

   // sum and return
   TruncSumQX = RPSUM( 1, TruncElmQXA, TruncElmQXB );      
   }

   return SaclibQXT( TruncSumQX );
}
 
inline SaclibQXT SaclibQXT::operator-( const SaclibQXT& SaclibQXT_i ) {
   Word RatMultA, ElmZXA, TruncElmZXA, TruncElmQXA;
   Word RatMultB, ElmZXB, TruncElmZXB, TruncElmQXB;
   Word TruncDiffQX;
 
   if ( ( (*this).ElmQXT == 0 ) && ( SaclibQXT_i.ElmQXT == 0 ) ) {
      TruncDiffQX = PMON( RNINT( 0 ), 0 );
   }
   else
   if ( (*this).ElmQXT == 0 ) {
      TruncDiffQX = RPNEG( 1, SaclibQXT_i.ElmQXT );
   }
   else
   if ( SaclibQXT_i.ElmQXT == 0 ) {
      TruncDiffQX = (*this).ElmQXT;
   }
   else {
   // compute TruncElmZXA, TruncElmZXB, TruncElmQXA, TruncElmQXB
   IPSRP( 1, ElmQXT, &RatMultA, &ElmZXA );
   TruncElmZXA = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZXA );
   TruncElmQXA = RPPROD( 1, RPFIP( 1, TruncElmZXA) ,
                         PMON( RNINV( RatMultA ), 0 ) );
   IPSRP( 1, SaclibQXT_i.ElmQXT, &RatMultB, &ElmZXB );
   TruncElmZXB = IPTRUN( 1, LIST1(SACLIB_TruncDeg), ElmZXB );
   TruncElmQXB = RPPROD( 1, RPFIP( 1, TruncElmZXB) ,
                         PMON( RNINV( RatMultB ), 0 ) );
 
   // subtract and return
   TruncDiffQX = RPDIF( 1, TruncElmQXA, TruncElmQXB );
   }
 
   return SaclibQXT( TruncDiffQX );
}
 
inline SaclibQ SaclibQXT::operator[] ( unsigned int index_i ) {
   if ( index_i <=  PDEG( ElmQXT ) ) {
      return SaclibQ( LELTI( PCL( ElmQXT ), PDEG( ElmQXT ) - index_i + 1 ) );
   }
   else {
       MP_INT MPZero;
       mpz_init_set_ui( &MPZero,  0 );
       return SaclibQ( &MPZero );
   }
}
 
inline unsigned int SaclibQXT::dg( ) {
   return PDEG( ElmQXT );
}

inline SaclibQ SaclibQXT::eval( const SaclibQ* EvalPoint ) {
   return SaclibQ( RPEMV( 1, ElmQXT, EvalPoint->SaclibQElm ) );
}

inline SaclibQXT SaclibQXT::Diff( ) {
   return SaclibQXT( RPDMV( 1, ElmQXT ) );
}

/*-------------------------------------------------------------
  class: SaclibZP

  date: 10/4/96
-------------------------------------------------------------*/

class SaclibZP {
   private:
      Word SaclibZPElm;
   public:
      friend SaclibZPX;
      inline friend ostream& operator<<(ostream& os, const SaclibZP& SaclibZP_i);

      SaclibZP()                        { }
      SaclibZP(const Word& SaclibZPElm_i)  { SaclibZPElm = SaclibZPElm_i; }
      SaclibZP(const long& SaclibZPElm_i)  
         { SaclibZPElm = MIHOM(SACLIB_SLCard,SaclibZPElm_i); }
      SaclibZP(const MP_INT*);
      SaclibZP(const char* );
      SaclibZP(const SaclibZP& SaclibZP_i) { SaclibZPElm=SaclibZP_i.SaclibZPElm; }
      ~SaclibZP()                       { }

      inline SaclibZP& operator=( const SaclibZP& );
      inline friend int operator==( const SaclibZP&, const SaclibZP& );
      inline friend int operator!=( const SaclibZP&, const SaclibZP& );

      inline SaclibZP operator*(const SaclibZP&);
      inline SaclibZP operator/(const SaclibZP&);
      inline SaclibZP operator+(const SaclibZP&);
      inline SaclibZP operator-(const SaclibZP&);

      inline SaclibZP DistinctSelect( );
      inline MP_INT card( )          { return SACLIB_MPCard; }
      inline SaclibZP Exp( int );
      inline SaclibZP Denom( );
      inline SaclibZP LCM( SaclibZP*, unsigned int );
      inline void toString(char* Pntr_m) {IntegerToString(SaclibZPElm,0,Pntr_m);}
      inline void NumtoString(char*);
      inline void DenomtoString(char*);

      class RandomIterator {
         public:
            RandomIterator(const unsigned int, const MP_INT*);
            ~RandomIterator()        { mpz_clear(&MPCard); }
            SaclibZP operator()();
         private:
            long RandStateArray[32];
            unsigned int RandSeed;
            MP_INT MPCard;
            Word SLCard;
      };

      inline Word Native() { return SaclibZPElm; }
};

SaclibZP::SaclibZP( const MP_INT *MPInt_i ) {
   SaclibZPElm = MIHOM( SACLIB_SLCard,ConvertMPToSacInt( MPInt_i ) );
}

SaclibZP::SaclibZP(const char* String_i) {
   int i = 0;
   SaclibZPElm = StringToInteger( String_i, i );
}

inline ostream& operator<<(ostream& os, const SaclibZP& SaclibZP_i) {
   IWRITE( SaclibZP_i.SaclibZPElm );
   return os;
}

inline SaclibZP& SaclibZP::operator=(const SaclibZP& SaclibZP_i) {
   this->SaclibZPElm = SaclibZP_i.SaclibZPElm;
   return *this;
}

inline int operator==(const SaclibZP& SaclibZPA_i, const SaclibZP& SaclibZPB_i) {
   if ( ICOMP( SaclibZPA_i.SaclibZPElm, SaclibZPB_i.SaclibZPElm ) == 0 )
      return 1;
   else
      return 0;
}

inline int operator!=(const SaclibZP& SaclibZPA_i, const SaclibZP&  SaclibZPB_i) {
   if ( ICOMP( SaclibZPA_i.SaclibZPElm, SaclibZPB_i.SaclibZPElm) == 0 )
      return 0;
   else
      return 1;
}

inline SaclibZP SaclibZP::operator*(const SaclibZP& SaclibZP_i) {
   return SaclibZP( MIPROD( SACLIB_SLCard, this->SaclibZPElm, 
                    SaclibZP_i.SaclibZPElm ) );
}

inline SaclibZP SaclibZP::operator/(const SaclibZP& SaclibZP_i) {
   return SaclibZP( MIQ( SACLIB_SLCard, this->SaclibZPElm, 
                    SaclibZP_i.SaclibZPElm ) );
}

inline SaclibZP SaclibZP::operator+(const SaclibZP& SaclibZP_i) {
   return SaclibZP( MISUM( SACLIB_SLCard, this->SaclibZPElm, 
                    SaclibZP_i.SaclibZPElm ) );
}

inline SaclibZP SaclibZP::operator-(const SaclibZP& SaclibZP_i) {
   return SaclibZP( MIDIF( SACLIB_SLCard, this->SaclibZPElm, 
                    SaclibZP_i.SaclibZPElm ) );
}

inline SaclibZP SaclibZP::DistinctSelect() {
   Word One        = 1;
   SaclibZP NextSaclibZP = *this;

   this->SaclibZPElm  = MISUM( SACLIB_SLCard, this->SaclibZPElm, One );
   return NextSaclibZP;
}

inline SaclibZP SaclibZP::Exp( int n_i ) {
    int i;
    int Sign;
    Word Prod;

    if ( n_i == 0 ) {
       Prod = 1;
       return SaclibZP( Prod );
    }
    else
       Prod = SaclibZPElm;

    if ( n_i < 0 ) {
       Sign = -1;
       n_i = n_i * -1;
    }
    else
       Sign = 1;

    for(i=1;i<n_i;i++)
       Prod = MIPROD( SACLIB_SLCard, Prod, SaclibZPElm );

    if ( Sign == 1 )
       return SaclibZP( Prod );
    else
       return SaclibZP( MIINV( SACLIB_SLCard, Prod ) );
}

inline SaclibZP SaclibZP::Denom( ) {
   return SaclibZP( (Word)1 );
}

inline SaclibZP SaclibZP::LCM(
   SaclibZP* SaclibZPElmArray_i,
   unsigned int NumElms_i
   )
{
   return SaclibZP( (Word)1 );
}

inline void SaclibZP::NumtoString(
   char* Pntr_m
   )
{
 Word Num;
 Num = SaclibZPElm;
 IntegerToString( Num, 0, Pntr_m );
}

inline void SaclibZP::DenomtoString(
   char* Pntr_m
   )
{
 Word Den;
 Den = 1;
 IntegerToString( Den, 0, Pntr_m );
}

SaclibZP::RandomIterator::RandomIterator(const unsigned int Seed_i,
const MP_INT* Card_i) : RandSeed(Seed_i) {
   if ( mpz_cmp_ui( Card_i, 0 ) < 0 )
      OutputErrorExit( BBERROR_SACLIB,
         "SaclibZP::RandomIterator : input card < 0" );

   mpz_init_set( &MPCard, Card_i );
   SLCard = ConvertMPToSacInt( &MPCard );

   for(int i=0;i<32;i++)
      RandStateArray[i] = (long)i;
   initstate( Seed_i, (char*) RandStateArray, sizeof(RandStateArray) );
}

SaclibZP SaclibZP::RandomIterator::operator()() {
   long r;
   Word rr;
   Word c,x;
   char *PrevState;

   x = 0; c = 1;

   PrevState = (char*)setstate((char*)RandStateArray);

   while ( ICOMP(c,SLCard) == -1 ) {
      r = random()-1;           // random long [0,2147483647)
      rr = ConvertLongToSacInt(r);
      x = IPROD(x,SACLIB_SLMaxRand);
      x = ISUM(x,rr);
      c = IPROD(c,SACLIB_SLMaxRand);
   }

   x = IREM(x,SLCard);
   setstate(PrevState);
   return(SaclibZP( (Word)x ));
}

/*-------------------------------------------------------------
  class: SaclibZPX

  date: 10/4/96
-------------------------------------------------------------*/

class SaclibZPX {
   private:
      Word ElmZPX;
   public:
      inline friend ostream& operator<<( ostream&, const SaclibZPX& );

      SaclibZPX( ) { }
      SaclibZPX( const long& );
      SaclibZPX( const Word& );
      // MUST ADD STRING TO ZPX
      SaclibZPX( const char* );
      SaclibZPX( const Word&, const unsigned int& );
      SaclibZPX( const SaclibZP&, const unsigned int&  );
      SaclibZPX( const SaclibZP& );
      SaclibZPX( const SaclibZPX& );
      SaclibZPX( const long&, const int& );
      ~SaclibZPX( ) { }

      inline SaclibZPX& operator=( const SaclibZPX& );
      inline friend int operator==( const SaclibZPX&, const SaclibZPX& );
      inline friend int operator!=( const SaclibZPX&, const SaclibZPX& );

      inline SaclibZPX operator*( const SaclibZPX& );
      inline SaclibZPX operator/( const SaclibZPX& );
      inline SaclibZPX operator+( const SaclibZPX& );
      inline SaclibZPX operator-( const SaclibZPX& );

      inline SaclibZP operator[] ( unsigned int );
      inline unsigned int dg( );
      inline SaclibZP eval( const SaclibZP* );
      inline SaclibZP lcoeff( );
      inline void prem( const SaclibZPX&, const SaclibZPX&, SaclibZPX& );
      inline void pquorem( const SaclibZPX&, const SaclibZPX&, SaclibZPX&, SaclibZPX
&);
      inline void pgcd( const SaclibZPX&, const SaclibZPX&, SaclibZPX& );
      inline void peea( const SaclibZPX&, const SaclibZPX&, SaclibZPX&,
                        SaclibZPX&, SaclibZPX& );
      inline void factors( const SaclibZPX&, int&, SaclibZP&, unsigned int&,
         unsigned int*, SaclibZPX* );
      inline SaclibZPX Diff( );
      // MUST ADD ZPX TO STRING
      inline void toString( char* Pntr_m ) {  }
};

inline ostream& operator<<( ostream& os, const SaclibZPX& SaclibZPX_i ) {
   IPWRITE( 1, SaclibZPX_i.ElmZPX, SACLIB_VariableName );
   return os;
}

SaclibZPX::SaclibZPX( const long& SaclibZPElm_i ) {
   ElmZPX = PMON( MIHOM( SACLIB_SLCard, SaclibZPElm_i ) ,0 );
}

SaclibZPX::SaclibZPX( const Word& ElmZPX_i ) {
   ElmZPX = ElmZPX_i;
}

SaclibZPX::SaclibZPX( const char* String_i ) {
}

SaclibZPX::SaclibZPX( const Word& ElmZPX_i, const unsigned int& Deg_i ) {
   ElmZPX = PMON( ElmZPX_i, Deg_i );
}

SaclibZPX::SaclibZPX( const SaclibZP& SaclibZPElm_i, const unsigned int& Deg_i ) {
   ElmZPX = PMON( SaclibZPElm_i.SaclibZPElm, Deg_i );
}

SaclibZPX::SaclibZPX( const SaclibZP& SaclibZPElm_i ) {
   ElmZPX = PMON( SaclibZPElm_i.SaclibZPElm, 0 );
}

SaclibZPX::SaclibZPX( const SaclibZPX& SaclibZPX_i ) {
   ElmZPX = SaclibZPX_i.ElmZPX;
}

SaclibZPX::SaclibZPX( const long& pe, const int& Deg_i ) {
   ElmZPX = PMON( MIHOM( SACLIB_SLCard, pe ), Deg_i );
}

inline SaclibZPX& SaclibZPX::operator=( const SaclibZPX& SaclibZPX_i ) {
   this->ElmZPX = SaclibZPX_i.ElmZPX;
   return *this;
}

inline int IPEQL(
   Word IPolyA_i,
   Word IPolyB_i
   )
{
   int deg = PDEG( IPolyA_i );

   if ( deg != PDEG( IPolyB_i ) )
      return BBDEFS_FALSE;

   // handle the zero cases
   if ( ( IPolyA_i == 0 ) && ( IPolyB_i == 0 ) ) return  BBDEFS_TRUE;
   if ( ( IPolyA_i == 0 ) && ( IPolyB_i != 0 ) ) return  BBDEFS_FALSE;
   if ( ( IPolyA_i != 0 ) && ( IPolyB_i == 0 ) ) return  BBDEFS_FALSE;

   Word c1,c2;
   c1 = PCL( IPolyA_i );
   c2 = PCL( IPolyB_i );

   for(int i=1;i<=deg+1;i++)
      if ( ICOMP(LELTI(c1,i),LELTI(c2,i)) < 0 ) return  BBDEFS_FALSE;

   return  BBDEFS_TRUE;
}

inline int operator==( const SaclibZPX& SaclibZPXA_i,
const SaclibZPX& SaclibZPXB_i ) {
   return( IPEQL( SaclibZPXA_i.ElmZPX, SaclibZPXB_i.ElmZPX ) );
}

inline int operator!=( const SaclibZPX& SaclibZPXA_i,
const SaclibZPX& SaclibZPXB_i ) {
   return !( SaclibZPXA_i == SaclibZPXB_i );
}

inline SaclibZPX SaclibZPX::operator*( const SaclibZPX& SaclibZPX_i ) {
   return SaclibZPX( MIPPR( 1, SACLIB_SLCard, ElmZPX, SaclibZPX_i.ElmZPX ) );
}

inline SaclibZPX SaclibZPX::operator/( const SaclibZPX& SaclibZPX_i ) {
   Word Q,R;
   MIUPQR( SACLIB_SLCard, ElmZPX, SaclibZPX_i.ElmZPX, &Q, &R );
   return SaclibZPX( Q );
}

inline SaclibZPX SaclibZPX::operator+( const SaclibZPX& SaclibZPX_i ) {
   return SaclibZPX( MIPSUM( 1, SACLIB_SLCard, ElmZPX, SaclibZPX_i.ElmZPX ) );
}

inline SaclibZPX SaclibZPX::operator-( const SaclibZPX& SaclibZPX_i ) {
   return SaclibZPX( MIPDIF( 1, SACLIB_SLCard, ElmZPX, SaclibZPX_i.ElmZPX ) );
}

inline SaclibZP SaclibZPX::operator[] ( unsigned int index_i ) {
   if ( index_i <= PDEG( ElmZPX ) ) {
      return SaclibZP( LELTI( PCL( ElmZPX ), PDEG( ElmZPX ) - index_i + 1 ) );
   }
   else {
       MP_INT MPZero;
       mpz_init_set_ui( &MPZero,  0 );
       return SaclibZP( &MPZero );
   }
}

inline unsigned int SaclibZPX::dg( ) {
   return PDEG( ElmZPX );
}

inline SaclibZP SaclibZPX::eval( const SaclibZP* EvalPoint ) {
   Word A;
   A = IPEMV( 1, ElmZPX, EvalPoint->SaclibZPElm );
   A = MIHOM( SACLIB_SLCard, A );
   // MPEMV( 1, SACLIB_SLCard, ElmZPX, EvalPoint->SaclibZPElm );
   return SaclibZP( A );
}

inline SaclibZP SaclibZPX::lcoeff( ) {
   return SaclibZP( PLDCF( ElmZPX ) );
}

inline void SaclibZPX::prem(
   const SaclibZPX& SaclibZPXA_i,
   const SaclibZPX& SaclibZPXB_i,
   SaclibZPX& SaclibZPXR_o
   )
{
   Word Q;
   MIUPQR( SACLIB_SLCard, SaclibZPXA_i.ElmZPX , SaclibZPXB_i.ElmZPX, &Q,
           &(SaclibZPXR_o.ElmZPX) );
}

inline void SaclibZPX::pquorem(
   const SaclibZPX& SaclibZPXA_i,
   const SaclibZPX& SaclibZPXB_i,
   SaclibZPX& SaclibZPXQ_o,
   SaclibZPX& SaclibZPXR_o
   )
{
   MIUPQR( SACLIB_SLCard, SaclibZPXA_i.ElmZPX , SaclibZPXB_i.ElmZPX, 
           &(SaclibZPXQ_o.ElmZPX), &(SaclibZPXR_o.ElmZPX) );
}

inline void SaclibZPX::pgcd( const SaclibZPX& SaclibZPXA_i,
const SaclibZPX& SaclibZPXB_i, SaclibZPX& SaclibZPXC_o ) {
   SaclibZPXC_o.ElmZPX = MUPGCD( SACLIB_SLCard, SaclibZPXA_i.ElmZPX, 
                                  SaclibZPXB_i.ElmZPX );
}

inline void SaclibZPX::peea(
   const SaclibZPX& SaclibZPXA_i,
   const SaclibZPX& SaclibZPXB_i,
   SaclibZPX& SaclibZPXG_m,
   SaclibZPX& SaclibZPXS_m,
   SaclibZPX& SaclibZPXT_m
   )
{
// MUST ADD EEA
//   MUPEGC( SACLIB_SLCard, SaclibZPXA_i, SaclibZPXB_i, &SaclibZPXG_m,
//           &SaclibZPXS_m, &SaclibZPXT_m );
}

inline void SaclibZPX::factors(
   const SaclibZPX& SaclibZPX_i,
   int& Sign_o,
   SaclibZP& Cont_o,
   unsigned int& NumFactors_o,
   unsigned int *Exps_o,
   SaclibZPX *Factors_o
   )
{
// MUST ADD FACTORIZATION
}

inline SaclibZPX SaclibZPX::Diff( ) {
   return SaclibZPX( MUPDER( SACLIB_SLCard, ElmZPX ) );
}

#endif
