;;; -*- Mode: LISP; Syntax: Zetalisp; Package: STRAIGHT; Base: 10 -*-

;; $STRAIGHTFACTOR is an implementation of the algorithm described in
;; (Kaltofen...).  Many of the variable names in the following code are
;; are the same as in the paper.  Two major modifications were made to the
;; original algorithm:
;;
;;  (1) The input polynomial is mapped to a univariate polynomial instead of
;;      a bivariate polynomial.
;;  (2) The original algorithm calls for computing the partial fraction
;;      expansion of the reciprocal of the product of all the univariate
;;      factors.  We do that here but also compute the partial fraction
;;      expansions of all x**i/(product of univ. factors), i=0,...,d-1, where
;;      d is the total degree of the input polynomial.  This makes for a
;;      very easy last step in each lifting step.
;;
;; Input:
;;   program:  a straight-line program
;;   cbound:   a bound on the numerators and comman denominator of the
;;             polynomial given by program
;;   dbound:   a bound on the total degree of the polynomial given by program
;;
;; Variables:
;;   fbar:     A translated form of the input program
;;   varlist:  A list of triples (xi,ai,bi) representing the translation in
;;             fbar as follows
;;                        x1 <-- x1 + a1
;;                        xi <-- xi + bi*x1 + ai    i=2,3,...
;;             The xi are straight-line input instructions.
;;   mainvar:  The Macsyma symbol for the main variable
;;   numbound: A bound used to convert the intermediate univariate polynomial
;;             to sparse representation.
;;   glist:    A list of Macsyma-format univariate factors.
;;   elist:    The list of multiplicities corresponding to the factors in glist.
;;   fbarcoeffs:   A list of straight-line instructions representing the
;;                 coefficients of fbar in the main variable.
;;   targetvalues: This represents an evaluation of fbar and is used to check
;;                 if a list of factors is complete.
;;
;;
;; When the global variable $factormonitor is non-nil, intermediate results are
;; converted into Macsyma-readable format and saved in several global variables
;; which can then be referenced from within  Macsyma.  The monitor should only
;; be used on very small problems.
;;
(defun $straightfactor (program cbound &optional dbound)
  (checkstraight program)
  (verbdo (princ "Beginning factorization.") (terpr)
	  (princ "Factor Monitor ")
	  (princ (if $factormonitor then "ON" else "OFF"))
	  (terpr))
  (if (null dbound) then
      (setq dbound ($straightdegree program))
      (verbdo (princ "Computed Total Degree is ") (princ dbound) (terpr)))

  (let* ( (fbar       (translate-for-fact program))
	  (varlist    (loop for xa in (straight-get-all-inputs
					          (straightlist fbar))
			    collect (cons xa (straightvalue xa))))
	  (mainvar    (straightarg1 (caar varlist)))
	  (macfactors (univariatefactor fbar varlist dbound cbound))
	  (numbound       (car    macfactors))
	  (constantfactor (cadr   macfactors))
	  (glist          (caddr  macfactors))
	  (elist          (cadddr macfactors))
	  fbarcoeffs
	  fbarcoeff-program
          targetvalues )

    (if (equal elist '(1)) then
	(princ "Polynomial is Irreducible.") (terpr)
	program
     else
	(verbdo (princ "Translation constants: ")
		(princ (loop for triple in varlist collect (cdr triple)))
		(terpr) (princ "Univariate factors: ")
		($print `((mtimes simp) . ,glist))
		(terpr) (princ "Exponents of univariate factors: ")
		(princ elist) (terpr)
		(princ "Bound: ") (princ numbound) (terpr))
	(if constantfactor then
	    (verbdo (princ "Constant factor: ") (princ constantfactor) (terpr))
	    (setq fbar (simulation-principle fbar
			   (sdiv (straightlast fbar) (sscalar constantfactor))
			   (currentprogram))))
	(setq targetvalues (target-evaluation fbar varlist))
	;; Get the coefficients of fbar in mainvar.  $straightcoeff returns a
	;; Macsyma list of straight-line programs.
	(let ( (coeff-proglist (cdr ($straightcoeff fbar mainvar))) )
	    (setq fbarcoeff-program (car coeff-proglist)
		  ;; pluck the poly-instructions from the programs...
		  fbarcoeffs (loop for prog in (butlast coeff-proglist)
				   collect (straightlast prog))))
	;; The input instructions in varlist are the inputs of the fbar program,
	;; and these need to be changed to the corresponding inputs in
	;; fbarcoeff-program.  Also, need to set the values of the fbarcoeff
	;; inputs to the values of the fbar inputs (for use in the routine
	;; Factors-Complete).
	(loop for triple in (cdr varlist)
	      for value  in (cdr (third targetvalues))
	      for inst in (straight-get-all-inputs
			          (straightlist fbarcoeff-program))
	      do (setvalue inst value)
		 (setf (car triple) inst))

	(simulation-principle fbarcoeff-program
	  (setq glist
		(hensel-lift
		    glist elist fbarcoeffs mainvar targetvalues
		    dbound numbound))
	  (verbdo (princ "Lifting Complete.") (terpr))
	  (if $factormonitor then (savestate3 varlist))
	  (reverse-translate-for-fact varlist mainvar)
	  ;; Reverse-translate-for-fact always puts the main input third in
	  ;; the list...
	  (setq mainvar (third (currentlist)))

	  `((mlist simp)
	     . ,(loop for gi in (build-factors glist mainvar)
		      for ei in elist
		      collect
			`((mlist simp)
			   ,(list (straightheader) (currentlist) gi)
			   ,ei))) ))))

;; HENSEL-LIFT receives a list of univariate factor images and returns a list 
;; of fully lifted multivariate factors.
;;
;; Variables:
;;
;;   glist:    The list of factors we are lifting.  Each factor in the list
;;             is represented as a list of homogeneous parts, and each
;;             homogeneous part is represented as a polynomial in the main
;;             variable.
;;   gproducts:  After each lifting step, the last element in this list is
;;             the product of all factors powered to their corresponding
;;             multiplicities.  The other elements of the list hold
;;             intermediate expressions in the product computation that are
;;             used by the routine NEWGPRODUCTS to avoid re-computing the same
;;             values several times. 
;;   rhodenominator:  If all factors have multiplicity 1 then this is nil,
;;             otherwise this is the product of each univariate factor raised
;;             to the power one less its multiplicity.
;;   partial-fractions:  A list of the partial fraction components for each
;;             factor, where the components are univariate polynomials (i.e.,
;;             this is a list of lists of polynomials).
;;   gvalues:  Keeps track of the value of each factor at the target
;;             evaluation point at the end of each lifting step. See the
;;             routine NEWGVALUES.
;;   max-lift-steps:  The maximum total degree of the factors.  This is used
;;             as a fail-safe stopping mechanism, and if it does come into play
;;             you can bet your answers are wrong.
;;
(defun hensel-lift (macglist elist fbarcoeffs mainvar targetvalues
		    dbound numbound)
  (let* ( (lastcoeff (currentlast))
	  (glist     (loop for gi in macglist collect
		         (list (macsyma-to-straight-coeffs gi mainvar))))
	  (gproducts (loop for (gi0) in glist
			   for ei in elist
			   for repeat-num first (1- ei) then ei
			   with product = (caar glist)
			   nconc
			     (loop repeat repeat-num collect
				 `( (,(sscalar 0))
				    ,(setq product (spolymult product gi0))))))
	  (rhodenominator    (getrhodenominator macglist elist mainvar))
	  (partial-fractions (getpartfracs macglist glist mainvar dbound))
	  ;;(partial-fractions (getpartfracs glist dbound numbound))
	  (gvalues   (newgvalues glist (loop repeat (length glist) collect 0)
				 targetvalues (currentlist)))
	  (max-lift-steps    (loop for (gi0) in glist
				   maximize (- (length gi0) 2))) )

      (if $factormonitor then
	  (setq $glist nil $gproducts nil $taulist nil)
	  (savestate1 partial-fractions)
	  (savestate2 glist gproducts nil))

      (verbdo (princ "Program length = ")
	      (princ (length (currentlist))) (terpr)
              (princ "Beginning Lifting.") (terpr))
      (loop for k = 0 then (1+ k)
	    for progtail = (cdr $straightprogram)
	    for tau first (incremental-coeffs fbarcoeffs lastcoeff 'initialize)
	            then  (progn (setq gproducts
				       (new-gproducts glist elist gproducts))
				 (spolysub
				     (incremental-coeffs fbarcoeffs lastcoeff)
				     (car (car (last gproducts)))))
	    do
	      (setq glist
		    (lift-glist glist elist partial-fractions
				(if rhodenominator then
				    (spolydiv tau rhodenominator)
				 else  ; No multiplicities
				    tau)))
	      (setq gvalues (newgvalues glist gvalues targetvalues progtail))

	      (if $factormonitor then
		  (savestate2 glist gproducts tau))

	      (verbdo (princ "   Parts of degree ") (princ (1+ k))
		      (princ " lifted.") (princ "   ")
		      (princ (length progtail)) (terpr))
	    until
	      (or (factors-complete gvalues elist targetvalues)
		  (if (= k max-lift-steps) then
		      (princ "FACTORS-COMPLETE missed ... "))) )
      glist )) 

;; Here we add one more homogeneous part to each factor.  The new part is 
;; computed by just taking the dot product of the corresponding partial 
;; fraction list and rho, where the elements in the each partial fraction list
;; are polynomials and the elements of rho are scalar coefficients.
(defun lift-glist (glist elist partial-fractions rho)
  (loop for gi in glist
	for ei in elist
	for partfrac in partial-fractions
	collect
	  (cons (loop for poly in partfrac
		      for rho-coeff in rho
		      with sum = (list (sscalar 0)) do
		    (setq sum
			  (spolyadd sum
				   ;; Multiply the polynomial by the scalar
				   (loop for coeff in poly
					     collect (stimes coeff rho-coeff))))
		      finally
			(return
			  (if (= ei 1) then
			      sum
			   else  ;divide each coefficient by ei
			      (loop for xa in sum
				    with ei-inverse
				            = (sscalar `((rat simp) 1 ,ei))
				    collect (stimes xa ei-inverse)))))
		gi) ))
	

;; Evaluate fbar at a random point for use by FACTORS-COMPLETE.
(defun target-evaluation (fbar-program varlist)
  (let* ( (modulus (fixrandomprime))
	  (inputvalues (loop for triple in varlist
			     for random-int = (modularize-int (random) modulus)
			     do  (setvalue (car triple) random-int)
			     collect random-int)) )
    (simulation-principle fbar-program
      (list (evalmodp (straightlast fbar-program) modulus nil)
	    modulus
	    inputvalues) )))

;; Check if finished by evaluating, at the target evaluation point, the product
;; of the gi powered to their multiplicities and checking if the result matches
;; the value of fbar at the same point.
(defun factors-complete (gvalues elist targetvalues)
  (let ( (target  (car targetvalues))
	 (modulus (cadr targetvalues)) )
    (equal target
	   (loop for gval in gvalues
		 for ei in elist
		 with product = 1
		 do
	           (setq product
			 (timesmod product (expmod gval ei modulus) modulus))
		 finally (return product))) ))

;; Compute the value of each gi at the target evaluation point.  The values of
;; the gi after the previous lifting step have been saved in gvalues, so just
;; need to add the value of the new homogeneous part.
(defun newgvalues (glist gvalues targetvalues progtail)
  (let ( (modulus   (cadr targetvalues))
	 (mainvalue (car (third targetvalues))) )
    (evalmodp (currentlast) modulus nil progtail)
    (loop for gi in glist
	  for oldgvalue in gvalues
	  collect
	    (addmod oldgvalue
		    ;; Get value of new homogeneous part using Horner's rule.
		    (loop for cikm in (reverse (car gi))
        		  with polyvalue = 0
			  do (setq polyvalue
				   (addmod
				       (timesmod polyvalue mainvalue modulus)
				       (straightvalue cikm)   modulus))
			    finally (return polyvalue))
		    modulus)) ))

;; Each factor is a list of homogeneous parts, and each homogeneous part is
;; a polynomial in the main variable.  This routine puts the pieces together.
(defun build-factors (glist mainvar)
  (loop for gi in glist
	collect
	  ;; Build factor using Horner's rule.
	  (loop with poly = (sscalar 0)
		for xa in (rearrange-gi gi)
		for coeff = (loop for c in xa
				  with sum = (sscalar 0)
				    do (setq sum (sadd sum c))
				  finally (return sum))
		do (setq poly (sadd (stimes mainvar poly) coeff))
		finally (setdegvalue poly
				     (1- (length (car (last gi)))))
			(return poly))) )

;; Group coefficients of like powers of the main variable.
(defun rearrange-gi (gi)
  (reverse
      (loop repeat (loop for poly in gi maximize (length poly))
	    for pointers = gi then (loop for xa in pointers collect (cdr xa))
	    collect
	      (loop for xa in pointers
		    when xa collect (car xa)))) )


;; Compute the product of the univariate factors powered to one less the
;; corresponding multiplicities.
;; If no factor has multiplicity greater than 1 then return nil.
(defun getrhodenominator (glist elist mainvar)
  (if (loop for ei in elist when (> ei 1) do (return t)) then
      (loop for coeff in (macsyma-coeffs
			     `((mtimes simp)
				. ,(loop for gi in glist
					 for ei in elist
					 when (> ei 1)
					   collect
					     (if (= ei 2) then
						 gi
					      else
					         `((mexpt simp)
						      ,gi ,(1- ei)))))
			     mainvar)
	    collect (sscalar coeff))
   else
      nil) )


;; Compute the partial-fraction decomposition of each of:
;;
;;                      x**k          
;;               -------------------      k=0,...,(dbound-1)
;;               (product of the gi0)
;;
;; Macsyma:$partfrac is called to get the k=0 case, and then each k=(j+1) case
;; is computed via modular shifting of the k=j case.  The result for a given k
;; and a given gi0 is a polynomial in the main variable.
;; The returned list is ordered first by gi0, then by k.
(defun getpartfracs (macglist glist mainvar dbound)
  (verbdo (princ "Starting partial fraction computation..."))
  (let* ( (zero (sscalar 0))
	  (partfracs (if (onep (length glist)) then
			 ;; special case for when only one factor
			 (list
			   (cons (sscalar 1)
				 (loop repeat (- (length (caar glist)) 2)
				       collect zero)))
		      else  ;more than one factor
		         ($partfrac
			     `((mexpt simp) ((mtimes simp) . ,macglist)  -1)
			       mainvar)))
	  (deglist (loop for (gi0) in glist collect (1- (length gi0))))
	  coeffs )
    (verbdo (terpr))
    (if (> (length glist) 1) then  ;more than one factor
	(setq partfracs
	      (loop for partfrac in partfracs
		    with d
		    when (or (patmatch partfrac '((mtimes . $) . $))
			     (patmatch partfrac '((mexpt . $) $ -1)))
		      do
			(setq d (car deglist)
			      deglist (cdr deglist))
			(setq coeffs
			      (if (patmatch partfrac '((mtimes . $) . $)) then
				  (partfrac-coeffs partfrac mainvar)
			       else
				  (list (sscalar 1))))
			(nconc coeffs
			       (loop repeat (- d (length coeffs))
				     collect zero))
		      and
			collect coeffs )))
    (loop for (gi0) in glist
	  for revgi = (reverse gi0)
	  for leadinverse = (sdiv (sscalar 1) (car revgi))
	  for partfrac in partfracs
	  collect
	    (cons partfrac
		  ;; Modular shifting for cases k=1,...,(dbound-1)
		  (loop repeat (1- dbound)
			with prevpart = (reverse partfrac)
			with multiplier
			do
		          (setq multiplier (stimes (car prevpart) leadinverse))
			collect
			  (reverse
			    (setq prevpart
				  (loop for xa = (cdr prevpart) then (cdr xa)
					for xb in (cdr revgi)
					collect
					  (ssub (or (car xa) zero)
						(stimes multiplier xb)))))))
	  finally
	    (verbdo (princ "Done computing partial fractions.") (terpr))) ))

;; Extract a partial-fraction coefficient from a Macsyma sub-expression.
;; This routine is based on faith in Macsyma's consistency, and there are
;; a couple of checks here in case something unexpected pops up.  The argument
;; PARTFRAC is an 'mtimes' expression, and it is assumed that there are either
;; two or three arguments with the possibilities being a constant, a polynomial,
;; or the gi0 denominator (which gets ignored).
(defun partfrac-coeffs (partfrac var)
  (if (> (length partfrac) 4) then
      (ferror "Partial fraction with more than 4 components."))
  (let ( (constant (second partfrac))
	 (poly     (third  partfrac)) )
    (if (and (listp constant) (not (patmatch constant '((rat . $) . $)))) then
	(if (> (length partfrac) 3) then
	    (ferror "Trouble in PARTFRAC-COEFFS."))
	(setq poly constant
	      constant 1))
    (if (patmatch poly '((mexpt . $) $ -1)) then
	(setq poly 1))
    (setq constant (sscalar constant))
    (loop for i = 0 then (1+ i)
	  with d = ($hipow poly var)
	  collect (stimes constant
			  (sscalar ($coeff poly var i)))
	  until (= i d)) ))


;##|
;#;;***************************************************************************
;#;; This large piece of commented-out code is a partial-fraction routine which
;#;; does not use Macsyma's $partfrac.  It works but is faster only for certain
;#;; types of problems.  There should be a selection routine to decide which
;#;; method to use, but I (GMI) don't have time to write it.
;#(defun getpartfracs (glist dbound numbound)
;#  (verbdo (princ "Starting partial fraction computation.") (terpr))
;#  (let* ( (zero (sscalar 0))
;#	  (partfracs (if (onep (length glist)) then
;#			 (list
;#			   (cons (sscalar 1)
;#				 (loop repeat (- (length (caar glist)) 2)
;#				       collect zero)))
;#		      else
;#		         (partfrac1 glist numbound))) )
;#    (loop for (gi0) in glist
;#	  for revgi = (reverse gi0)
;#	  for leadinverse = (sdiv (sscalar 1) (car revgi))
;#	  for partfrac in partfracs
;#	  collect
;#	    (cons partfrac
;#		  (loop repeat (1- dbound)
;#			with prevpart = (reverse partfrac)
;#			with multiplier
;#			do
;#		          (setq multiplier (stimes (car prevpart) leadinverse))
;#			collect
;#			  (reverse
;#			    (setq prevpart
;#				  (loop for xa = (cdr prevpart) then (cdr xa)
;#					for xb in (cdr revgi)
;#					collect
;#					  (ssub (or (car xa) zero)
;#						(stimes multiplier xb)))))))
;#	  finally
;#	    (verbdo (princ "Done computing partial fractions.") (terpr))) ))
;#(defun partfrac1 (glist bound)
;#  (let* ( (savemodulus  (prog1 *modulus*
;#			       (setq *modulus*
;#				     (powergreater (fixrandomprime)
;#						   (expt bound 2)))))
;#	  (mod-glist    (loop for (gi0) in glist
;#			      collect
;#			        (loop for coeff in gi0
;#				      collect (modularize (straightarg1 coeff)
;#							  *modulus*))))
;#	  (fullproduct  (loop for gi0 in mod-glist
;#			      with product = (list 1)
;#			      do  (setq product (polymult-mod product gi0))
;#			      finally (return product)))
;#	  (num-unknowns (1- (length fullproduct)))
;#	  (products     (loop for gi0 in mod-glist
;#			      collect (polydiv-mod fullproduct gi0)))
;#	  (right-hand-side (cons 1 (loop repeat (1- num-unknowns) collect 0)))
;#	  (linearsolutions (find-linear-solutions mod-glist products))
;#	  solution )
;#    (loop for gi0 in mod-glist
;#	  for product in products
;#	  for linearsol in linearsolutions
;#	  when (not (zerop linearsol))
;#	  do
;#	    (setq right-hand-side
;#		  (polysub-mod right-hand-side
;#			       (loop for xa in product
;#				     collect (timesmod linearsol xa))))
;#	    (setq num-unknowns (1- num-unknowns)))
;#    (if (> num-unknowns 0) then
;#	(setq solution
;#	      (linsolvemodp
;#		  (form-partfrac-system
;#			    mod-glist products right-hand-side num-unknowns)
;#		  *modulus*)))
;#    (verbdo (princ "    Recovering coefficients...") (terpr))
;#    (loop for gi0 in mod-glist
;#	  for linearsol in linearsolutions
;#	  with coeffs
;#	  with maxsize = 0
;#	  with bound = (isqrt (quotient *modulus* 2))
;#	  do
;#            (setq coeffs
;#	          (if (zerop linearsol) then
;#		      (loop repeat (1- (length gi0))
;#		            collect (sscalar ($padic_recover (car solution)
;#						             *modulus* bound))
;#		            do      (setq solution (cdr solution)))
;#	           else
;#	              (list (sscalar ($padic_recover linearsol
;#						     *modulus* bound)))))
;#	    (loop for xa in coeffs
;#		  for size = (rat-size (straightarg1 xa))
;#		  when (> size maxsize)
;#		    do (setq maxsize size))
;#	  collect coeffs
;#	  finally
;#	    (verbdo (princ "    Partfrac1 bound:          ")
;#		    (princ bound) (terpr)
;#		    (princ "    Largest number generated: ")
;#		    (princ maxsize) (terpr))
;#	    (setq *modulus* savemodulus)) ))
;#(defun rat-size (num)
;#  (if (patmatch num '((rat . $) . $)) then
;#      (max (abs (cadr num)) (abs (caddr num)))
;#   else
;#      (abs num)) )
;#(defun find-linear-solutions (glist products)
;#  (loop for gi0 in glist
;#	for product in products
;#	when (= (length gi0) 2)
;#	  collect
;#            (inversemod
;#	        (evalpoly-mod product
;#		              (divmod (times -1 (car gi0)) (cadr gi0))))
;#	else
;#	  collect 0) )
;#(defun evalpoly-mod (poly value)
;#  (loop for coeff in (reverse poly)
;#	with sum = 0
;#	do  (setq sum (addmod (timesmod sum value) coeff))
;#	finally (return sum)) )
;#(defun form-partfrac-system (glist products right-hand-side num-unknowns)
;#  (loop repeat num-unknowns
;#	for rhs in right-hand-side
;#	for revproducts first (loop for prod in products
;#				    collect (list (car prod)))
;#	                then  (progn
;#				(setq products (loop for prod in products
;#					             collect (cdr prod)))
;#				(loop for revprod in revproducts
;#				      for prod in products
;#				      collect (if prod then
;#						  (cons (car prod) revprod)
;#					       else
;#					          (cons 0 revprod))))
;#	collect
;#	  (nconc
;#	      (loop for gi0 in glist
;#		    for d = (1- (length gi0))
;#		    for revprod in revproducts
;#		    when (> d 1)
;#		      nconc
;#		        (loop repeat d
;#			      for revprodtail first revprod
;#					      then  (cdr revprodtail)
;#			      collect (or (car revprodtail) 0)))
;#	      (list rhs))
;#	finally
;#	  (verbdo (princ "    The system has ") (princ num-unknowns)
;#		  (princ " equations.") (terpr))) )
;#;;****************************************************************************
;#|#



;; In the program for fbar all minor variables are set to 0 and the resulting
;; program is converted to a Macsyma-format univariate polynomial via
;; $straighttosparse.  This polynomial is factored with Macsyma $factor and 
;; a re-arranged version of the Macsyma output is returned.
(defun univariatefactor (program varlist dbound denombound)
  (let ( (maxrandom (abs (cadr (car varlist))))
	 numbound
	 factors
	 constantfactor )
    ;; Find biggest translation constant and disguise minor variables as
    ;; zero constants.  Variable symbols are save in the straightarg2 slot
    ;; of the input instruction.
    (loop for (xi ai bi) in (cdr varlist)
	  for varsym = (straightarg1 xi)
	  do
            (setf (cdr xi) `(constant 0 ,varsym))
	    (setq maxrandom (max maxrandom (abs ai) (abs bi))))
    ;; Numerator bound for $straighttosparse
    (setq numbound (* denombound
		      (expt (* 2 maxrandom) dbound)
		      ($ratsimp `((%binomial simp) ,(+ dbound (length varlist))
						   ,dbound))))

    (setq factors
	  ($factor
	      ($straighttosparse program denombound 
			         `((mequal) $nb ,numbound)
				 `((mequal) $db ,dbound)
				 `((mequal) $norm no) )))

    ;; Restore input instructions...
    (loop for za in (cdr varlist)
	  for xi = (car za)
	  for varsym = (straightarg2 xi)
	   do
             (setf (cdr xi) `(input ,varsym)) )
    (setq factors
	  (cond
	    ((patmatch factors '((mplus . $) . $))    ;Irreducible poly
	        (list (list factors) (list 1)))
	    ((patmatch factors '((mexpt . $) . $))    ;Single Factor
	        (list (list (cadr factors)) (list (caddr factors))))
	    (t
	      (transpose
		  (loop for xa in (cdr factors)
			when (or (atom xa) (eq (caar xa) 'rat))
			  do (setq constantfactor xa)
			else
			  collect (if (eq (caar xa) 'mexpt) then
				      `(,(cadr xa) ,(caddr xa))
				   else
				      `(,xa 1)))))))
    (list numbound constantfactor (car factors) (cadr factors)) ))

 
;; Compute Product(gik**ei,i=1,...,r).
;; The only thing that has changed since we last computed this product is 
;; that one new homogeneous part has been added to each gi.  The product is
;; computed by staightforward repeated multiplication, and we have available
;; (in GPRODUCTS) the result of each intermediate multiplication as it
;; happened in the previous lifting step.
;; Each gi in glist and each product in gproducts is a list of homogeneous
;; parts, and each homogeneous part is a polynomial in the main variable.
(defun new-gproducts (glist elist gproducts)
  (loop for prod in gproducts
	for ei first (1- (car elist))
	       then  (if (zerop ei) then
			 (setq elist (cdr elist))
			 (1- (car elist))
		      else (1- ei))
	for gi = (progn (if (zerop ei) then (setq glist (cdr glist)))
			(car glist))
	with prevprod = (cons (list (sscalar 0)) (car glist))
	with carry = (caar glist)
        do
	  (rplaca prod
		  (spolyadd (car prod)
			    (setq carry
			          (spolyadd (spolymult carry (car (last gi)))
				            (spolymult (car gi)
						      (car (last prevprod)))))))
        collect
	  (setq prevprod
	      (cons (loop for xa in gi
			  for xb in (cdr (reverse prevprod))
			  with sum = (list (sscalar 0))
			    do (setq sum (spolyadd sum (spolymult xa xb)))
			  finally (return sum))
		    prod)) ))
     

;; Translate the given program as follows:
;;                 x1 <-- x1 + a1
;;                 xi <-- xi + bi*x1 +ai    i=2,3,...
;; where the ai and bi are random integers in the range -255 to 255 (excl. 0).
(defun translate-for-fact (program)
  (simulation-principle (emptyprogram)
    (loop for inst in (straightlist program)
	  with lastinst = (straightlast program)
	  with mainvar
	  with newinput
	  with ai
	  with bi    do 
        (setvalue inst
	    (caseq (straightop inst)
	      (input
		  (setq newinput (sinput (straightarg1 inst)))
		  (setvalue newinput (list (setq ai (random256))
					   (setq bi (random256))))
		  (if (null mainvar) then
		      (sadd (setq mainvar newinput) (sscalar ai))
		   else
		      (sadd newinput
			    (sadd (stimes (sscalar bi) mainvar)
				  (sscalar ai))))) 
	      (constant
		  (sscalar (straightarg1 inst))) 
	      (t
		  (funcall (get '(() plus sadd minus ssub
				     times stimes quotient sdiv)
				(straightop inst))
			   (straightvalue (straightarg1 inst))
			   (straightvalue (straightarg2 inst)))) ))
	  until (eq inst lastinst) )
    (currentprogram) ))


;; Append instructions to the beginning of the current program to reverse
;; what TRANSLATE-FOR-FACT did. 
(defun reverse-translate-for-fact (varlist mainvar)
  (let ( (a1 (cadr (car varlist)))
	 (saveprogram (currentlist))
	 (savecdr     (cdr $straightprogram)) )
    (setq $straightprogram nil
	  $straightzero    nil
	  $straightone     nil)
    (setq mainvar (ssub (sinput mainvar) (sscalar a1)))
    (loop for (oldinput ai bi) in (cdr varlist)
	  for xi = (straightarg1 oldinput) do
	(setf (straightop oldinput) 'minus)
	(setf (straightargs oldinput)
	      (list (sinput xi)
        	    (sadd (stimes (sscalar bi) mainvar)
		          (sscalar ai)))) )
    (setq $straightprogram
	  (cons (nconc (currentlist) saveprogram)
		savecdr)) ))

 
;; The 'incremental-' functions in the following macro are defined in 
;; the file Polyarith.
(defmacro incremental-operation (inst)
  `(funcall (get '(() plus incremental-add minus incremental-sub
		      times incremental-times quotient incremental-div)
		 (straightop ,inst))
	    ,inst
	    (straightcoeffvalue (straightarg1 ,inst))
            (straightcoeffvalue (straightarg2 ,inst))))

;; This routine computes the next homogeneous part of each instruction in the
;; current program.  Previously computed homogeneous parts are stored in the
;; 'coeff' value of each instruction.  On the first call (INIT-FLAG non-nil) 
;; both the constant and linear parts are computed.  This is just an on-line
;; version of DIRECT-COEFFS.
;; The new homogeneous parts for each instruction in polys is returned.
(defun incremental-coeffs (polys lastinst &optional init-flag)
  (let ( (zeroinst (sscalar 0)) )
    (if init-flag then
	(loop for inst in (currentlist) do
	    (setcoeffvalue inst
	        (caseq (straightop inst)
		  (input    (list inst zeroinst))
		  (constant (list zeroinst inst))
		  (t
         	    (setcoeffvalue inst
			(list
			  (funcall
			    (get '(() plus sadd minus ssub times stimes
				      quotient first-incremental-division)
				 (straightop inst))
			    ;; Arguments are constant coefficients of
			    ;; straightarg1 and straightarg2
			    (car (last
				   (straightcoeffvalue
				         (straightarg1 inst))))
			    (car (last
				   (straightcoeffvalue
				         (straightarg2 inst)))))))
		    (incremental-operation inst) )))
	      until (eq inst lastinst) )
     else
	 (loop for inst in (currentlist) do
	     (setcoeffvalue inst
		 (caseq (straightop inst)
		   ((input constant)
		     (cons zeroinst (straightcoeffvalue inst)))
		   (t
                     (incremental-operation inst))))
	       until (eq inst lastinst) ))
    (loop for xa in polys collect
	(car (straightcoeffvalue xa))) ))


(defun savestate1 (partial-fractions)
  (setq $straightpartfracs partial-fractions) )

(defun savestate2 (glist gproducts tau)
  (setq $glist glist)
  (setq $gproducts (cons (car (last gproducts)) $gproducts))
  (if tau then
      (setq $taulist (cons tau $taulist))) )

(defun savestate3 (varlist)
  (straighttopoly-all (currentlast))
  (setq $varlist
	`((mlist simp)
	   . ,(loop for (xi ai bi) in varlist
		    collect `((mlist simp) ,(straightarg1 xi) ,ai ,bi))))
  (setq $straightpartfracs
	`((mlist simp)
	   . ,(loop for partfrac in $straightpartfracs
		    collect
		      `((mlist simp)
			 . ,(loop for poly in partfrac
				  collect
				    `((mlist simp)
				       . ,(loop for coeff in poly
						collect
						  (straightvalue coeff))
				       ))))))
  (setq $glist
	`((mlist simp)
	   . ,(loop for gi in $glist
		    collect
		      `((mlist simp)
			 . ,(loop for gik in gi
				  collect
				    `((mlist simp)
				       . ,(loop for cikm in gik
					      collect (straightvalue cikm))
				       ))))))
  (setq $gproducts
	`((mlist simp)
	   . ,(loop for product in $gproducts
		    collect
		      `((mlist simp)
			 . ,(loop for poly in product
				  collect
				    `((mlist simp)
				       . ,(loop for coeff in poly
						collect
						  (straightvalue coeff))
				       ))))))
  (setq $taulist
	`((mlist simp)
	   . ,(loop for tau in $taulist
		    collect
	              `((mlist simp)
		         . ,(loop for coeff in tau
				  collect (straightvalue coeff)))))) )

(defun $testpartfracs (var &rest polys)
  (simulation-principle (emptyprogram)
    (let* ( (glist (loop for gi in polys
		         collect
			   (list (macsyma-to-straight-coeffs gi var))))
	    (partfracs (getpartfracs glist
				     (loop for (gi0) in glist
					   sum (1- (length gi0))))) )
      (straighttopoly-all (currentlast))
      `((mlist simp)
	 . ,(loop for xa in partfracs 
		  collect
		    `((mlist simp)
		       . ,(loop for xb in xa
				collect
				  `((mlist simp)
				     . ,(loop for xc in xb
					      collect (straightvalue xc))
				     ))))) )))

(defun $testnewgproducts (var &rest polys)
  (simulation-principle (emptyprogram)
    (let* ( (deglist (loop for gi in polys
			   collect ($hipow ($ratsimp gi) var)))
	    (merged-gcoeffs (loop for gi in polys
				  for di in deglist
				  nconc
				    (loop for j = 0 then (1+ j)
					  collect
					    (polytostraight1
					        ($ratcoef gi var j))
					  until (= j di))))
	    (lastinst (currentlast))
	    (firstcoeffs (incremental-coeffs
			     merged-gcoeffs lastinst 'initialize))
	    (glist (loop for di in deglist
			 with coeffs = merged-gcoeffs
			 collect
			   (list
			     (loop repeat (1+ di)
				   collect
				     (cadr (straightcoeffvalue (car coeffs)))
				   do      (setq coeffs (cdr coeffs))))))
	    (gproducts (loop for gi in (cdr glist)
			     with product = (caar glist)
			     collect
			       `( (,(sscalar 0))
			       	  ,(setq product
					 (spolymult product (car gi))))))
	    (elist (loop repeat (length polys) collect 1)) )
      (loop repeat (apply 'straightdegree merged-gcoeffs)
	    for newhomparts first firstcoeffs
			    then  (incremental-coeffs merged-gcoeffs lastinst)
	    do
	  (setq glist (loop for gi in glist
			    for di in deglist
			    collect
			      (cons (loop repeat (1+ di)
					  collect (car newhomparts)
					  do      (setq newhomparts
							(cdr newhomparts)))
				    gi) ))
	  (setq gproducts (new-gproducts glist elist gproducts)) )
      (setq lastinst
	    (coeffs-to-poly
	        (loop for gi in glist
		      with product = (list (sscalar 1))
		      do (setq product
			       (spolymult product
					  (loop for cik in gi
					        with sum = (list (sscalar 0))
					        do (setq sum (spolyadd sum cik))
					        finally (return sum))))
		      finally (return product))
		(sinput var)))
      (list (straightheader) (currentlist) lastinst) )))
