FUNCTION: sinterp - sparse multivariate modular polynomial interpolation
   
CALLING SEQUENCE:
   sinterp(f, x, t, p)
   
PARAMETERS:
   f - a Maple procedure
   x - a list of variables to be interpolated
   t - a positive integer
   p - a prime modulus
   
SYNOPSIS:   
- Assume that f is a procedure which, given n integers and a prime p, returns
  an integer in Z mod p.  Then sinterp returns a polynomial in x whose values
  modulo p equal those of f.  The list x must contain n variables.  The func-
  tion sinterp uses sparse interpolation.
   
- For the interpolation to succeed, the modulus p should be sufficiently large.
  If it is not, FAIL is returned.
   
- The degree bound t is a bound on the total degree of the polynomial being
  interpolated.
   
- This function should be defined by the command readlib(sinterp) before it is
  used.   
   
EXAMPLES:   
> readlib(sinterp):
> f := proc(x,y,p) local t1, t2;
>         t1 := x^2+y^2;
>         t2 := x^2-y^2;
>         t1 * t2 mod p;
>      end:
> sinterp(f, [x,y], 10, 101);
                                       4    4
                                  100 y  + x
   
SEE ALSO: interp

> interface(verboseproc=2);
> readlib(`sinterp`);

proc(f,x,t,p)
local a,b,c,d,e,i,j,k,l,m,n,r,u,v,z;
options `Copyright 1993 by Waterloo Maple Software`;
    if type(x,name) then RETURN(sinterp(f,[x],d,p)) fi;
    if not type(f,procedure) and type(x,list) and type(t,integer) or t < 0 or
        not type(p,integer) then
        ERROR(`invalid arguments`)
    fi;
    n := nops(x);
    for i to n do
        d[i] := readlib(dinterp)(f,n,i,t,p);
        if d[i] = FAIL then RETURN(FAIL) fi
    od;
    r := rand(p);
    v := ['r()' $ n];
    c := 0;
    for i to n do
        if type(c,`+`) then c := {op(c)} else c := {c} fi;
        if c = {0} then c := {1} fi;
        userinfo(2,sinterp,`number of terms in polynomial now`,nops(c),
            lprint('interpolating',x[i],`to degree`,d[i]),
            lprint(`maximum # of equations required`,nops(c)*(d[i]+1)));
        m := 0;
        a := table();
        for j to nops(c) do
            l := min(d[i],t-degree(c[j]));
            m := m+l+1;
            a[j] := c[j]*b[j,k]*x[i]^k $ (k = 0 .. l);
            a[j] := convert([a[j]],`+`)
        od;
        userinfo(2,sinterp,`number of equations generated`,m);
        v := subsop(1 = NULL,v);
        a := convert([a[k] $ (k = 1 .. nops(c))],`+`);
        u := indets(a) minus {op(x)};
        e := table();
        for j to m do
            z := ['r()' $ i];
            e[j] := traperror(f(op(z),op(v),p));
            if e[j] = lasterror then RETURN(FAIL) fi;
            z := {(x[k] = z[k]) $ (k = 1 .. i)};
            e[j] := (subs(z,a)-e[j]) mod p
        od;
        e := {e[k] $ (k = 1 .. m)};
        c := subs(readlib(`msolve/linear`)(e,u,p),a) mod p;
        if has(c,u) then RETURN(FAIL) fi
    od;
    RETURN(c)
end

