#-----------------------------------------------------------------------
# module: sparse.mpl
#
# author: Angel Diaz
# date: 6/15/96
#
# (c) Copyright 1996 Angel Diaz and Eirch Kaltofen, All Rights Reserved
# This code is provided WITHOUT WARRANTY either expressed or implied.
#-----------------------------------------------------------------------

read `bbpoly.mpl`:        # black box polynomial package
read `sparseaux.mpl`:     # auxiliary functions

#-----------------------------------------------------------------------
# function: sparseinterp_mod
# date: 6/15/96
#
# reference: the varibale names mirror those utilized in  the sparse 
#    conversion algorithm detailed in "Factorization of polynomials 
#    given by straight-line programs", E. Kaltofen, Randomness and 
#    Computation Vol 5, Advances in Computing Research, S. Micali ed., 
#    Greenwhich, Connecticut, 1989, pg 387--394
#
# description: This is an improved version of E. Kaltofen's sparse 
#    conversion algorithm. This improved version is due to A. Diaz
#    and E. Kaltofen. It takes advantage of what we call a 
#    "pruning" of monomials and a method for rapidly determining the 
#    degree of such monomials. A formal referece is TBA.
#-----------------------------------------------------------------------

sparseinterp_mod := proc(
   bb,        # a maple procedure that evaluates a polynomail modulo m
   varlist_i, # list of interpolation variables [ x1 , ... , xn ]
   deglist_i, # list of exact degress of each var [ deg(x1) , ... , deg(xn) ]
   degpoly_i, # total degree of the polynomial represented by the black box
   card_i,    # cardinality of the set from which to pick random numbers
   modulus_i
)
   local  n,          # number of variables x1 , ... , xn
          vars,       # augmented list of variables
          d0,         # bound on all of the degrees [ deg(x1) , ... , deg(xn) ]
          ii,e,d,zz,  # loop variables
          i,          # loop variable (loop over each elment of vars)
          k,          # loop variable (loop over the augmented Ji list)
          rnd,        # random number generator
          Ji,         # list of exponent vectors [ [e1,...,ei-1], ... ]
          nJi,        # augmented list of exponents
          ji,         # size of linear system (ji x ji)
          jibound,    # upper bound on ji ( ((d0+1)^n)*d0 )
          a,          # initial random elements
          l,          # list of "remaining degree" for each monomial
          t,          # number of terms
          p,          # random points p1,...,pi
          inputpnts,  # (p1)^k,...,(pi)^k,ai+1,...,an
          bbval,      # black box evaluated at (p1)^k,...,(pi)^k,ai+1,...,an
          prunedval,  # pruned poly evaluated at (p1)^k,...,(pi)^k,ai+1,...,an
          gki,        # adjusted black box evaluation ( bbval - prunedval ) 
          v,          # v1...v_jibound describes a transp. vandermonde matrix
          x,          # x1...x_jibound solution vector
          prunedpoly, # polynomial of fully determined monomials
          prunedans,  # return value of the "prune" function call
          prunedeg,   # "remaining degree" of a particular monomial
          bbcnt,      # number of black box calls
          poly;       # final interpolated polynomial (in Maple format)

   # initialize variables
   bbcnt      := 0; 
   prunedpoly := 0; 
   d0         := max( op( deglist_i ) );
   vars       := [ `z`, op( varlist_i) ];
   n          := nops( vars );
   rnd        := rand( 1..card_i );
   p          := array( 1..n ); 
   inputpnts  := array( 1..n );
   jibound    := ( ( d0 + 1)^n ) * d0;  
   v          := array( 1..jibound ); 
   x          := array( 1..jibound );
   prunedans  := array( 1..3 );
   t          := 1;
   Ji         := [ [ ] ];

   # select initial values a2,...,an
   a := array( 2..n ); 
   for i from 2 to n do 
      a[i] := rnd(); 
      inputpnts[i] := a[i]; 
   od;

   trace3(`sparseinterp_mod`,`random initial value a[`,2,n,a);

   # interpolate one more variable
   for i from 1 to n do
      l  := [ ]; 
      ji := 0;
      for ii from 1 to t do;
         if i = 1 then 
            l  := [ degpoly_i ]; 
            ji := degpoly_i + 1; 
         else 
            prunedeg := compprunedeg( Ji[ ii ], i - 1 );
            l        := [ op( l ), min( deglist_i[ i - 1 ], prunedeg ) ]; 
            ji       := ji + min( deglist_i [ i - 1 ], prunedeg ) + 1; 
         fi;
      od; 

      trace1(`sparseinterp_mod`,i,vars,t,l,ji,Ji);

      prunedans  := prune(l,Ji,vars,x);
      l          := prunedans[1];
      Ji         := prunedans[2];
      prunedpoly := prunedpoly + prunedans[3];

      t  := nops(Ji);
      l  := []; 
      ji := 0;
      for ii from 1 to t do;
         if i = 1 then
            l  := [ degpoly_i ];
            ji := degpoly_i + 1;
         else
            prunedeg := compprunedeg( Ji[ ii ], i - 1 );
            l        := [ op( l ), min( deglist_i[ i - 1 ], prunedeg ) ];
            ji       := ji + min( deglist_i[ i - 1 ],  prunedeg ) + 1;
         fi;
      od;

      trace2(`sparseinterp_mod`,t,l,ji,Ji,prunedpoly);

      # select random points p1,...,pk
      for ii from 1 to i do
          p[ii] := rnd();                    
      od;

      # compute gki
      for k from 1 to ji do

         for ii from 1 to i do;
            inputpnts[ii] := modp( p[ii]^( k - 1 ), modulus_i );
         od;

         trace3(`sparseinterp_mod`,`inputpnts[`,1,n,inputpnts);

         prunedval := modp( subs( ( vars[zz] = inputpnts[zz] *inputpnts[1]  ) 
                      $ ( zz = 1..n ), prunedpoly ), modulus_i );
         bbval     := homoeval_mod( bb, inputpnts, n, modulus_i );
         gki[k]    := modp( bbval - prunedval, modulus_i );
         bbcnt     := bbcnt + 1;

         trace4(`sparseinterp_mod`,bbval,prunedval,gki[k]);

      od; 
       
      # set up the ji by ji linear system
      k   := 1;
      nJi := [ ];
      # for every element in Ji (exponent vector)
      for e from 1 to t do 
         trace5(`sparseinterp_mod`,e,Ji);
         # for every degree
         for d from 0 to l[e] do
            v[k] := prodraisedpnts_mod( p , Ji[ e ], d, i, modulus_i );
            k := k + 1;
            if i = 1 then 
              nJi := [ op( nJi ), [ d ] ];
            else
              nJi := [ op( nJi ), [ op( Ji[ e ] ), d ] ]; 
            fi;
         od;
      od;
      
      trace3(`sparseinterp_mod`,`vandermonde entry v[`,1,ji,v);

      # solve vandermonde system 
      x  := vansolve_kl_mod( v, gki, ji, modulus_i );
      Ji := selectJi( x, nJi, k - 1);
      t  := nops( Ji );

      trace6(`sparseinterp_mod`,x,nJi);
   od;

   trace8(`sparseinterp_mod`,t,l,ji,Ji);

   poly := prunedpoly + subs( 'z' = 1, 
           converttomaple( x, nJi, k - 1, vars, n, modulus_i ) );

   trace7(`sparseinterp_mod`,`Kaltofen and Diaz sparse conversion`,
           poly ,bbcnt ); 

   RETURN( poly );
end:

#-----------------------------------------------------------------------
# functions: Tracing procedures
# date: 6/15/96
#-----------------------------------------------------------------------

trace1 := proc(
   procname_i, # procedure name
   i_i,        # variable index
   vars_i,     # list of variable 
   t_i,        # number of terms
   l_i,        #
   ji_i,       #
   Ji_i        # list of exponent vectors 
)
   if ( printlevel > 2 ) then
      printf(`%s : Working on variable %a\n`,procname_i,vars_i[i_i]);
      trace8(procname_i,t_i,l_i,ji_i,Ji_i);
      printf(`%s : Pruning currently interpolated polynomial\n`,procname_i);
   fi;
end:

trace2 := proc(
   procname_i,  # procedure name
   t_i,         # number of terms
   l_i,         #
   ji_i,        #
   Ji_i,        #
   prunedpoly_i # 
)
   if ( printlevel > 2 ) then
      trace8(procname_i,t_i,l_i,ji_i,Ji_i);
      printf(`%s : Pruned polynomial = %a\n`,procname_i,prunedpoly_i);
   fi;
end:

trace3 := proc(
   procname_i,   # procedure name
   comment_i,
   startindex_i,
   endindex_i, 
   pnts_i
)
   local i;
   if ( printlevel > 3 ) then
      for i from startindex_i to endindex_i do
         printf(`%s : %s%a] = %a\n`,procname_i,comment_i,i,pnts_i[i]);
      od;
   fi;
end:

trace4 := proc(
   procname_i,  # procedure name
   bbval_i,     # value of black box at a specified point
   prunedval_i, # value of the pruned polynomial at the same point
   adjvalue     # adjusted value of black box at the specified point
)
   if ( printlevel > 3 ) then
      printf(`%s : value of black box = %a\n`,procname_i,bbval_i);
      printf(`%s : value of pruned polynomial = %a\n`,procname_i,prunedval_i);
      printf(`%s : adjusted value of black box = %a\n`,procname_i,adjvalue);
   fi;
end: 

trace5 := proc(
   procname_i, # procedure name
   expindex_i, # a particular exponent vector index
   Ji_i        # list of exponent vectors
)
   if ( printlevel > 2 ) then
      printf(`%s : Working on exponent vector (Ji[%a]) = %a\n`,procname_i,
             expindex_i,Ji_i[expindex_i]); 
   fi;
end:

trace6 := proc(
   procname_i, # procedure name
   solution_i, # solution vector of the vandermonde system
   nJi_i
)
  if ( printlevel > 2 ) then
     printf(`%s : solution vector (x) = %a\n`,procname_i,solution_i);
     printf(`%s : augmented exponent vector (nJi) = %a\n`,procname_i,
            nJi_i);
  fi; 
end:

trace7 := proc(
   procname_i,     # procedure name
   algname_i,      # name of the interpolation algorithm
   poly_i,         # interpolated polynomial
   bbcnt_i         # number of black box calls
)
   if ( printlevel > 1 ) then
      printf(`%s : Algorithm   : %s\n`,procname_i,algname_i);
      printf(`%s : Interpolated polynomial (poly) = %a\n`,procname_i,poly_i);
      printf(`%s : Number of black box calls = %a\n`,procname_i,bbcnt_i);
   fi;
end:

trace8 := proc(
   procname_i, # procedure name
   t_i,        # number of terms
   l_i,        #
   ji_i,       #
   Ji_i        # list of exponent vectors
)
   if ( printlevel > 2 ) then
      printf(`%s : Number of terms (t) = %a\n`,procname_i,t_i);
      printf(`%s : Remaining degree of each monomial (l) = %a\n`,procname_i,
             l_i);
      printf(`%s : size of linear system (ji x ji) = %a\n`,procname_i,ji_i);
      printf(`%s : Set of exponent vectors (Ji) = %a\n`,procname_i,Ji_i);
   fi;
end:

