#-----------------------------------------------------------------------
# module: sparseaux.mpl
#
# author: Angel Diaz
# date: 6/15/96
#
# (c) Copyright 1996 Angel Diaz and Eirch Kaltofen, All Rights Reserved
# This code is provided WITHOUT WARRANTY either expressed or implied.
#-----------------------------------------------------------------------

with(linalg):   # Linear Algebra package

#-----------------------------------------------------------------------
# function: vansolve_zippel_mod
# date: 6/15/96
#
# reference: R. Zippel, "Interpolating polynomials from their values", 
#    J. Symbolic Comput., 9/3, 1990 pg. 375--403
#
# description: Richard Zippel's algorithm for solving a transposed 
#    vandermonde system of equations ( A x = a ). Variable names mirror 
#    thoses utilized in Kaltofen's and  Lakshman's paper referenced below.
#
# return: Vector x = transpose(x1,...,xn_i), where x = inverse(A)*a
#-----------------------------------------------------------------------

vansolve_zippel_mod := proc(
   v_i,       # list which describes a Vandermonde matrix A
   a_i,       # vector a = transpose(a1,...,an_i)
   n_i,       # dimension of A and a
   modulus_i
)
   local j,k,i, 
         B,      # array of polynomials
         x,      # solution vector
         xj;

   B := array(1..n_i); x := [];

   # compute the polynomials Bj(z)
   for j from 1 to n_i do
      B[j] := 1;
      for k from 1 to n_i do
         if  not( j = k ) then
            B[j] := modp( B[j] * ( ( 'z' - v_i[k] ) / ( v_i[j] - v_i[k] ) ),
                    modulus_i );
         fi;
      od;
   od;

   # compute solution vector from coefficients of Bj(z)
   for j from 1 to n_i do
      xj := modp( coeff( collect( B[j], 'z' ), 'z', 0 ) * a_i[1], 
                  modulus_i );
      for i from 2 to n_i do
         xj := modp( xj + coeff( collect( B[j], 'z' ), 'z', i-1) * a_i[i],
                     modulus_i );
      od;
      x := [op(x),xj];
   od;

  RETURN( eval( x ) );
end:

#-----------------------------------------------------------------------
# function: vansolve_kl_mod
# date: 6/15/96
#
# reference: E. Kaltofen and Yagati Lakshman, "Improved sparse 
#    multivariate polynomial interpolation algorithms",Proc. ISSAC '88, 
#    Springer Lect. Notes Comput. Sci., P. Gianni ed., pg. 467--474 
#
# description: Kaltofen's and Lakshman's efficient algorithm to solve 
#    a transposed vandermonde system of equations ( A x = a ). Variable
#    names mirror thoses utilized in Kaltofen's and  Lakshman's paper.
#
# return: Vector x = transpose(x1,...,xn_i), where x = inverse(A)*a
#-----------------------------------------------------------------------

vansolve_kl_mod := proc(
   v_i,       # list which describes a Vandermonde matrix A
   a_i,       # vector a = transpose(a1,...,an_i)
   n_i,       # dimension of A and a
   modulus_i
)
   local j,k,i, 
         B,BP,  
         D, 
         BD, 
         QN,     
         x; 

   # compute the polynomials B(z) and BP(z)
   B := modp( ( `z` - v_i[1] ), modulus_i );
   for i from 2 to n_i do
      B := modp( B*(`z` - v_i[i]), modulus_i );
   od;
   B  := modp( sort( expand( B ), `z` ) , modulus_i );
   BP := modp( sort( expand( diff( B, `z` ) ), `z` ), modulus_i );

   # compute the polynomials D(z) and B(z)*D(z)
   D := `z`^n_i * a_i[1];
   for i from 2 to n_i do
      D :=D + `z`^( n_i - i + 1 ) * a_i[i];
   od;
   BD := modp( sort( expand( B * D ), `z` ), modulus_i );

   # compute the polynomial QN(w)
   QN := coeff( BD, `z`, n_i + 1 );
   for i from n_i + 2 to 2 * n_i do
      QN := modp( QN + `w`^( i - n_i - 1 ) * 
            coeff( BD, `z`, i ), modulus_i );
   od;

   # compute solution vector
   x := [];
   for i from 1 to n_i do;
      x := [ op(x) , modp( subs( w = v_i[i], QN )  /
             subs( z = v_i[i], BP ), modulus_i ) ];
   od;

  RETURN( eval( x ) );
end:

#-----------------------------------------------------------------------
# function: homoeval_mod 
# date: 6/15/96
#
# description: evaluate the polynomial modulo m utilizing pntlist_i[1]
#    as the homogenizing variable.
#-----------------------------------------------------------------------

homoeval_mod := proc(
   blackbox_i,  # a maple procedure that evaluates a polynomail modulo m
   pntlist_i, 
   numpnts_i,
   modulus_i
)
   local i,
         newpntlist;  # adjusted list of points

   newpntlist := array(1..numpnts_i-1);
   for i from 2 to numpnts_i do
      newpntlist[i-1] := modp( pntlist_i[i]*pntlist_i[1], modulus_i);
   od;
   RETURN( blackbox_i( newpntlist, modulus_i) );
end:

#-----------------------------------------------------------------------
# function: prodraisedpnts_mod
# date: 6/15/96
#
# reference: this corresponds to the product b_k,1^e1*...b_k,i^d described
#    on page 389 of "Factorization of polynomials given by straight-line 
#    programs", E. Kaltofen, Randomness and Computation Vol 5, Advances 
#    in Computing Research, S. Micali ed., Greenwhich, Connecticut, 1989,
#  
# description: compute the product of each of the points raised to its
#    corresponding exponent. The last point (representing the next
#    varible to interpolate) is raised to d_i.
#
# returns: pnts_i[1]^(exps[1])*...* pnts_i[i_i-1]^(exps[i_i-1])* 
#          pnts_i[i_i]^(d_i)
#-----------------------------------------------------------------------

prodraisedpnts_mod := proc(
   pnts_i,   # list of points
   exps_i,   # list of corresponding exponents
   d_i,      # exponent of last point 
   i_i,      # number of pnts to "raise"
   modulus_i
)
   local j,
         prod;

   if i_i = 1 then
      prod := modp( pnts_i[ 1 ]^d_i, modulus_i );
   else
      prod := modp( pnts_i[ 1 ]^exps_i[ 1 ], modulus_i );
      for j from 2 to i_i-1 do
         prod := modp( prod * ( pnts_i[ j ]^exps_i[ j ] ), modulus_i );
      od;
      prod := modp( prod * pnts_i[ i_i ]^d_i, modulus_i );
   fi;

   RETURN( prod );
end:

#-----------------------------------------------------------------------
# function: selectJi 
# date: 6/15/96
#
# description: Select exponents from the full set of exponent vectors
#    which have a corresponding non zero solution in the solution 
#    vector
#
# returns: a list of exponent vectors
#-----------------------------------------------------------------------

selectJi := proc(
   x_i,     # solution vector
   nJi_i,   # full set of exponent vectors
   n_i      # length of nJi_i
)
   local i,
         Ji;

   Ji := [ ];
   for i from 1 to n_i do
      if x_i[ i ] <> 0 then
         Ji := [ op( Ji ), nJi_i[ i ] ];
      fi;
   od;

   RETURN( Ji );
end:

#-----------------------------------------------------------------------
# function: compprunedeg
# date: 6/15/96
#
# description: compute the total degree of the remaining variables in
#    a monomial given an exponent vector. The fist entry in the exponent
#    vector is the total degree of the monomial and the remaining 
#    entries are the degrees of the variables interpolated thus far.
#
# returns: a non negative integer, total degree of the remaining variables
#-----------------------------------------------------------------------

compprunedeg := proc(
   exps_i, # a list representing an exponent vector
   n_i     # length of list
)
   local i,
         d;

   # the first entry is the total degree of the monomial (extra variable)
   d := exps_i[ 1 ];

   # compute the total degree of the remaining variables (if 0 we can prune)
   for i from 2 to n_i do
      d := d - exps_i[ i ];
   od;

   RETURN( d );
end:

#-----------------------------------------------------------------------
# function: prune
# supportfunction: getmonomial
# date: 6/15/96
#
# description: from the list of "remaining degree" of each monomial, 
#    determine which monomials are fully interpolated and adjust the
#    status variables accordingly.
#
# returns: a list of an updated l and Ji as well as a polynomial of
#    pruned monomials.
#-----------------------------------------------------------------------

prune :=proc(
   l_i,         # list of "remaining degree" for each monomial
   Ji_i,        # list of exponent vectors
   varlist_i,   # list of variables
   x_i          # solution vector
)
   local i,          # loop variable
         t,          # number of terms
         newx,       # solution vector w/o zeros
         numsol,     # number of solutions
         prunepoly,  # polynomial of fully determined monomials
         newl,       # l w/o fully determined monomials 
         newJi,      # Ji w/o fully determined monomials
         pruneans;

    numsol    := nops( x_i );
    t         := nops( l_i );
    prunepoly := 0;
    pruneans  := [];
    newl      := []; 
    newJi     := [];
    newx      := [];

    for i from 1 to numsol do
       if x_i[i] <> 0 then
          newx := [ op( newx ), x_i[i] ];
       fi;
    od;

    for i from 1 to t do
       # if this monomial can be pruned
       if l_i[i] = 0 then
          prunepoly := prunepoly + newx[i] * getmonomial( Ji_i[i], varlist_i );
       else
          newl  := [ op( newl ), l_i[i] ];
          newJi := [ op( newJi ), Ji_i[i] ];
       fi;
    od;

    pruneans := [ newl, newJi, prunepoly ];
    RETURN( pruneans );
end:

getmonomial := proc(
   exps_i,
   vars_i
)
   local i,
         numexps,
         monomial;

   numexps  := nops(exps_i);
   monomial := 1;

   for i from 2 to numexps do
      monomial := monomial * vars_i[i]^exps_i[i];
   od;

   RETURN( monomial );
end:

#-----------------------------------------------------------------------
# function: converttomaple
# date: 6/15/96
#
# description: convert from our internal representation of a polynomial 
#    into a Maple format. 
#
# returns: a Maple format polynomial
#-----------------------------------------------------------------------

converttomaple := proc(
   x_i,        # solution vector
   Ji_i,       # list of exponent vectors
   ji_i,       # total number of exponents 
   varlist_i,  # list of variables
   numvars_i,  # number of variables
   modulus_i
)
   local i,j,  # loop variables
         poly, # Maple format polynomial
         m;    # product of variables raised to their corresp. exp.

   poly := 0;

   for i from 1 to ji_i do
      if x_i[i] <> 0 then
         m := 1;
         for j from 1 to numvars_i do
            m := m * varlist_i[j]^Ji_i[i][j];
         od;
         poly := modp( poly + x_i[i]*m, modulus_i );
      fi;
   od;

   RETURN( poly );
end:

